// package JavaCode;

/*
SLIDER: Local Short Reads DNA Sequence Alignment
Beta Release; Ver 0.1, 22th Mar. 2008
By: Nawar Malhis
Canada's Michael Smith Genome Sciences Centre
*/

public class Oligo implements Comparable<Oligo> {
	
	static final int A = 0;          // 
	static final int C = 1;          // 
	static final int G = 2;          // 
	static final int T = 3;          //
	
	static final char F = 'F';          // Forward direction
	static final char R = 'R';          // Backward direction
	static final char B = 'B';          // Both directions
	
	static final int MAXU = 111;
	
	static final byte FPT = 0x01;  //  PET/SET
	static final byte FS0 = 0x02;  //  Side 0
	static final byte FMP = 0x04;  //  MP
	static final byte FU0 = 0x08;  //  U0
	static final byte FRC = 0x10;  //  RC
	static final byte FUQ = 0x20;  //  Crisp
	static final byte FUM = 0x40;  //  UM
	static final byte FLQ = -128;  //  LQ

	static final int SEQ      = 0;
	static final int ID       = 1;
	static final int WEIGHT   = 2;
	static final int LOCATION = 3;
	static final int MATCH    = 4;
	static final int[] prbMask  = {0x88000000, 0x00880000, 0x00008800, 0x00000088};
	
    public static boolean bac = true;
    static final char[] nucl  = {'A', 'C', 'G', 'T', 'N'}; // used to convert int to char
	private static final long mask[] = {
		0x0000000000000003L, 0x000000000000000CL, 0x0000000000000030L, 0x00000000000000C0L,
		0x0000000000000300L, 0x0000000000000C00L, 0x0000000000003000L, 0x000000000000C000L,
		0x0000000000030000L, 0x00000000000C0000L, 0x0000000000300000L, 0x0000000000C00000L,
		0x0000000003000000L, 0x000000000C000000L, 0x0000000030000000L, 0x00000000C0000000L,
		0x0000000300000000L, 0x0000000C00000000L, 0x0000003000000000L, 0x000000C000000000L,
		0x0000030000000000L, 0x00000C0000000000L, 0x0000300000000000L, 0x0000C00000000000L,
		0x0003000000000000L, 0x000C000000000000L, 0x0030000000000000L, 0x00C0000000000000L,
		0x0300000000000000L, 0x0C00000000000000L, 0x3000000000000000L
		};
	public static final long fullMask[] = { 0x0000000000000000L,
		0x3000000000000000L, 0x3C00000000000000L, 0x3F00000000000000L, 0x3FC0000000000000L, 
		0x3FF0000000000000L, 0x3FFC000000000000L, 0x3FFF000000000000L, 0x3FFFC00000000000L, 
		0x3FFFF00000000000L, 0x3FFFFC0000000000L, 0x3FFFFF0000000000L, 0x3FFFFFC000000000L, 
		0x3FFFFFF000000000L, 0x3FFFFFFC00000000L, 0x3FFFFFFF00000000L, 0x3FFFFFFFC0000000L, 
		0x3FFFFFFFF0000000L, 0x3FFFFFFFFC000000L, 0x3FFFFFFFFF000000L, 0x3FFFFFFFFFC00000L, 
		0x3FFFFFFFFFF00000L, 0x3FFFFFFFFFFC0000L, 0x3FFFFFFFFFFF0000L, 0x3FFFFFFFFFFFC000L, 
		0x3FFFFFFFFFFFF000L, 0x3FFFFFFFFFFFFC00L, 0x3FFFFFFFFFFFFF00L, 0x3FFFFFFFFFFFFFC0L, 
		0x3FFFFFFFFFFFFFF0L, 0x3FFFFFFFFFFFFFFCL, 0x3FFFFFFFFFFFFFFFL 
		};
	public static char[] str;
    public static long compMask[] = {0x3FFFFFFFFFFFFFFFL, 0x3FFFFFFFFFFFFFFFL, 0x3FFFFFFFFFFFFFFFL};
    public static long code[][];
    public static int sortIndex = 0;
    public static long dTmp[] = {0, 0};
    public static int numberOfBACs;
    public static String[] BACs;
    public static int  sfi;
    public static int  sfw;
    public static int  sri;
    public static int  srw;
    
    public static long countSet;
    public static long countPetP;
    public static long countPetN;

    public long[] d;
    public float  weight;
    public int    start;
    public byte   ch;
    public int    size;
    public int    id1;
    public byte   flg;
    public byte   u0;
    public byte[][] prb = null;
    public short  prbSize;
    public short  petSize;

    public static void clearMask() {
		   for (int i = 0; i < Parameters.now; i++) {
			   compMask[i] = 0x3FFFFFFFFFFFFFFFL;
		   }	   
	   }
    /*/
	   // --- abs(num) < wordSize && abs(num) < size
	//*/
	   public  static void fillCode() { // to fill the matrix for coding string to long
		   
			code        = new long[4][Parameters.wordSize * 2];
			for (int bpIdx = 0; bpIdx < 4; bpIdx++) {
			    code[bpIdx][Parameters.wordSize * 2 - 2] = bpIdx;
			    code[bpIdx][Parameters.wordSize * 2 - 2] = code[bpIdx][Parameters.wordSize * 2 - 2] << (Parameters.wordSize * 2 - 2);
			    for (int i = Parameters.wordSize * 2 - 4; i >= 0; i -=2) {
				    code[bpIdx][i] = code[bpIdx][i+2] >> 2;
			    }
			    for (int i = 0; i < Parameters.wordSize; i++) {
				    code[bpIdx][i] = code[bpIdx][2*i];
			    }		
			}
		}
    public static int getBIdx(int x) {
		   int bidx = Parameters.wordSize - x % Parameters.wordSize - 1;;
		   return bidx;
	   }

	   public static int  getBpInt(char b) {
	    	if (b == 'A')
	    	    return A;
	    	if (b == 'C')
	    	    return C;
	    	if (b == 'G')
	    	    return G;
	    	if (b == 'T')
	    	    return T;
	    	// System.err.println("Error in getBpInt:\t" + b);
	    	return -1;
	   }
	   public static int  getBpInt(char b, int strand) {
		   if (strand > 0) {
		    	if (b == 'A')
		    	    return A;
		    	if (b == 'C')
		    	    return C;
		    	if (b == 'G')
		    	    return G;
		    	if (b == 'T')
		    	    return T;
		   } else {
		    	if (b == 'A')
		    	    return T;
		    	if (b == 'C')
		    	    return G;
		    	if (b == 'G')
		    	    return C;
		    	if (b == 'T')
		    	    return A;			   
		   }
	    	// System.err.println("Error in getBpInt:\t" + b);
	    	return -1;
	   }
    public static int getWIdx(int x) {
		   int widx = x / Parameters.wordSize;;
		   return widx;
	   }
    private static void setMask(int x) {
		   int w = x / Parameters.wordSize;
		   int i = x % Parameters.wordSize;
		   compMask[w] = compMask[w] & ~mask[i];
	   }

public Oligo() throws Exception {
	   d = new long[Parameters.now];
	   // u = new byte[nou]; 
	   start = -1;
	   weight = 1;
	   size   = Parameters.dbOligoSize;
	   prbSize = -1;
	   petSize = 0;
	   prb    = null;
	   id1    = 0;
	   flg    = 0;
	   for (int i = 0; i < Parameters.now; i++) {
		   d[i] = 0;
	   }
	   u0 = -1;
   }
   
   public Oligo(Oligo oligo) {
	   d = new long[Parameters.now];
	   u0 = oligo.u0; 
	   for (int i = 0; i < Parameters.now; i++) {
		   d[i] = oligo.d[i];
	   }
	   weight = oligo.weight;
	   start  = oligo.start;
	   ch     = oligo.ch;
	   size   = oligo.size;
	   prbSize = oligo.prbSize;
	   petSize = oligo.petSize;
	   prb     = oligo.prb;
	   id1    = oligo.id1;
	   flg    = oligo.flg;
	   if (prb != null) {
		   for(int i = 0; i < prbSize; i++) {
			   for(int b = 0; b < 4; b++) {
				   prb[b][i] = oligo.prb[b][i];
			   }
		   }
	   }
   }
   
   public Oligo(String s) throws Exception {
	   int bidx;
	   int widx;
	   int cat;
	   d      = new long[Parameters.now];
	   u0     = -1; 
	   weight = 1;
	   id1    = 0;
	   flg    = 0;
	   start  = -1;
	   size   = s.length();
	   prbSize = -1;
	   petSize = 0;
	   prb    = null;
	   for (int i = 0; i < size; i++) {
		   cat = getBpInt(s.charAt(i));
		   if (cat == -1)
			   cat = 0;
		   bidx = getBIdx(i);
		   widx = getWIdx(i);
		   // System.err.println(widx + "\t" + bidx + "\t" + s.charAt(i));
		   d[widx] = d[widx] | code [cat][bidx];
	   }
   }
   
   public void shift(int num) throws Exception { 
	   if (num > 0) {
		   long tmp;
		   int nw = getWIdx(size);
		   for (int i = 0; i < nw; i++) {
			   d[i] = d[i] << (num * 2);
			   tmp = d[i+1];
			   tmp = tmp >>> ((31 - num) * 2);
			   System.err.println(nw + "\t" + i + "\t" + tmp);
			   d[i] = d[i] | tmp;
		   }
		   d[nw] = d[nw] << (num * 2);
	   } 
	   size = size - Math.abs(num);
   }

   
   public void comp() throws Exception {
	   setRC(true);
	   for (int i = 0; i < Parameters.now; i++) {
		   d[i] = (~d[i]) & 0x3FFFFFFFFFFFFFFFL;
	   }
   }
   public int compareTo(Oligo o){
   if (sortIndex == SEQ) {
	   return this.equal(o, size);
   } else if (sortIndex == ID) {
	   if (o.id1 == id1) {
		   if (o.getSide() == getSide()) {
 			   if (o.weight == weight) {
 				   return 0;
 			   } else if (o.weight > weight) {
 				   return 1;
 			   } else {
 				   return -1;
 			   }
		   } if (isSide0()) {
			   return -1;
		   } else {
			   return 1;
		   }
	   } else if (o.id1 < id1) {
		   return -1;
	   } else {
		   return 1;
	   }
   } else if (sortIndex == WEIGHT) {
	   if (o.weight > weight) {
		   return 1;
	   } else if (o.weight < weight) {
		   return -1;
	   } else {
		   return 0;
	   }
   } else if (sortIndex == MATCH) {
	   if (o.getMatch() > getMatch()) {
		   return 1;
	   } else if (o.getMatch() < getMatch()) {
		   return -1;
	   } else {
		   return 0;
	   } //*/
   } else if (sortIndex == LOCATION) {
	   if (o.ch == ch) {
 		   if (o.start > start) {
 			   return -1;
 		   } else if (o.start < start) {
 			   return 1;
 		   } else {
 			   return 0;
 		   }
 	   } else if (o.ch > ch) {
 		   return -1;
 	   } else {
 		   return 1;
 	   }
   }
   return 0;	   
}
   private void deCodeIt() {
	   long a;
	   int tmp;
	   int nw   = getWIdx(size-1);
	   int[] x = new int[nw+1];
	   for (int i = 0; i < nw; i++) {
		   x[i] = Parameters.wordSize;
	   }
	   x[nw] = Parameters.wordSize - getBIdx(size-1);
	   str = new char[size];
	   for (int w = 0; w <= nw; w++) {
		   for (int i = 0; i < x[w]; i++) {
			   tmp = w * Parameters.wordSize + i;
			   a = d[w];
			   a = a << ((i + 1) * 2);
			   a = a >>> (62);
			   str[tmp] = nucl[(int) a];
		   }
	   }
   }

   public int equal(Oligo o, int sz) {
	   int nw = getWIdx(sz-1);
	   int[] x = new int[nw+1];
	   for (int i = 0; i < nw; i++) {
		   x[i] = Parameters.wordSize;
	   }
	   x[nw] = Parameters.wordSize - getBIdx(sz-1);
	   for (int i = 0; i <= nw; i++) {
		   if ((o.d[i] & fullMask[x[i]]) > (this.d[i] & fullMask[x[i]])) {
			   return -1;
		   } else if ((o.d[i] & fullMask[x[i]]) < (this.d[i] & fullMask[x[i]])) {
			   return 1;
		   }
	   }
	   return 0;
   }
   public int extend(Seq seq) throws Exception { // DATB
	   int i;
	   int misMatches = 0; // expandMismatches
	   int lastmm  = -2;
	   int bpat;
	   for (i = 0; i < seq.prbSize; i++) { // seq.prbSize
		   bpat = getBpInt(getBPat(i));
		   if (seq.p[i][bpat] < (1-Parameters.seqSegNucP)) {
			   if (misMatches >= Parameters.expandMismatches) {
				   break;
			   }
			   if (i == lastmm + 1) {
				   i--;
				   break;
			   }
			   misMatches++;
			   lastmm  = i;
		   } 
	   }
	   if (i - lastmm <= 2) {
		   size = lastmm;
		   misMatches--;
	   } else {
		   size = i;
	   } 
	   return size - misMatches - u0;
   }

   //*/
   public char getBPat(int x) {
	   char bp;
	   int i = Parameters.wordSize - getBIdx(x) -1;
	   int w = getWIdx(x);
	   long a;
	   a = d[w];
	   a = a << ((i + 1) * 2);
	   a = a >>> (62);
	   bp = nucl[(int) a];	   
	   return bp;
   }

   public byte getMPBat(int i) {
	   if (prb != null) {
		   byte mpb = 0;
		   for (byte b = 1; b < 4; b++) {
			   if (prb[b][i] > prb[mpb][i])
				   mpb = b;
		   }
		   return mpb;
	   } else {
		   return -1;
	   }
   }
   
   public byte getBPatByte(int x) {
	   int i = Parameters.wordSize - getBIdx(x) -1;
	   int w = getWIdx(x);
	   long a;
	   a = d[w];
	   a = a << ((i + 1) * 2);
	   a = a >>> (62);
	   // bp = nucl[(int) a];	   
	   return (byte) a;
   }

   private void getIdx(int fi, int ri) {
	   sfw = getWIdx(fi);
	   sfi = getBIdx(fi);
	   srw = getWIdx(ri);
	   sri = getBIdx(ri);
   }
   //*
   public float getMappingScore(byte uF0, byte uR0, float w) throws Exception{
	   return w;// + w;
   }

   public float getMatch() {
	   return size - u0;
   }

   public int getCompSide() {
	   if ((FS0 & flg) == 0) 
		   return 0;
	   else
		   return 1;
   }

   public int getSide() {
	   if ((FS0 & flg) == 0) 
		   return 1;
	   else
		   return 0;
   }

   public boolean isLQ() { // FMP
	   if ((FLQ & flg) == 0) 
		   return false;
	   else
		   return true;
   }

   public boolean isMP() { // FMP
	   if ((FMP & flg) == 0) 
		   return false;
	   else
		   return true;
   }
   public boolean isPET() {
	   if ((FPT & flg) == 0) 
		   return false;
	   else
		   return true;
   }

   public boolean isRC() { // FMP
	   if ((FRC & flg) == 0) 
		   return false;
	   else
		   return true;
   }
   
   public boolean isSide0() {
	   if ((FS0 & flg) == 0) 
		   return false;
	   else
		   return true;
   }
   public boolean isU0() {
	   if ((FU0 & flg) == 0) 
		   return false;
	   else
		   return true;
   }
  
   public boolean isUM() {
	   if ((FUM & flg) == 0) 
		   return false;
	   else
		   return true;
   }
   //*
   public boolean isUQ() {
	   if ((FUQ & flg) == 0) 
		   return false;
	   else
		   return true;
   }
 //*/
   public boolean isPrbLoaded() {
	   if (prb != null)
		   return true;
	   return false;
   }
   
   public void load(Oligo oligo) {
	   for (int i = 0; i < Parameters.now; i++) {
		   d[i] = oligo.d[i];
	   }
   }
   public byte match(Oligo o) {
	   byte l = MAXU;
	   int i;
	   
	   for (i = 0; i < Parameters.now; i++) {
		   if ((o.d[i] & compMask[i]) != (this.d[i] & compMask[i]))
			   break;
	   }
	   
	   if (i == Parameters.now)
		   return MAXU;
	   
	   l = matchOne(d[i], o.d[i], i);
	   
	   if (l < 100)
		   l += (i * 31);
	   
	   return l;
   }
   
   private byte matchOne(long l1, long l2, int j) {
	   byte i;
	   for (i = 30; i >= 0; i--) {
		   if ((((l1 ^ l2) & mask[i]) & compMask[j]) != 0) 
			   return (byte) (30 - i);
	   }
	   return MAXU;
   }

   public byte matchR(Oligo o) {
	   byte l = MAXU;
	   int i;
	   for (i = 0; i < Parameters.now; i++) {
		   if ((o.d[i] & compMask[i]) != (this.d[i] & compMask[i]))
			   break;
	   }
	   if (i == Parameters.now) {
		   return MAXU;
	   }
	   l = matchOne(d[i], o.d[i], i);
	   if (l < 100)
		   l += (i * 31);
	   return l;
   }
   
   public void revers() {
	   int fi;
	   int ri;
	   for (fi = 0; fi < size/2; fi++) {
		   ri = size - fi - 1;
		   swap(fi, ri);
	   }
   }
   public void revers_Comp() throws Exception {
	   comp();
	   revers();
   }
   public void setLQ(boolean lq) {
	   if (lq) {
		   flg = (byte) (flg | FLQ);
	   } else {
		   flg = (byte) (flg & ~FLQ);		   
	   }
   }
   
   public void setMP(boolean mp) {
	   if (mp) {
		   flg = (byte) (flg | FMP);
	   } else {
		   flg = (byte) (flg & ~FMP);		   
	   }
   }
   public void setPET(boolean pt) {
	   if (pt) {
		   flg = (byte) (flg | FPT);
	   } else {
		   flg = (byte) (flg & ~FPT);		   
	   }
   }
   
   public void setRC(boolean rc) {
	   if (rc) {
		   flg = (byte) (flg | FRC);
	   } else {
		   flg = (byte) (flg & ~FRC);		   
	   }
   }
   public void setSide(int s) {
	   if (s == 0) {
		   flg = (byte) (flg | FS0);
	   } else {
		   flg = (byte) (flg & ~FS0);		   
	   }
   }   

   public void setU0(boolean u0) {
	   if (u0) {
		   flg = (byte) (flg | FU0);
	   } else {
		   flg = (byte) (flg & ~FU0);		   
	   }
   }
   public void setUM(boolean um) {
	   if (um) {
		   flg = (byte) (flg | FUM);
	   } else {
		   flg = (byte) (flg & ~FUM);		   
	   }
   }
   //*
   public void setUQ(boolean uq) {
	   if (uq) {
		   flg = (byte) (flg | FUQ);
	   } else {
		   flg = (byte) (flg & ~FUQ);		   
	   }
   }
   //*/
   private void swap(int fi, int ri) {
	   long tmpf;
	   long tmpr;
	   getIdx(fi, ri);
	   tmpf = d[sfw] & mask[sfi];
	   tmpr = d[srw] & mask[sri];
	   d[sfw] = d[sfw] & ~mask[sfi];
	   d[srw] = d[srw] & ~mask[sri];
	   if (sfi < sri) {
		   tmpf = tmpf <<  (2 * (sri - sfi));
		   tmpr = tmpr >>> (2 * (sri - sfi));		   
	   } else if (sfi > sri) {
		   tmpf = tmpf >>> (2 * (sfi - sri));
		   tmpr = tmpr <<  (2 * (sfi - sri));
	   } 
	   d[sfw] = d[sfw] | tmpr;
	   d[srw] = d[srw] | tmpf;	   
   }
   
   public String toSeq() {
	   deCodeIt();
	   String sq = (new String(str)).substring(0, size);
	   return sq;
   }
   

   @Override
public String toString() {
	   deCodeIt();
	   String tmp;
	   String s = "";
	   String sq = "";
	   int sEnd;
	   char rn;

	   if (start >= 0) 
	   {
		   if(numberOfBACs > 0) {
			   sq = BACs[ch];
		   } else {
			   sq = Parameters.SEQ_VAR[ch];
		   }
		   if (isRC()) {
			   rn = 'R';
		   } else {
			   rn = 'F';
		   }
		   s =  sq + "\t" + start + "\t" + rn + "\t" + ((float) Math.round(weight * 1000)/1000) + "\t" + size + "\t" + id1 + "\t";
	   } else {
		   s =  ((float) Math.round(weight * 1000)/1000) + "\t" + size + "\t" + id1 + "\t";
	   } //*/
	   
	   if (isUQ())
		   s += "Crisp\t";
	   else if (isLQ())
		   s += "LQ\t";
	   else 
		   s += "GQ\t";

	   if (isMP())
		   s += "MPS\t";
	   else 
		   s += "PS\t";

	   if (isSide0())
		   s += "S0\t";
	   else 
		   s += "S1\t";

	   if (start >= 0) 
	   {
		   if (!isPrbLoaded()) {
			   if (isPET())
				   s += "PET\t";
			   else 
				   s += "SET\t";
		   }

		   if (isUM())
			   s += "UM\t";
		   else 
			   s += "MM\t";
		   s += "U" + u0;
		   s += "\t";
	   }
	   
	   if (isU0())
		   s += "u0\t";
	   else
		   s += "u1\t";

	   if (isPrbLoaded()) {
		   s += toText();
	   } else {
		   tmp = (new String(str)).substring(0, size);
		   for (int i = 0; i < size;) {
			   if ((i + 10) >= size) {
				   sEnd = size;
			   } else {
				   sEnd = i + 10;
			   }
			   s += " " + tmp.substring(i, sEnd);
			   i += 10;
		   }
	   }
	   // s += "\t" + (new String(str)).substring(0, size);
	   
	   return s;
   }
	
   public String toText() {
	   String ps = "SET\t";
	   String s = "";
	   String q = "\t";
	   String nms = "";
	   char base;
	   int iBase;
	   int iRefBase;
	   boolean first = true;;
	   int ix;
	   if (isRC()) {
		   for (int i = 0; i < size; i++) {
			   ix = size - i - 1;
			   iBase = getMPBP(ix);
			   base = nucl[iBase];
			   if (base != getBPat(i)) {
				   if (!first) {
					   nms += " ";
				   }
				   first = false;
				   iRefBase = getBpInt(getBPat(i));
				   if (prb[iRefBase][ix] <= (-Parameters.q_Cal[ix] + 2)) {
					   nms += "<" + ((char) getBPat(i)) + (i+1) + " (Q" + prb[iRefBase][ix] + ")>";
				   } else {
					   nms += ((char) getBPat(i)) + "" + (i+1);
				   }
				   base = Character.toLowerCase(base);
			   }
			   // s += "" + base;
			   if (i > 0 && i % 10 == 0) {
				   q += " ";
				   s += " ";
			   }
			   /*
			   if (i > 0 && i % 10 == 0)
				   s += " " + base;
			   else
				   //*/
			   s += "" + base;	
			   if (iBase < 4 && prb[iBase][ix] > -6) { // size - i - 1
				   q += "" + (char) ((int) prb[iBase][ix] + 64);
			   } else {
				   q += ((char) 59);
			   }
		   }
	   } else {
		   for (int i = 0; i < size; i++) {
			   iBase = getMPBP(i);
			   base = nucl[iBase];
			   if (base != getBPat(i)) {
				   if (!first) {
					   nms += " ";
				   }
				   first = false;
				   iRefBase = getBpInt(getBPat(i));
				   if (prb[iRefBase][i] <= (-Parameters.q_Cal[i] + 2)) {
					   nms += "<" + ((char) getBPat(i)) + "" + (i+1) + " (Q" + prb[iRefBase][i] + ")>";
				   } else {
					   nms += ((char) getBPat(i)) + "" + (i+1);
				   }
				   base = Character.toLowerCase(base);
			   }
			   if (i > 0 && i % 10 == 0) {
				   q += " ";
				   s += " ";
			   }

			   s += "" + base;	
			   
			   if (iBase < 4 && prb[iBase][i] > -6) { // size - i - 1
				   q += "" + (char) ((int) prb[iBase][i] + 64);
			   } else {
				   q += ((char) 59);
			   }
		   }
	   }
	   //*/
	   if (petSize != 0 && petSize != -1) {
		   ps = petSize + "\t";
		   if (petSize > 0) {
			   countPetP++;
		   } else {
			   countPetN++;
		   }
	   } else {
		   countSet++;
	   }
	   return ps + s + " " + q + "\t" + prbSize + "\t" + nms;
   }

	public static short baseComp(short b) {
		if (b == 4)
			return b;
		return (short) (~b & 3);
	}

	private int getMPBP(int i) {
		if (prb == null)
			return 4;
		int ret = 0;
		for (int b = 1; b < 4; b++) {
			if (prb[b][i] > prb[ret][i])
				ret = b;
		}
		return ret;
	}
}
