// package JavaCode;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

// import OligoFile.OligoFileTop;

/*
SLIDER: Local Short Reads DNA Sequence Alignment
Beta Release; Ver 0.1, 22th Mar. 2008
By: Nawar Malhis
Canada's Michael Smith Genome Sciences Centre
*/

public class ReferenceSequences {
	private float[]  readP = {0, 0, 0, 0};
    int[] bidxHist = new int[93];

	// public  float[]  qCal = new float[Parameters.qCal.length];
	private int      prbAvrSize;
	private int      readAvrSize;
	private float    prbAvrSizeF;
	private float    readAvrSizeF;
	private float[][][] prbCal = new float[2][93][81];
	public  float[][] q2prb  = new float[93][81];
	private int thisId;
	private int nextId;
	private byte thisSide;
	private byte[] aligne2Ref;
	private byte[] aligne2Map;
	private int    aligne2RefSize;
    private byte bfr[][][] = new byte[2][4][111];
    private int bfrLoadedSides;
    int[] iComp = {3, 2, 1, 0};
	// private byte nextSide;
	private short thisSize;
	// private short nextSize;
	private byte[] lastQs = new byte[400];
	private static boolean forward;

	public class Map {
		public int       size;
		public long      uSize;
		public String    fileName;
		public String    outf;

		public char[]    charSeq;
		public byte[]    refMap;
		public float[][][] Pb;

		public float[]   bidx;
		public int[][]   cnt;
		public float[]   info;
		public int[]     subSeq;
		private double[] avrCnt = {0, 0, 0};
		private float[]  avrCrv = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
		public short[]   gshift;
		public float[]   gval;
		public int[]     gcnt;
		public byte[][]  type;

		public int[][]   histogram1;
		public int[][]   histogram2;

		public int       cutLow;
		public int       cutHigh;

		int x0;
		int x1;

		public Runtime   rt;
		
		public void updatePb(int lx, int prbPoint, int direction) {
			float sum = 0;
			for (int b = 0; b < 4; b++) {
				Pb[direction][b][lx] = Pb[direction][b][lx] * readP[b] * ((1- Parameters.weightFactor) + Parameters.weightFactor * oligo[DATA].weight);
				sum += Pb[direction][b][lx];
			}
			for (int i = 0; i < 4; i++) {
				Pb[direction][i][lx] = Pb[direction][i][lx] / sum;
			}
		}
		
		public Map(int sz, String name, String fout, int refIdx) throws Exception {
			float tmpF;
			rt = Runtime.getRuntime();
			System.gc();
			fileName  = name;
			outf      = fout;
			size      = sz;

			charSeq   = new char [size];
			refMap    = new byte [size];
			bidx      = new float[size];
			subSeq    = new int  [size];
			gshift    = new short[size];
			gval      = new float[size];
			gcnt      = new int  [size];
			type      = new byte [3][size];		
			// pK_SNPs    = new byte [4][size];
			info      = new float[size];
			cnt       = new int  [3][size];
			Pb        = new float[3][4][size];
			
			for (int i = 0; i < size; i++) {
				info[i] = bidx[i] = gval[i] = gshift[i] = 0;
				type[0][i] = type[1][i] = type[2][i] = (byte) 0;
			}
			
			// 	------------------------------------- ---------------------------- ---------------------------- =======
			String line;
			char c;
			int sIdx;
			FileReader ch2File  = new FileReader(name);
			sIdx = 0;
			c = (char) ch2File.read();
			while(c != 10) {
				c = (char) ch2File.read();
			}
			while(ch2File.ready()) {
				c = (char) ch2File.read();
				if (c == '>') {
					// line = ch2File.re
					System.err.println();
				}
				if (c != 10 && c != 13) {
					charSeq[sIdx] = c;
					sIdx++;
				}
			}
			ch2File.close();
			// 	------------------------------------- ---------------------------- ---------------------------- =======
			for(int i = 0; i < size; i++) {
				gcnt[i] = cnt[0][i] = cnt[1][i] = cnt[2][i] = 0;
			}
	    	// reading sequencebility data into map
	    	String sapFl = Parameters.pathDB + Parameters.DATABASE + outf + ".map.36";
	    	DataInputStream		strIns = new DataInputStream(new BufferedInputStream(new FileInputStream(sapFl), 127000));
			for (int i = 0; i < size - Parameters.seedSize; i++) {
				refMap[i] = strIns.readByte();
			}
			strIns.close();
			String chName = Parameters.pathDB + Parameters.DATABASE + Parameters.ch_name1 + Parameters.SEQ_VAR[refIdx] + Parameters.ch_name2 + ".p";
			int i = 0;
			try {
				float[] pVal = {0, 0, 0, 0};
				float   sumPVal = 0;
				reader = new BufferedReader(new FileReader(chName));
				line = reader.readLine();
				while (line != null && line.length() > 2) {
					StringTokenizer tokens = new StringTokenizer(line);
					sumPVal = 0;
					for (int b = 0; b < 4; b++) {
						tmpF = ((float) Integer.parseInt(tokens.nextToken(), 10)) / 100;
						if (tmpF < 0.4 && tmpF > 0.1)
							tmpF = (float) 0.1;
						pVal[b] =  tmpF + Parameters.snpRatio / 3;
						sumPVal += pVal[b];
					}
					for (int b = 0; b < 4; b++) {
						Pb[2][b][i] = pVal[b] / sumPVal;
						/*
						if (Pb[2][b][i] > 0.1)
							pK_SNPs[b][i] = (byte) Math.round(100 * Pb[2][b][i]);
						else
							pK_SNPs[b][i] = 0;
						//*/
						Pb[1][b][i] = Pb[0][b][i] = (float) 0.25;
					}
					line = reader.readLine();	
					i++;
				}
				reader.close();
			} catch (Exception e) {
				System.err.println(">>>>>>>\tError reading " + chName + ", completed " + i);
			}
			System.err.println("Reading " + chName + ", completed " + i + "\t" + Parameters.q_Cal.length);
		}
		
		
		public void setNoise(int l, boolean s) {
			if (s) 
				type[2][l] = (byte) (type[2][l] | (byte) 2);
			else
				type[2][l] = (byte) (type[2][l] & (byte) -3);
		}

		private boolean isNoise(int l) {
			if ((type[2][l] & (byte) 2) == 2)
				return true;
			return false;
		} 
				
	
		public void setSnp(int dr, int l, boolean s) {
			switch(dr) {
			case 0: 
				if (s) {
					type[1][l] = (byte) (type[1][l] | (byte) -128);
				} else {
					type[1][l] = (byte) (type[1][l] & (byte) 127);
				}
				break;
			case 1:
				if (s) {
					type[1][l] = (byte) (type[1][l] | (byte) 64);
				} else {
					type[1][l] = (byte) (type[1][l] & (byte) -65);
				}
				break;
			case 2:
				if (s) {
					type[0][l] = (byte) (type[0][l] | (byte) -128);
				} else {
					type[0][l] = (byte) (type[0][l] & (byte) 127);
				}				
				break;
			default:
				System.err.println("Err in setSnp:\t" + l + ", dr:\t" + dr);
			}
		}

		//*
		private boolean isSnp(int dr, int l) {
			switch(dr) {
			case 0:
				if ((type[1][l] & (byte) -128) == -128)
					return true;
				return false;
			case 1: 
				if ((type[1][l] & (byte) 64) == 64)
					return true;
				return false;
			case 2:
                if (type[0][l] < 0)
                    return true;
                return false;
			default:
				System.err.println("Err in isSnp at:\t" + l + ", dr:\t" + dr);
			return false;
			}
		} //*/
				
		private void set_MPBP(int dr, int l, int mpbp) {

			if (dr == 0) {
				type[1][l] = (byte) (type[1][l] | (byte) 1);
				switch (mpbp) {
				case 0: 
					return;
				case 1: type[1][l] = (byte) (type[1][l] | (byte) 16); 
					return;
				case 2: type[1][l] = (byte) (type[1][l] | (byte) 32); 
					return;
				case 3: type[1][l] = (byte) (type[1][l] | (byte) 48); 
					return;
				default: 
					return; 
				}
			} else 
			if (dr == 1) {
				type[1][l] = (byte) (type[1][l] | (byte) 2);
				switch (mpbp) {
				case 0: 
					return;
				case 1: type[1][l] = (byte) (type[1][l] | (byte) 4); 
					return;
				case 2: type[1][l] = (byte) (type[1][l] | (byte) 8); 
					return;
				case 3: type[1][l] = (byte) (type[1][l] | (byte) 12); 
					return;
				default: 
					return; 
				}
			} 
			System.err.println("Err in setMPBP at:\t" + l + ", dr:\t" + dr);
		}
		
		public char getMPBP(int dr, int l) {
			int most = 0;
			if (map.cnt[dr][l] > 0) {
				for(int i = 1; i < 4; i++) {
					if (Pb[dr][i][l] > Pb[dr][most][l])
						most = i;
				}
				if (Pb[dr][most][l] > 0.9)
					return Character.toUpperCase(nucl[most]);
				return Character.toLowerCase(nucl[most]);
			} else {
				return '_';
			}
		}


		public int getMPBPInt(int direction, int l) {
			if (map.cnt[direction][l] > 0) {
				int most = 0;
				for(int i = 1; i < 4; i++) {
					if (Pb[direction][i][l] > Pb[direction][most][l])
						most = i;
				}
				return most;
			} else {
				return -1;
			}
		}

		public void setSnp(int l, boolean s) {
			if (s) {
				type[0][l] = (byte) (type[0][l] | (byte) -128);
			} else {
				type[0][l] = (byte) (type[0][l] & (byte) 127);
			}
		}
		
		public boolean detectSnp(int lx, int direction, float cut) {
			int bpAtRef = bp2Int(map.charSeq[lx]);
			if (Pb[direction][bpAtRef][lx] < cut)
				return true;
			return false;
		}
		
		public void setMPBP(int l, int mpbp) {
			switch (mpbp) {
			case 0: 
				return;
			case 1: type[0][l] = (byte) (type[0][l] | (byte) 32); 
				return;
			case 2: type[0][l] = (byte) (type[0][l] | (byte) 64); 
				return;
			case 3: type[0][l] = (byte) (type[0][l] | (byte) 96); 
				return;
			default: 
				return; 
			}
		}
		/*
		public char getMPBP(int l) {
			if (getDirection(l) != '.' || (snp_Sc[1][l] + snp_Sc[2][l] > 0 && isNeighbourhood(l))) {
				if ((type[0][l] & (byte) 96) == 96)
					return 'T';
				if ((type[0][l] & (byte) 64) == 64)
					return 'G';
				if ((type[0][l] & (byte) 32) == 32)
					return 'C';
				return 'A';
			} else {
				return '_';
			}
		} //*/
		// direction is 'F', 'R', 'B', or 'N' (N is undefined)
		public void setDirection(int l, char direc) {
			if (direc == 'F') {
				type[0][l] = (byte) (type[0][l] | (byte) 1);
			} else if (direc == 'R') {
				type[0][l] = (byte) (type[0][l] | (byte) 2);
			} else if (direc == 'B') {
				type[0][l] = (byte) (type[0][l] | (byte) 3);
			} else {
				type[0][l] = (byte) (type[0][l] & (byte) -4);
			}
		}
		
		public char getDirection(int d) {
			switch (d) {
			case 0: return 'F';
			case 1: return 'R';
			case 2: return 'B';
			default: return '.'; 
			}
		}
		//*
		public void setSnpEdge(int l, boolean e) {
			if (e) {
				type[0][l] = (byte) (type[0][l] | (byte) 4);
			} else {
				type[0][l] = (byte) (type[0][l] & (byte) -5);
			}
		}
	//*	
		public boolean isSnpEdge(int l) {
			if ((byte) (type[0][l] & (byte) 4) > 0)
				return true;
			else
				return false;
		}
//*/
		public void setGap(int l, boolean g) {
			if (g) {
				type[0][l] = (byte) (type[0][l] | (byte) 8);
			} else {
				type[0][l] = (byte) (type[0][l] & (byte) -9);
			}	
		}
		
		public boolean isGap(int l) {
			if ((byte) (type[0][l] & (byte) 8) > 0)
				return true;
			else
				return false;
		}

		public void setNeighbourhood(int l, boolean g) {
			if (g) {
				type[0][l] = (byte) (type[0][l] | (byte) 16);
			} else {
				type[0][l] = (byte) (type[0][l] & (byte) -17);
			}	
		}
		
		public boolean isNeighbourhood(int l) {
			if ((byte) (type[0][l] & (byte) 16) > 0)
				return true;
			else
				return false;
		}

		
		public void ceil(String fl1) throws Exception {
			System.err.println("RefSize:\t" + size + "\tUX:\t" + Parameters.trim(UXf));
			// Parameters.map_HistSize = (int) (3 * UX);
			int h1 = 0;
			System.err.println("Parameters.map_HistSize:\t" + Parameters.map_HistSize);
			histogram1= new int[2][Parameters.map_HistSize + 1];
			histogram2= new int[Parameters.map_HistSize + 1][Parameters.map_HistSize + 1];
			for (int i0 = 0; i0 < Parameters.map_HistSize + 1; i0++) {
				histogram1[1][i0] = histogram1[0][i0] = 0;
				for (int i1 = 0; i1 < Parameters.map_HistSize + 1; i1++) {
					histogram2[i0][i1] = 0;
				}
			}
			/*
			for(int i = 0; i < size; i++) {
				x0 = Math.round(snp_Sc[0][i]); 
				x1 = Math.round(snp_Sc[1][i] + snp_Sc[2][i]);
				if ( x0 >= Parameters.map_HistSize) {
					x0 = Parameters.map_HistSize;
				}
				if ( x1 >= Parameters.map_HistSize) {
					x1 = Parameters.map_HistSize;
				}
				h1 = x0 + x1;
				if (h1 > Parameters.map_HistSize)
					h1 = Parameters.map_HistSize;
				histogram1[0][h1]++;
				histogram2[x0][x1]++;
			}
			//*/
			histogram1[1][0] = histogram1[0][0];
			for (int i = 1; i <= Parameters.map_HistSize; i++) {
				histogram1[1][i] = histogram1[1][i-1] + histogram1[0][i];
			}
			String histogramsFileX1 = Parameters.pathOut + "ScoreResults/" + outf + ".X1_histograms"; 
			System.err.println(histogramsFileX1);
			BufferedWriter writerX1 = new BufferedWriter(new FileWriter(histogramsFileX1));
			// writerX1.write("Total Coverage <not normolized>:\n");
			for (h1 = 0; h1 <= Parameters.map_HistSize; h1++) {
				writerX1.write(h1 + "\t" + histogram1[0][h1] + "\t" + histogram1[1][h1] + "\n");
			}
			writerX1.write("\n");
			writerX1.close();

			String histogramsFileX2 = Parameters.pathOut + "ScoreResults/" + outf + ".X2_histograms"; 
			System.err.println(histogramsFileX1);
			BufferedWriter writerX2 = new BufferedWriter(new FileWriter(histogramsFileX2));
			// writerX2.write("X=U1, Y=U0 Coverage <not normolized>:\n");
			for (int i0 = 0; i0 <= Parameters.map_HistSize; i0++) {
				for (int i1 = 0; i1 <= Parameters.map_HistSize; i1++) {
					writerX2.write(histogram2[i0][i1] + "\t");
				}
				writerX2.write("\n");
			}
			writerX2.close();
			
			/*
			 *    From: contigs
			 *    To .: transposons
			 */
			// ====================================================================================================================
			boolean used = false;
			// String line;
			StringTokenizer st;
			int cStarts;
			int cEnds;
			int contigID = 0;
			// float mn = 0;
			float[][] ct = new float[2][size];
			for (int i = 0; i < size; i++) {
				ct[0][i] = ct[1][i] = 0;
			}
			BufferedReader reader = new BufferedReader(new FileReader(Parameters.pathOut + "ScoreResults/" + outf + ".contigs"));
			String transposonsFile = Parameters.pathOut + "ScoreResults/" + outf + ".transposons"; 
			// System.err.println("\n--> " + transposonsFile);
			BufferedWriter writerTransposons = new BufferedWriter(new FileWriter(transposonsFile));
			String line = reader.readLine();
			// System.err.println(line);
			while (true) {
				line = reader.readLine();
				st = new StringTokenizer(line);
				// System.err.println("countTokens:\t" + st.countTokens());
				if (st.countTokens() < 8) {
					// System.err.println(line);
					break;
				}
				contigID = new Integer(st.nextToken()).intValue();
				st.nextToken();
				st.nextToken();
				st.nextToken();
				cStarts = new Integer(st.nextToken()).intValue();
				cEnds   = new Integer(st.nextToken()).intValue();
				// System.err.println("cStarts:\t" + cStarts + "\tcEnds:\t" + cEnds);
				if (cEnds - cStarts >= Parameters.transposonMinSize) {
				/*
					for (int i = cStarts; i < cStarts + 5; i++) {
						mn += snp_Sc[0][i] + snp_Sc[1][i] + snp_Sc[2][i];
					}
				
					for (int i = cStarts + 5; i < cEnds; i++) {
						mn = mn - (snp_Sc[0][i - 5] + snp_Sc[1][i - 5] + snp_Sc[2][i - 5]) + (snp_Sc[0][i] + snp_Sc[1][i] + snp_Sc[2][i]);
						if (mn > (2 * UXf * 5) ) {
							if (!used) {
								for (int j = 0; j < 5; j++) {
									ct[0][i-j] += mn / (UXf * 5);
									ct[1][i-j]++;
								}
								used = true;
							} else {
								ct[0][i] += mn / (UXf * 5);
								ct[1][i]++;
							}
						} else {
							used = false;
						}
					} //*/
					for (int i = cStarts; i < cEnds; i++) {
						if (ct[1][i] > 1.5)
							ct[0][i] = ct[0][i] / ct[1][i];
					}
					used = false;
					int start = 0;
					float sum = 0;
					for (int i = cStarts; i < cEnds; i++) {
						// writerTransposonsTmp.write(i + "\t" + snpSc[0][i] + "\t" + snpSc[1][i] + "\n");
						if (ct[0][i] >= 2) {
							if (!used) {
								// writerTransposons.write(i + "\t");
								used = true;
								start = i;
							}
							sum += ct[0][i];
						} else {
							if (used) {
								if ((i - start) >= Parameters.transposonMinSize)
									writerTransposons.write(contigID + "\t" + start + "\t" + i + "\t" + (i - start) + "\t" + Parameters.trim(sum / (i - start)) + "\n");
								used = false;
								sum = 0;
							}
						}
					}
				}
			}
			writerTransposons.close();
			// ====================================================================================================================
			// snp_Sc = null;
			Pb    = null;
			// cnt   = null;
			System.gc();
			System.err.println("End Map");
		}
	}

	
	static final int    A             = 0;          // 
	static final int    C             = 1;          // 
	static final int    G             = 2;          // 
	static final int    T             = 3;          // 
	static final int    N             = 4;          // 
    static final int[]  comp          = {T, G, C, A, N}; // used to convert int bp to its complement int
    static final char[] nucl          = {'A', 'C', 'G', 'T', 'N', '>', '<', '.'}; // used to convert int to char
    static final int    ALL           = 1;
    static final int    CLEAN         = 0;
    static final int    MMSZ          = 31;
    static int[]        BPint;
    
	static final int    UM            = 0;         // 
	static final int    MM            = 1;         // 
	static final int    NM            = 2;         //
	static final int    DATA          = 0;
	static final int    DATB          = 1;

    static final int    non           = 2; // max number of Ns in an oligo
    static final int    wordSize      = 31; // length of each word
    static       int    subResult;

	static       int    ux            = 0;
	
	static public boolean display     = false;
	
	// private static long[][] code;                     // for coding c_oligo into oligo
	private float    UXf;
	private int      idx;                      // index in the chromosome
	private int		 oId;                      // oligo id for those with low weight
	private byte     chr;                      // chromosome index
	// private int      olid;
	private int[]    nList;                    // a list of the Ns in oligo
	private int      nCounter;                 // number of unknown nuc in a seq.
	private String   c_oligo;
	private char[]   bak_c_oligo;
	private Oligo[]  oligo;
	
	private OligoFile ofile;
	private Map       map;
	private OligoFile ofOne;
	private OligoFile ofTwo;
	private OligoFile ofNMatch;
	private OligoFile ofMatch;

	private Seq      seq;
	private long     numberOfOligos;
	private double   weightOfUM;
	private double   weightOfMM;
	private long     numberOfUQ;
	private long     numberOfUD;
	public  int[]    refSeqSize;

	private int      npbp;
	private int      mpbp;
	private boolean  in_a_gap;
	private long[]   mmHist;
	
	private char snpDirection;
	// private float[]  snpDirectCvr = {0, 0};

	private FileOutputStream fwPrb; 
	private BufferedOutputStream bwPrb;
	private DataOutputStream filePrb;

	private FileInputStream fiPrb; 
	private BufferedInputStream biPrb;
	private DataInputStream strInPrb;

	private BufferedReader reader;

	private Vector<Oligo>[] bunk; 
	private byte[][] idArray; 
	private int   IdCount;
	
	private int petGapMin;
	private int petGapMax;
	float spanAvr;
	public boolean   cutString = false;

	static public int  bp2Int(char b) {
		// System.err.print(b);
    	return BPint[java.lang.Character.toUpperCase(b) - 'A'];
    }
    static public char getBpChar(int i) {
    	return nucl[i];
    }

    public  void finalize()  throws Exception {
	}
	public  ReferenceSequences() throws Exception {

		nList       = new int[Parameters.dbOligoSize];
		c_oligo     = new String();
		bak_c_oligo = new char[Parameters.dbOligoSize+1];
		
		oligo       = new Oligo[2];
		oligo[DATA] = new Oligo();
		oligo[DATB] = new Oligo();
		
		Oligo.fillCode();
		
		bunk        = new Vector[2]; 
		bunk[0]     = new Vector<Oligo>(); 
		bunk[1]     = new Vector<Oligo>(); 

		numberOfOligos = 0;
		weightOfUM     = 0;
		weightOfMM     = 0;

		BPint = new int['Z' - 'A' + 1];
		for (char c = 'A'; c <= 'Z'; c++) {
			BPint[c-'A'] = -1;
		}
		BPint['A'-'A'] = A;
		BPint['C'-'A'] = C;
		BPint['G'-'A'] = G;
		BPint['T'-'A'] = T;
		seq = new Seq();
	}
	
	int getSkip(String fl) throws Exception {
		int s = 0;
		int count = 0;
		int[] x = {0, 0, 0, 0};
		int[] st= {0, 0, 0, 0};
		float stTotal = 0;
		// System.err.println(fl);
		BufferedReader reader;
		if ((fl.toString()).indexOf(".gz") >= 0) {
			reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(fl))));			
		} else {
			reader = new BufferedReader(new InputStreamReader(new FileInputStream(fl)));
		}
		for (String line = reader.readLine(); line != null; line = reader.readLine()) {
			if (count >= 100000)
				break;
			StringTokenizer tokens = new StringTokenizer(line.substring(0, 20));
		    for (int ii = 0; ii < 4; ii++) {
		    	x[ii] = Integer.parseInt(tokens.nextToken(), 10);
		    }
		    int b =  0;
		    for(int ii = 1; ii < 4; ii++) {
		    	if (x[ii] > x[b])
		    		b = ii;
		    }
		    st[b]++;
		    count++;
		}
		float min = 1;
		float max = 0;
		float tmp;
		stTotal = st[0] + st[1] + st[2] + st[3];
		for(int i = 0; i < 4; i++) {
			tmp = ((float) st[i])/stTotal;
			if (tmp < min)
				min = tmp;
			if(tmp > max)
				max = tmp;
			System.err.print(Parameters.trim(tmp) + "%\t");
		}
		if (min / max < 0.1)
			s = 1;
		reader.close();
		return s;
	}	
	
	public void generateOligosFromPRB_SET2(String[] prbFiles, int nid, int clusterSize) throws Exception {
		if (clusterSize > 0 && clusterSize < nid) {
			System.err.println("node " + nid + " greater than cluster size " + clusterSize);
			return;
		}
		OligoFile outOligoFile;
		int  id   = Parameters.maxId;
		int  idQc = 0;
		boolean  used = false;
		int rd;
		long totalIn  = 0;
		long totalOut = 0;
		long totalMPS = 0;
		long totalCrisp  = 0;
		long totalLQ  = 0;
		boolean mps = false;
		boolean uq  = false;
		boolean lq  = false;
		int  ok;
		numberOfOligos = 0;
		weightOfUM     = 0;
		weightOfMM     = 0;
		clearMeta();
		String outOlFile;
		String outPrb;
		if (clusterSize == 0) {
			outOlFile = Parameters.pathData + Parameters.outFile + ".ol" + ux;
			outPrb    = Parameters.pathData + Parameters.outFile;
		} else {
			outOlFile = Parameters.pathParallel + Parameters.outFile + "." + nid + ".ol" + ux;
			outPrb    = outOlFile + "/" + Parameters.outFile;			
		}
		outOligoFile = new OligoFile(outOlFile, "", OligoFile.W);

		openPrbOut(outPrb);
		for(int i=0; i<prbFiles.length; i++) {
			if (i % 20 == 0) {
				System.err.println("\nUsing " + Parameters.seedSize + " bases");
				if (Parameters.skipValue)
					System.err.println("#\tA%\tC%\tG%\tT%\tSkip\tout\tCrisp\tGQ\tLQ\ttotal");
				else
					System.err.println("#\tSkip\tout\tCrisp\tGQ\tLQ\ttotal");				
			}
        	if (clusterSize == 0 || (i) % clusterSize == (nid - 1)) {
        		System.err.print(i + "\t");
            	if (Parameters.skipValue)
            		Parameters.skipCycles = getSkip(prbFiles[i]);
            	System.err.print(Parameters.skipCycles);
    			// BufferedReader reader = new BufferedReader(new FileReader(prbFiles[i]));
    			BufferedReader reader;
    			if ((prbFiles[i].toString()).indexOf(".gz") >= 0) {
    				reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(prbFiles[i]))));			
    			} else {
    				reader = new BufferedReader(new InputStreamReader(new FileInputStream(prbFiles[i])));
    			}
    			for (String line = reader.readLine(); line != null; line = reader.readLine()) {
    				StringTokenizer tokens = new StringTokenizer(line);
    				seq.size = Parameters.seedSize;
    				if (Parameters.skipCycles > 0) {
    					seq.prbSize = (tokens.countTokens()/4) - Parameters.skipCycles;
    				} else {
    					seq.prbSize = tokens.countTokens()/4;
    				}
    		    	if (seq.size > seq.prbSize) {
    		    		seq.size = seq.prbSize;
    		    		System.err.println("Size reduced, ERROR");
    		    	}
    		    	// ------------------------------------------------ skip
    				for (int sk = 0; sk < Parameters.skipCycles; sk++) {
    			    	for (int ii = 0; ii < 4; ii++) {
    			    		tokens.nextToken();
    			    	}
    				}
    		    	// ------------------------------------------------
    				for (int is = 0; is < seq.prbSize; is++) {
    					for (int js = 0; js < 4; js++) {
    						seq._q[0][is][js] = (byte) Integer.parseInt(tokens.nextToken(), 10);
    					}
    				}
    				ok = seq.solvePrbLine(0);
    				totalIn++;
					uq = false;
					lq = false;
    				if (ok == 0) {
    					seq.getPS();  // equal sizes for now
						uq = true;
    				} else if (ok == 1) {
    					seq.getPS(); // must be short one
						Oligo.sortIndex = Oligo.WEIGHT;
						Collections.sort(seq.bunk);
						Oligo.sortIndex = Oligo.SEQ;
    				} else {
    					seq.getPSLQ(); // must be short one
						Oligo.sortIndex = Oligo.WEIGHT;
						Collections.sort(seq.bunk);
						Oligo.sortIndex = Oligo.SEQ;
    					lq = true;
    				}
					Iterator<Oligo> it = seq.bunk.iterator();
					rd = 0;
					mps = true;
					while (it.hasNext()) {
						if (rd > Parameters.maxMu) {
							break;
						} 	    							
						oligo[DATA] = (Oligo) it.next ();
						if (oligo[DATA].weight > Parameters.minWeight) {
							oligo[DATA].setMP(mps);
							oligo[DATA].setUQ(uq);
							oligo[DATA].setLQ(lq);
   							mps = false;
							oligo[DATA].id1 = id;
							oligo[DATA].setSide(0);
							outOligoFile.sort_WriteOligo(new Oligo(oligo[DATA]), Oligo.SEQ);
							used = true;
							totalOut++;
						} else {
							break;
						}
						rd++;
					}
					if (used) {
						seq.id = id;
						writePrb((byte) 0);	    					    		
						id--; // ++
						if(lq)
							totalLQ++;
						totalMPS++;
						if(uq)
							totalCrisp++;
					} else {
						idQc++;
					}
					used = false;
    			}
    			reader.close();
    			System.err.println("\t" + totalMPS + "\t" + totalCrisp + "\t" + (totalMPS - totalCrisp - totalLQ) + "\t" + totalLQ +
    					"\t" + totalOut + "\t" + prbFiles[i]);
        	}
		} 
		closePrbOut();
		Oligo.sortIndex = Oligo.SEQ;
		outOligoFile.saveSBunk(Oligo.SEQ);
		outOligoFile.minID = id;
		outOligoFile.dfn = outOligoFile.dfns;
		outOligoFile.close();
		outOligoFile = new OligoFile(outOlFile, "", OligoFile.R);
		if (Parameters.contDB) {
			outOligoFile.joinFilterContaminationSort();
		} else {
			outOligoFile.joinSort();
		}
		outOligoFile.close();
	    System.out.println(
	    		"\nTotal lines .....................:\t" + totalIn + 
	    		"\nLines removed by QC .............:\t" + idQc + 
	    		"\nNumber of MPS only lines ........:\t" + totalMPS + 
	    		"\nNumber of UQ lines ..............:\t" + totalCrisp +
	    		"\nFinal number of Oligos ..........:\t" + totalOut +
	    		"\nMultiplication ratio of UD Oligos:\t" + ((float) (totalOut - (totalCrisp + totalLQ)) / (float) (Parameters.maxId - id - (totalCrisp + totalLQ)))
	    		);
	  //*/

	}

	public void generateOligosFromPRB_PET2(String[] prbFiles, int nid, int clusterSize) throws Exception {
		if (clusterSize > 0 && clusterSize < nid) {
			System.err.println("node " + nid + " greater than cluster size "
					+ clusterSize);
			return;
		}
		OligoFile outOligoFile;
		int id = Parameters.maxId;
		int oldId = id;
		int idQc = 0;
		boolean used = false;
		int rd;
		int ok = 0;
		long totalIn = 0;
		long totalOut = 0;
		long totalMPS = 0;
		long totalCrisp = 0;
		long totalLQ = 0;
		boolean mps = false;
		boolean uq;
		boolean idUsed;
		boolean lq  = false;
		// int rid = 0;
		numberOfOligos = 0;
		weightOfUM = 0;
		weightOfMM = 0;
		clearMeta();

		String outOlFile;
		String outPrb;
		if (clusterSize == 0) {
			outOlFile = Parameters.pathData + Parameters.outFile + ".ol" + ux;
			outPrb    = Parameters.pathData + Parameters.outFile;
		} else {
			outOlFile = Parameters.pathParallel + Parameters.outFile + "." + nid + ".ol" + ux;
			outPrb    = outOlFile + "/" + Parameters.outFile;			
		}

		outOligoFile = new OligoFile(outOlFile, "", OligoFile.W);

		openPrbOut(outPrb); 
 
		for (int i = 0; i < prbFiles.length; i++) {
			if (clusterSize == 0 || (i) % clusterSize == (nid - 1)) {
				if (i % 20 == 0) {
					System.err.println("\nUsing " + Parameters.seedSize + " bases");
					if (Parameters.skipValue)
						System.err.println("#\tA%\tC%\tG%\tT%\tSkip\tout\tCrisp\tGQ\tLQ\ttotal");
					else
						System.err.println("#\tSkip\tout\tCrisp\tGQ\tLQ\ttotal");				
				}
				System.err.print(i + "\t");
				if (Parameters.skipValue)
					Parameters.skipCycles = getSkip(prbFiles[i]);
				System.gc();
				System.err.print(Parameters.skipCycles);
				// BufferedReader reader = new BufferedReader(new FileReader(prbFiles[i]));
    			BufferedReader reader;
    			if ((prbFiles[i].toString()).indexOf(".gz") >= 0) {
    				reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(prbFiles[i]))));			
    			} else {
    				reader = new BufferedReader(new InputStreamReader(new FileInputStream(prbFiles[i])));
    			}
				for (String line = reader.readLine(); line != null; line = reader.readLine()) {
					idUsed = false;
					oldId = id;
					split(line);
					for (byte l = 0; l < 2; l++) {
						ok = seq.solvePrbLine(l);
						uq   = false;
						lq   = false;
						mps  = true;
						used = false;
						totalIn++;
						if (ok == 0) {
							seq.getPS();
							uq = true;
						} else if (ok == 1) {
							seq.getPS();
							Oligo.sortIndex = Oligo.WEIGHT;
							Collections.sort(seq.bunk);
							Oligo.sortIndex = Oligo.SEQ;
						} else {
							seq.getPSLQ();
							lq = true;
							Oligo.sortIndex = Oligo.WEIGHT;
							Collections.sort(seq.bunk);
							Oligo.sortIndex = Oligo.SEQ;
						}
						Iterator<Oligo> it = seq.bunk.iterator();
						rd = 0;
						while (it.hasNext() && rd <= Parameters.maxMu) {
							oligo[DATA] = (Oligo) it.next();
							if (oligo[DATA].weight > Parameters.minWeight) {
								oligo[DATA].setMP(mps);
								oligo[DATA].setUQ(uq);
								if (mps) {
									mps = false;
								}
								oligo[DATA].id1 = oldId;
								oligo[DATA].setSide(l);
								oligo[DATA].setLQ(lq);
								// System.err.println(oligo[DATA]);
								// Thread.sleep(2000);
								outOligoFile.sort_WriteOligo(new Oligo(oligo[DATA]), Oligo.SEQ);
								used   = true;
								idUsed = true;
								totalOut++;
							} else {
								break;
							}
							rd++;
							seq.id = oldId;
						}
						if (used) {
							writePrb(l);
							if(lq)
								totalLQ++;
							totalMPS++;
							if(uq)
								totalCrisp++;
						} else {
							idQc++;
						}
					}
					if (idUsed) {
						id--;
					}
				}
				reader.close();
    			System.err.println("\t" + totalMPS + "\t" + totalCrisp + "\t" + (totalMPS - totalCrisp - totalLQ) + "\t" + totalLQ +
    					"\t" + totalOut + "\t" + prbFiles[i]);
			}
		}
		closePrbOut();
		Oligo.sortIndex = Oligo.SEQ;
		outOligoFile.saveSBunk(Oligo.SEQ);
		outOligoFile.minID = id - 1;
		outOligoFile.dfn = outOligoFile.dfns;
		outOligoFile.close();
		//*/
		outOligoFile = new OligoFile(outOlFile, "", OligoFile.R);
		outOligoFile.joinSort();
		outOligoFile.close();
		System.out.println("\nTotal lines .....................:\t"
				+ totalIn
				+ "\nLines removed by QC .............:\t"
				+ idQc
				+ "\nNumber of IDs ...................:\t"
				+ (Parameters.maxId - id)
				+ "\nNumber of MPS only sides ........:\t"
				+ (totalMPS)
				+ "\nNumber of Crisp sides ...........:\t"
				+ totalCrisp
				+ "\nFinal number of Oligos ..........:\t"
				+ totalOut
				+ "\nMultiplication ratio of UD Oligos:\t"
				+ ((float) (totalOut - totalCrisp) / (float) (Parameters.maxId
						- id - totalCrisp)));
		// */
	}
	
	public  long statMPS(String prbPath) throws Exception {
		int[][] histogram = new int[2][Parameters.seedSize + 1];
		int[][] mpsHis = new int[3][101]; // [0][] for all; [1][] for pass; [2][] for LQ
		float[][] infoHist = new float[2][Parameters.seedSize + 1];
		float sprb;
		for(int i = 0; i < histogram[0].length; i++) {
			histogram[1][i] = histogram[0][i] = 0;
		}
		for(int i = 0; i < mpsHis[0].length; i++) {
			mpsHis[2][i] = mpsHis[1][i] = mpsHis[0][i] = 0;
		}
		long sumUDin  = 0;
		long GQ         = 0;
		long GQTotal    = 0;
		long LQ         = 0;
		long LQTotal    = 0;
		long CRISP      = 0;
		long CRISPTotal = 0;
		long IN         = 0;
		long INTotal    = 0;
		long OUT        = 0;
		long OUTTotal   = 0;
		long NCBin      = 0;
		long NCBgo      = 0;
		long NCBlq      = 0;
		long NCBinTotal = 0;
		long NCBgoTotal = 0;
		long NCBlqTotal = 0;
		
		int length = 0;
		clearMeta();
		int[][]   histo  = new int[2][Parameters.seedSize];
		float[][] cnt    = new float[2][Parameters.seedSize];
		float[][] histoW = new float[2][Parameters.seedSize];
		int[] sumLQBases = {0, 0};
		
		for (int i = 0; i < histo[0].length; i++) {
			histoW[1][i] = histoW[0][i] = cnt[1][i] = cnt[0][i] = histo[1][i] = histo[0][i] = 0;
		}
		BufferedWriter writerLst = new BufferedWriter(new FileWriter(Parameters.pathData + "stat.lst." + ((int) (100 * Parameters.seqSegNucP)) + "." + Parameters.maxUDN));
		BufferedWriter writerCluster = new BufferedWriter(new FileWriter(Parameters.pathData + "files.list"));
		writerLst.write("Dataset:\t" + Parameters.DATASET + "\n");
		writerLst.write("Using " + Parameters.seedSize + " bases, no skip\n");
		writerLst.write("Length\tSkip\tIn\tNCBin%\tRmv\tOut\tCrisp\tGQ\tLQ\tNCBgq%\n");

		int fid = 0;
		int rid = 0;
		File fl = new File(prbPath);
		System.err.println("PRB_Path:\t" + prbPath);
		File[] dirs = fl.listFiles();
		for (int dd = 0; dd < dirs.length; dd++) {
			if(dirs[dd].isDirectory()) {
				// System.err.println("Dir:\t" + dirs[dd]);
		        File[] files = dirs[dd].listFiles();
		        for(int i=0; i<files.length; i++) {
		        	LQ    = 0;
		        	GQ    = 0;
		        	CRISP = 0;
		        	IN    = 0;
		        	OUT   = 0;
		        	NCBin = 0;
		        	NCBgo = 0;
		        	NCBlq = 0;
		            if(!files[i].isDirectory() && (files[i].toString()).indexOf("prb") >= 0) {
		            	writerCluster.write(files[i] + "\n");
		            	fid++;
		            	if (Parameters.skipValue) {
		            		if (fid % 20 == 1) {
				            	System.err.println("\nUsing " + Parameters.seedSize + " bases");
		            			System.err.println("A%\tC%\tG%\tT%\tLength\tSkip\tIn\tNCBin%\tRmv\tOut\tCrisp\tGQ\tLQ\tNCBgq%\t");
		            		}
		            		Parameters.skipCycles = getSkip("" + files[i]);
		            	} else {
		            		if (fid % 20 == 1) {
				            	System.err.println("\nUsing " + Parameters.seedSize + " bases, no skip");
		            			System.err.println("Length\tSkip\tIn\tNCBin%\tRmv\tOut\tCrisp\tGQ\tLQ\tNCBgq%\t");
		            		}
		            	}
		    			// BufferedReader reader = new BufferedReader(new FileReader(files[i]));
		            	// System.err.println(files[i]);
		    			BufferedReader reader;
		    			if ((files[i].toString()).indexOf(".gz") >= 0) {
		    				reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(files[i]))));			
		    			} else {
		    				reader = new BufferedReader(new InputStreamReader(new FileInputStream(files[i])));
		    			}
		    			rid = 0;
		    			for(int ia = 0; ia <= Parameters.seedSize; ia++) {
		    				infoHist[0][ia] = infoHist[1][ia] = 0;
		    			}
		    			for (String line = reader.readLine(); line != null; line = reader.readLine()) {
		    				if (Parameters.ETag == Parameters.PET) {
			    				length = split(line);
			    				IN++;
			    				seq.solvePrbLine(0); 
			    				seq.infoContent(0);
			    				for(int k = 0; k < Parameters.seedSize; k++) {
			    					infoHist[0][k] += seq.infoc[k];
			    				}
			    				histogram[0][seq.totalUd]++;
			    				NCBin += seq.totalUd;
		    					if(seq.totalUd <= Parameters.maxUDN) {
		    						OUT++;
			    					if(seq.totalUd == 0) {
			    						CRISP++;
			    					} else {
			    						GQ++;
			    						NCBgo += seq.totalUd;
			    					}
		    						sprb = seq.getP_MPS(histo, histoW, cnt, 0);
			    					mpsHis[1][(int) (100 * sprb)]++;
		    					} else {
			    					sprb = seq.getP_MPS(histo, histoW, cnt, 1);
		    						if(sprb > Parameters.minWeight) {
		    							histogram[1][seq.totalUd]++;
			    						LQ++;
			    						OUT++;
			    						NCBlq += seq.totalUd;
				    					mpsHis[2][(int) (100 * sprb)]++;		    							
		    						}
		    					}
		    					mpsHis[0][(int) (100 * sprb)]++;
			    				IN++;
			    				seq.solvePrbLine(1); 
			    				seq.infoContent(1);
			    				for(int k = 0; k < Parameters.seedSize; k++) {
			    					infoHist[1][k] += seq.infoc[k];
			    				}
			    				histogram[0][seq.totalUd]++;
			    				NCBin += seq.totalUd;
		    					// sprb = seq.getP_MPS(histo, histoW, cnt);
		    					if(seq.totalUd <= Parameters.maxUDN) {
		    						OUT++;
			    					if(seq.totalUd == 0) {
			    						CRISP++;
			    					} else {
			    						GQ++;
			    						NCBgo += seq.totalUd;
			    					}
			    					sprb = seq.getP_MPS(histo, histoW, cnt, 0);
			    					mpsHis[1][(int) (100 * sprb)]++;
		    					} else {
			    					sprb = seq.getP_MPS(histo, histoW, cnt, 1);
		    						if(sprb > (Parameters.minWeight)) {
		    							histogram[1][seq.totalUd]++;
			    						LQ++;
			    						OUT++;
			    						NCBlq += seq.totalUd;
				    					mpsHis[2][(int) (100 * sprb)]++;		    							
		    						}
		    					}
		    					mpsHis[0][(int) (100 * sprb)]++;
		    				} else {
		    					length = split(line);
			    				IN++;
			    				seq.solvePrbLine(0); 
			    				seq.infoContent(0);
			    				for(int k = 0; k < Parameters.seedSize; k++) {
			    					infoHist[0][k] += seq.infoc[k];
			    				}
			    				histogram[0][seq.totalUd]++;
			    				NCBin += seq.totalUd;
		    					// sprb = seq.getP_MPS(histo, histoW, cnt);
		    					if(seq.totalUd <= Parameters.maxUDN) {
		    						OUT++;
			    					if(seq.totalUd == 0) {
			    						CRISP++;
			    					} else {
			    						GQ++;
					    				NCBgo += seq.totalUd;
			    					}
			    					sprb = seq.getP_MPS(histo, histoW, cnt, 0);
			    					mpsHis[1][(int) (100 * sprb)]++;
		    					} else {
			    					sprb = seq.getP_MPS(histo, histoW, cnt, 1);
		    						if(sprb > Parameters.minWeight) {
		    							histogram[1][seq.totalUd]++;
			    						LQ++;
			    						OUT++;
			    						NCBlq += seq.totalUd;
				    					mpsHis[2][(int) (100 * sprb)]++;		    							
		    						}
		    					}
		    					mpsHis[0][(int) (100 * sprb)]++;
		    				}
		    				rid++;
		    			}
		    			reader.close();
		    			sumUDin  = 0;
		    			for(int j = 0; j < histogram[0].length; j++) {
		    				sumUDin += j * histogram[0][j];
		    			}
	    				System.err.print(length + "\t" + Parameters.skipCycles + "\t" + IN);
						System.err.print("\t" + Parameters.trim(((float) 100 * NCBin)  / (IN * Parameters.seedSize)) + "%");
						System.err.print("\t" + (IN - (LQ + GQ + CRISP)));
						System.err.print("\t" + (LQ + GQ + CRISP));
						System.err.print("\t" + CRISP);
						System.err.print("\t" + GQ);
						System.err.print("\t" + LQ);
						System.err.print("\t" + Parameters.trim(((float) 100 * NCBgo) / (GQ * Parameters.seedSize)) + "%");
						System.err.println("\t" + fid + ":" + files[i]);
						
						writerLst.write(length + "\t" + Parameters.skipCycles + "\t" + IN);
						writerLst.write("\t" + Parameters.trim(((float) 100 * NCBin)  / (IN * Parameters.seedSize)) + "%");
						writerLst.write("\t" + (IN - (LQ + GQ + CRISP)));
						writerLst.write("\t" + (LQ + GQ + CRISP));
						writerLst.write("\t" + CRISP);
						writerLst.write("\t" + GQ);
						writerLst.write("\t" + LQ);
						writerLst.write("\t" + Parameters.trim(((float) 100 * NCBgo) / (GQ * Parameters.seedSize)) + "%");
						writerLst.write("\t" + fid + ":" + files[i] + "\n");
	    				for(int k = 0; k < Parameters.seedSize; k++) {
		    				for (int s = 0; s < 2; s++) {
		    					writerLst.write(Parameters.trim(infoHist[s][k]/rid) + "\t");
		    				}
		    				writerLst.write("\n");
	    				}
	    				writerLst.flush();
			        }
		            LQTotal    += LQ;
		            GQTotal    += GQ;
		            CRISPTotal += CRISP;
		            INTotal    += IN;
		            OUTTotal   += OUT;
		            NCBinTotal += NCBin;
		            NCBgoTotal += NCBgo;
		            NCBlqTotal += NCBlq;
		        } // i files
			} // if dd 
		} // dd dirs
		writerLst.close();
		writerCluster.close();
		BufferedWriter writerInfo = new BufferedWriter(new FileWriter(Parameters.pathData + "filesInfo.txt"));
		writerInfo.write("number of PRB files\t" + fid + "\n");
		writerInfo.close();
		BufferedWriter writer = new BufferedWriter(new FileWriter(Parameters.pathData + "stat." + ((int) (100 * Parameters.seqSegNucP)) + "." + Parameters.maxUDN));
		for (int i = 0; i < histo[0].length; i++) {
			sumLQBases[0] += histo[0][i];
			sumLQBases[1] += histo[1][i];
		}

		writer.write("Dataset:\t" + Parameters.DATASET + "\n");
		writer.write("total Sequences:\t" + INTotal + "\n");
		writer.write("Histogram number of UDP%:\n");
		for(int i = 0; i < histogram[0].length; i++) {
			writer.write(i + "\t" + ((float) histogram[0][i]) / INTotal + "\t" + histogram[0][i] + 
					"\t" + ((float) histogram[1][i]) / LQ + "\t" + histogram[1][i] + "\n");
		}
		
		writer.write("\nHistogram P_mps:\n");
		for(int i = 0; i < mpsHis[0].length; i++) {
			writer.write(i + "\t" + ((float) mpsHis[0][i]) / INTotal + "\t" + ((float) mpsHis[1][i]) / GQTotal + 
						"\t" + ((float) mpsHis[2][i]) / LQ + "\t" + mpsHis[0][i] + "\t" + mpsHis[1][i] + "\t" + mpsHis[2][i] + "\n");
		}
		writer.write("\n\tUD%\t\t\tWeighMPBt\t\tnumBasesPerUD"+ "\n");
		for (int i = 0; i < histo[0].length; i++) {
			writer.write(i + 
					"\t" + Parameters.trim(100 * ((float) (histo[0][i] + histo[1][i])/ (float) INTotal)) + 
					"\t" + Parameters.trim(100 * ((float) histo[0][i]/ (float) GQTotal)) + 
					"\t" + Parameters.trim(100 * ((float) histo[1][i]/ (float) (INTotal - GQTotal))) + 
					
					"\t" + Parameters.trim( 100 * (histoW[0][i] + histoW[1][i]) / (histo[0][i] + histo[1][i])) + 
					"\t" + Parameters.trim( 100 * histoW[0][i] / histo[0][i]) + 
					"\t" + Parameters.trim( 100 * histoW[1][i] / histo[1][i]) + 
					
					"\t" + Parameters.trim((cnt[0][i] + cnt[1][i]) / (histo[0][i] + histo[1][i])) + 
					"\t" + Parameters.trim(cnt[0][i] / histo[0][i]) + 
					"\t" + Parameters.trim(cnt[1][i] / histo[1][i]) + "\n");
		}
		writer.write("\n");
		writer.write("Input Data Set:\t" + Parameters.DATASET + "\n");
		writer.write("Lines In .....:\t" + INTotal + "\n");
		writer.write("%NCBin .......:\t" + Parameters.trim(((float) 100 * NCBinTotal)  / (INTotal * Parameters.seedSize)) + "%" + "\n");
		writer.write("Lines Removed :\t" + (INTotal - (LQTotal + GQTotal + CRISPTotal)) + "\n");
		writer.write("Lines PassQC .:\t" + (LQTotal + GQTotal + CRISPTotal) + "\n");
		writer.write("   Crisp .....:\t" + CRISPTotal + "\n");
		writer.write("   GQ ........:\t" + GQTotal + "\n");
		writer.write("   LQ ........:\t" + LQTotal + "\n");
		writer.write("%NCBgq .......:\t" + Parameters.trim(((float) 100 * NCBgoTotal) / (GQTotal * Parameters.seedSize)) + "%\n");
		writer.close();
		System.err.println();
		System.err.println("Input Data Set:\t" + Parameters.DATASET);
		System.err.println("Lines In .....:\t" + INTotal);
		System.err.println("%NCBin .......:\t" + Parameters.trim(((float) 100 * NCBinTotal)  / (INTotal * Parameters.seedSize)) + "%");
		System.err.println("Lines Removed :\t" + (INTotal - (LQTotal + GQTotal + CRISPTotal)));
		System.err.println("Lines PassQC .:\t" + (LQTotal + GQTotal + CRISPTotal));
		System.err.println("   Crisp .....:\t" + CRISPTotal);
		System.err.println("   GQ ........:\t" + GQTotal);
		System.err.println("   LQ ........:\t" + LQTotal);
		System.err.println("%NCBgq .......:\t" + Parameters.trim(((float) 100 * NCBgoTotal) / (GQTotal * Parameters.seedSize)) + "%");
		return INTotal;
	}
	private  void  getMPB(int dr, int i) {
		if (map.Pb[dr][0][i] < map.Pb[dr][1][i]) {
			mpbp = 1;
			npbp = 0;
		} else {
			mpbp = 0;
			npbp = 1;			
		}
		for (int j = 2; j < 4; j++) {
			if (map.Pb[dr][mpbp][i] < map.Pb[dr][j][i]) {
				npbp = mpbp;
				mpbp = j;
			} else {
				if (map.Pb[dr][npbp][i] < map.Pb[dr][j][i])
					npbp = j;
			}
		}
	}
	
	private  void  getMPB(int i) {
		if ((map.Pb[0][0][i] + map.Pb[1][0][i]) < (map.Pb[0][1][i] + map.Pb[1][1][i])) {
			mpbp = 1;
			npbp = 0;
		} else {
			mpbp = 0;
			npbp = 1;			
		}
		for (int j = 2; j < 4; j++) {
			if ((map.Pb[0][mpbp][i] + map.Pb[1][mpbp][i]) < (map.Pb[0][j][i] + map.Pb[1][j][i])) {
				npbp = mpbp;
				mpbp = j;
			} else {
				if ((map.Pb[0][npbp][i] + map.Pb[1][npbp][i]) < (map.Pb[0][j][i] + map.Pb[1][j][i]))
					npbp = j;
			}
		}
	}
	/*
	public  float getBPP(int i, int bp, int inc) {
		if (bp < 0 || bp > 3) {
			System.err.println("bad bp:\t" + bp + "\ti:\t" + i);
			return 0;
		}
		if (i < 0) {
			System.err.println("bad i:\t" + i);
			return 0;
		}
		return map.Pb[inc][bp][i];
	}
	//*/
	private void openPrbOut(String fl) throws Exception {
		fwPrb     = new FileOutputStream(fl + ".prb");
		bwPrb     = new BufferedOutputStream(fwPrb, 256000);
		filePrb   = new DataOutputStream(bwPrb);
	}
	private void closePrbOut() throws Exception {
		filePrb.close();
		bwPrb.close();
		fwPrb.close();
	}
	private void writePrb(byte l) throws Exception {
		// short prbSize = (short) seq.used.length;
    	filePrb.writeInt(seq.id);
    	filePrb.writeByte(l);
    	filePrb.writeShort((short) seq.prbSize);
    	for (int ii = 0; ii < seq.prbSize; ii++) {
    		for (int jj = 0; jj < 4; jj++) {
    			filePrb.writeByte(seq._q[l][ii][jj]);
    		}
    	}
	}
	private void openPrbIn(String f) throws Exception {
		fiPrb    = new FileInputStream(f); 
		biPrb    = new BufferedInputStream(fiPrb, 256000);
		strInPrb = new DataInputStream(biPrb) ;
    }
    private void closePrbIn() throws Exception {
    	strInPrb.close();
    	fiPrb.close();
    }
    
	private boolean read_Q_Comp(int id, int sid, boolean cmp) throws Exception {
		byte l = -1;
		seq.size = Parameters.seedSize;
		seq.id  = strInPrb.readInt();
		// System.err.println(seq.id + "\t" + sid);
		l = strInPrb.readByte();
		seq.prbSize = strInPrb.readShort();
		if (id < 0 ||(id == seq.id && sid == l)) { //
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
	    		for (int jj = 0; jj < 4; jj++) {
	    			if (cmp) {
	    				seq._q[0][ii][comp[jj]] = strInPrb.readByte();
	    			} else {
	    				seq._q[0][ii][jj] = strInPrb.readByte();
	    			}
	    		}
	    	}
	    	return true;
		} else {
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
    			strInPrb.readInt();
	    	}	
		}
		if (id > seq.id) {
			System.err.println("\t<prb:\t" + seq.id + ", " + l + ">\t,Target:\t" + id + "\t" + oligo[DATA].getSide());
			Thread.sleep(2000);
		}
    	return false;
	}

	private boolean readPrb2dCallComp(int id, int sid) throws Exception {
		byte l = -1;
		byte q;
		seq.size = Parameters.seedSize;
		seq.id  = strInPrb.readInt();
		seq.side = l = strInPrb.readByte();
		seq.prbSize = strInPrb.readShort();
		if (id < 0 ||(id == seq.id && sid == l)) { 
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
	    		for (int jj = 0; jj < 4; jj++) {
	    			q = strInPrb.readByte();
	    			seq.p[ii][comp[jj]] = q2prb[ii][q+40];
	    		}
	    	}
	    	return true;
		} else {
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
    			strInPrb.readInt();
	    	}	
		}
		if (id > seq.id) {
			System.err.println("\t<prb:\t" + seq.id + ", " + l + ">\t,Target:\t" + id + "\t" + oligo[DATA].getSide());
			Thread.sleep(2000);
		}
    	return false;
	}
	
	private boolean readPrb2dCall(int id, int sid) throws Exception {
		byte l = -1;
		int q;
		seq.size = Parameters.seedSize;
		seq.id  = strInPrb.readInt();
		seq.side = l = strInPrb.readByte();
		seq.prbSize = strInPrb.readShort();
		if (id < 0 ||(id == seq.id && sid == l)) { 
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
	    		for (int jj = 0; jj < 4; jj++) {
	    			q = strInPrb.readByte();
	    			seq.p[ii][jj] = q2prb[ii][q+40]; 
	    		}
	    	}
	    	return true;
		} else {
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
    			strInPrb.readInt();
	    	}	
		}
		if (id > seq.id) {
			System.err.println("\t<prb:\t" + seq.id + ", " + l + ">\t,Target:\t" + id + "\t" + oligo[DATA].getSide());
			Thread.sleep(2000);
		}
    	return false;
	}

	private boolean read_PrbExtend(int id, int sid) throws Exception {
		byte l = -1;
		int q;
		seq.size = Parameters.seedSize;
		seq.id  = strInPrb.readInt();
		seq.side = l = strInPrb.readByte();
		seq.prbSize = strInPrb.readShort();
		if (id < 0 ||(id == seq.id && sid == l)) { //
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
	    		for (int jj = 0; jj < 4; jj++) {
	    			q = strInPrb.readByte();
	    			seq.p[ii][jj] = seq.qCal[q+40]; // (float) (1/(1+Math.pow(10,(((float)-q)/10)))); // 
	    			// System.err.println("Q:\t" + q + "\tp:\t" + q2prb[ii][q+40]);
	    		}
	    	}
	    	return true;
		} else {
	    	for (int ii = 0; ii < seq.prbSize; ii++) {
    			strInPrb.readInt();
	    	}	
		}
		if (id > seq.id) {
			System.err.println("\t<prb:\t" + seq.id + ", " + l + ">\t,Target:\t" + id + "\t" + oligo[DATA].getSide());
			Thread.sleep(2000);
		}
    	return false;
	}

	public  void processReference2(String[] files, int[] refSeqSize, int wSize) throws Exception {
		//*
		ofile = new OligoFile(Parameters.pathDB + Parameters.DATABASE, "", OligoFile.W);
		for (chr = 0; chr < Parameters.SEQ_VAR.length; chr++) {//HUMAN_CHR.length
			processFile2(files[chr], refSeqSize, chr, wSize);
			System.err.println(files[chr] + ":\t" + refSeqSize[chr]);
		}
		ofile.saveSBunk(Oligo.SEQ);
		ofile.dfn = ofile.dfns;
		ofile.close();
		//*/
		ofile = new OligoFile(Parameters.pathDB + Parameters.DATABASE, "", OligoFile.R);
		ofile.joinSortU2(); 
		ofile.close();
	}

	public  void processReference(String[] files, int[] refSeqSize) throws Exception {
		//*
		ofile = new OligoFile(Parameters.pathDB + Parameters.DATABASE, "", OligoFile.W);
		for (chr = 0; chr < Parameters.SEQ_VAR.length; chr++) {//HUMAN_CHR.length
			processFile(files[chr], refSeqSize, chr);
			System.err.println(files[chr] + ":\t" + refSeqSize[chr]);
		}
		ofile.saveSBunk(Oligo.SEQ);
		ofile.dfn = ofile.dfns;
		ofile.close();
		//*/
		ofile = new OligoFile(Parameters.pathDB + Parameters.DATABASE, "", OligoFile.R);
		ofile.joinSortU2(); 
		ofile.close();
	}

	public  void processContamination(String pathConSeq, String ConDB) throws Exception {
		int oldDBSize = Parameters.dbOligoSize;
		/*
		if(Parameters.seedSize <= Oligo.wordSize) {
			Parameters.dbOligoSize = Oligo.wordSize;
		}
		//*/
		Parameters.dbOligoSize = Parameters.seedSize;
		int[] tmp = new int[1];
		try {
			File fl = new File(pathConSeq);
			System.err.println("ConDB:\t" + ConDB);
			ofile = new OligoFile(ConDB, "", OligoFile.W);
			System.err.println("pathConSeq:\t" + pathConSeq);
			File[] dirs = fl.listFiles();
			for (int dd = 0; dd < dirs.length; dd++) {
				if(dirs[dd].isDirectory()) {
			        File[] files = dirs[dd].listFiles();
			        for(int i=0; i<files.length; i++) {
						System.gc();
			            if(!files[i].isDirectory()) { // && (files[i].toString()).indexOf("prb") >= 0) {
			    			System.err.println("File:\t" + files[i].toString());
			    			processFile(files[i].toString(), tmp, 0);
			            }
			        }
				}
			}
			ofile.saveSBunk(Oligo.SEQ);
			ofile.dfn = ofile.dfns;
			ofile.close();
			Parameters.dbOligoSize = oldDBSize;
			ofile = new OligoFile(ConDB, "", OligoFile.R);
			ofile.joinSortU2(); 
			ofile.close();
			System.err.println("conamination sequences Database created");
		} catch(Exception e) {
			System.err.println("No conamination sequences");
			Parameters.contDB = false;
		}
	}
	
	private void processFile(String fn, int[] sz, int chr) throws Exception {
		char[] olc = new char[Parameters.dbOligoSize];
		String line;
		oId = 0;
		int c = 0;
		System.err.println(fn + "\n");
		BufferedReader reader = new BufferedReader(new FileReader(Parameters.pathRefSeq + fn));
		line = reader.readLine();
		System.err.println(fn + "\n" + line + "\n");
		clearMeta();

		for (int i = 0; i < Parameters.dbOligoSize; i++) {
			c = reader.read();
			while(c == 10 || c == 13) {
				c = reader.read();
			}
			
			olc[i] = Character.toUpperCase((char) c);
		}
		idx = 0;
		while (c > 0) {
			c_oligo = new String(olc);
			process(Parameters.dbOligoSize); // will create new DB line
			c = reader.read();
			idx++;
			while(c == 10 || c == 13) {
				c = reader.read();
			}
			for (int i = 1; i < Parameters.dbOligoSize; i++) {
				olc[i-1] = olc[i];
			}
			olc[Parameters.dbOligoSize-1] = Character.toUpperCase((char) c);
			if (idx%1000000 == 0)
				System.err.println("\tidx\t" + idx/1000000 + "M");
		}
		reader.close();
		c_oligo = new String(olc);
		sz[chr] = idx + Parameters.dbOligoSize;
	}

	private void processFile2(String fn, int[] sz, int chr, int wSize) throws Exception {
		char[] olc = new char[wSize];
		int olSize = 1;
		// String line;
		oId = 0;
		int c = 1;
		BufferedReader reader = new BufferedReader(new FileReader(Parameters.pathRefSeq + fn));
		BufferedWriter writer = new BufferedWriter(new FileWriter(Parameters.pathDB + Parameters.DATABASE + fn + ".contigs"));
		System.err.println(Parameters.pathDB + Parameters.DATABASE + fn + ".contigs");
		clearMeta();
		idx = 0;
		int counter = 0;
		while (c > 0) {
			if(olSize == wSize){
				c_oligo = new String(olc);
				process(wSize); // will create new DB line
			}
			c = getCharFromSeq(reader, writer, wSize);
			counter++;
			if (counter >= wSize)
				idx++;
			if(cutString) {
				olSize = 0;
				cutString = false;
			} 
			if(olSize == wSize){
				for (int i = 1; i < wSize; i++) {
					olc[i-1] = olc[i];
				}
				olc[wSize-1] = Character.toUpperCase((char) c);
			} else {
				olc[olSize] = Character.toUpperCase((char) c);
				olSize++;
			}
			if (counter%1000000 == 0)
				System.err.println("\tidx\t" + idx/1000000 + "M");
		}
		reader.close();
		writer.close();
		c_oligo = new String(olc);
		sz[chr] = idx + wSize;
	}
	private int getCharFromSeq(BufferedReader reader, BufferedWriter writer, int wSize) throws Exception {
		int c;
		int ofset = 0;
		if (idx > wSize)
			ofset = idx + wSize;
		c = reader.read();
		while(c == 10 || c == 13 || c == '>') {
			if (c == '>') {
				String line = reader.readLine();
				// System.err.println(">" + line);
				writer.write(ofset + "\t>" + line + "\n");
				cutString = true;
			} else {
				cutString = false;				
			}
			c = reader.read();
		}
		return c;
	}
    private void process(int wSize) throws Exception { 
    	char [] olc = c_oligo.toCharArray();
    	nCounter = 0;
    	for (int i = 0; i < wSize; i++) {
    	    bak_c_oligo[i] = c_oligo.charAt(i);
    	    nList[i] = 0;
    	}

    	for (int i = 0; i < wSize; i++) {
    	    if (olc[i] == 'N' || olc[i] == 'M' || olc[i] == 'R' || olc[i] == 'x') {
    		    nList[nCounter] = i;
    		    nCounter++;
    		    olc[i] = 'T';
    	    } 
    	}

    	if (nCounter == 0) {
    	    c_oligo = new String(olc);
    	    codeIt();
    	} else if (nCounter > non) {
    	    c_oligo = new String(bak_c_oligo);
    	    return;
    	} else {
    	    c_oligo = new String(olc);
		    codeIt();
    	    while (flipOligo() == 1) {
    		    codeIt();
    	    }
    	    oId++;
    	    for (int i = 0; i < wSize; i++) {
    		    olc[i] = bak_c_oligo[i];
    	    }
    	}
	    c_oligo = new String(olc);
		// System.err.println("C_OLIGO SIZE <process>: " + c_oligo.length());
    }
    private void codeIt() throws Exception { // to fill an oligo object
    	boolean uq;
    	Oligo ol = new Oligo(new String(c_oligo));
    	ol.start = idx; // 
    	ol.ch    = chr;
    	ol.size = c_oligo.length();
    	ol.weight = (float) 1.0/ (float) Math.pow(4.0,(double) nCounter);
    	ol.id1 = oId;
    	if (ol.weight < 0.97)
    		uq = false;
    	else
    		uq = true;
    	ol.setUQ(uq);
    	ol.id1 = 0; //olid;
		// System.err.println(idx + ">\t" + ol);
    	ofile.sort_WriteOligo(new Oligo(ol), Oligo.SEQ);
    	ol.comp();
    	ol.revers();
		// System.err.println(idx + ">\t" + ol);
    	ofile.sort_WriteOligo(ol, Oligo.SEQ);
    }

    private void addOligo(int fd) {
    	bunk[fd].add(new Oligo(oligo[fd]));
    }
   
    private int  flipOligo() { // to give different posible sequences of a seq., by changing Ns
    	char[] olc = c_oligo.toCharArray();
    	for (int i = nCounter - 1; i >= 0; i--) {
    	    if (bp2Int(olc[nList[i]]) == A) {
    	    	olc[nList[i]] = 'T';
    	    } else {
    	    	olc[nList[i]] = getBpChar(bp2Int(olc[nList[i]]) - 1);
    	    	c_oligo = new String(olc);
    		    // System.err.println("flipOligo:\t" + c_oligo + "\t" + nList[nCounter - 1]);
    		    return 1;
    	    }
    	}
    	c_oligo = new String(olc);
    	return -1;
    }
    //*
    private void deCodeIt(int od) { // to decode an oligo into C_oligo string
    	long a;
    	int sz = Parameters.now * wordSize + 1;
    	char[] olc = new char[sz + 1]; 
    	for (int w = 0; w < 2; w++) {
    		for (int i = 0; i < wordSize; i++) {
    			a = oligo[od].d[w];
    			a = a << (2 + 2 * i);
    			a = a >>> (62);
    			olc[w * wordSize + i] = getBpChar((int) a);
    		}
    	}
    	c_oligo = (new String(olc)).substring(0, oligo[DATA].size);
    }
    //*/
    public  void align(String fDatabase, String pData, String pOut, String fData, int uCompare) throws Exception {
    	//             String db, String pd, String po, String dt, int osz, int ux
		seq = new Seq();
    	String fOne   = pData + fData + ".ol" + ux;
    	String fTwo   = fDatabase;
    	String mMatch = pOut + fData + ".m"  + ux;
    	String nMatch = pOut + fData + ".nm" + ux;
    	subResult = UM;
		//System.gc();
    	subtractSQ2f1(fOne, fTwo, nMatch, mMatch, uCompare);
    }
    
    public  void subtractSQ2f1(String fOne, String fTwo, String nMatch, String mMatch, int uCompare) throws Exception {
    	Oligo  bakO;
    	int mmIndex;
    	boolean u0bool;
    	int idxDbDat;
    	int compResult;
    	int dbsz;
    	int sz = 0;
    	int evSize;
    	int readCount = 0;
    	int S0Count   = 0;
    	int S1Count   = 0;
    	int petCount  = 0;
    	BufferedWriter multiMatchHist = null;
    	
    	mmHist   = new long[Parameters.mmHistSize];
    	ofOne    = new OligoFile(fOne, "", OligoFile.R);
    	evSize   = (int) (Parameters.maxId - ofOne.minID + 1); // + ofOne.weightOfCG);
    	ofTwo    = new OligoFile(fTwo, "", OligoFile.R);
    	if (subResult == NM) {
    		ofNMatch = new OligoFile(nMatch, "", OligoFile.W);
    	} else {
    		ofMatch  = new OligoFile(mMatch + ".s.id", "", OligoFile.W);
    	}
    	
    	for (int i = 0; i < mmHist.length; i++) {
    		mmHist[i] = 0;
    	}
    	
    	if (evSize > 0) {
        	idArray   = new byte[4][evSize];
    		for(int i = 0; i < evSize; i++) {
    			idArray[0][i] = idArray[1][i] = idArray[2][i] = idArray[3][i] = 0;
    		}
    	}
		idxDbDat = 0;
		idx      = 0;
		clearMeta();

		ofOne.readOligo(oligo[DATA]);
		sz = oligo[DATA].size;
		readBunk(DATA, sz); 
		ofTwo.readOligo(oligo[DATB]);
		readBunk(DATB, sz); 
		
		if (uCompare < 0) {
			uCompare = 0;
		}
    	if (uCompare == 0) {
    		u0bool = true;
    		multiMatchHist = new BufferedWriter(new FileWriter(Parameters.pathOut + "multiMatch.hist"));
    	} else {
    		u0bool = false;
    	}
		System.err.print("\nSearching:\t");
		// int sIdx;
		while(ofOne.available() > 0 && ofTwo.available() > 0) {
	    	bakO = new Oligo(oligo[DATA]); 
			compResult = (bunk[DATA].get(bunk[DATA].size()-1)).equal(bunk[DATB].get(bunk[DATB].size()-1), sz);
			if (compResult == 0) { // 0 is equal
				dbsz = bunk[DATB].size();
				if (dbsz == 1) {
					if (subResult == UM) {
						Iterator<Oligo> it = bunk[DATA].iterator();
						Oligo oligoDB = new Oligo(bunk[DATB].get(0)); // ------------------------------------------ delete new?
						while (it.hasNext()) {
							Oligo o      = (new Oligo(oligoDB));
							oligo[DATA]  = (Oligo) it.next();
				    		o.id1    = oligo[DATA].id1;
				    		o.setSide(oligo[DATA].getSide());
							o.size   = oligo[DATA].size;
							if (Parameters.seedSize > (o.u0 + 2) && o.weight > 0.9)
								o.setU0(u0bool);
							else 
								o.setU0(false);
							o.weight = oligo[DATA].weight * o.weight;
							o.setMP(oligo[DATA].isMP());
							o.setUQ(oligo[DATA].isUQ());
							o.setLQ(oligo[DATA].isLQ());
							/*
							if(o.isRC()) {
								o.start = o.start + Parameters.dbOligoSize;
							} //*/
							idx++;
							o.setUM(true);
							if (evSize > 0) {
								idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] = (byte) (idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] | (byte) 1);
								idArray[oligo[DATA].getSide()+2][Parameters.maxId - oligo[DATA].id1] += 1;
							}
							ofMatch.sort_WriteOligo(new Oligo(o), Oligo.ID);
						}
					}
				} else {
					if (subResult == UM) {
						if (u0bool) {
							if (dbsz < mmHist.length) {
								mmIndex = dbsz;								
							} else {
								mmIndex = mmHist.length - 1;
							}
						    if (u0bool) {
						    	mmHist[mmIndex] += bunk[DATA].size();
						    }
						} 
						if (dbsz <= Parameters.MMBRK) { 
						    Iterator<Oligo> idb = bunk[DATB].iterator();
					        for (; idb.hasNext(); ) {
					        	Oligo oligoDB = new Oligo((Oligo) idb.next()); // ----------------------------- delete new? 2027CAAXX
							    Iterator<Oligo> it = bunk[DATA].iterator();
						        for (; it.hasNext(); ) {
						    		oligo[DATA] = (Oligo) it.next();
						    		Oligo o     = (new Oligo(oligoDB));
						    		o.id1       = oligo[DATA].id1;
						    		o.setSide(oligo[DATA].getSide());
									o.size      = oligo[DATA].size;
									o.weight    = oligo[DATA].weight * o.weight;
									o.setMP(oligo[DATA].isMP());
									o.setUQ(oligo[DATA].isUQ());
									o.setLQ(oligo[DATA].isLQ());
									/*
									if (oligoDB.isRC()) {
										o.start += oligoDB.size - oligo[DATA].size;
										o.revers_Comp();
										// o.u0 = (byte) (oligo[DATA].size - o.u0);
									}
									//*/
									/*
									if(o.isRC()) {
										o.start = o.start - Parameters.dbOligoSize;
									} //*/
									idx++;
									o.setUM(false);
									o.setU0(u0bool);
									if (evSize > 0)
										idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] = (byte) (idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1]|1);
									ofMatch.sort_WriteOligo(new Oligo(o), Oligo.ID);
								}
						    }
						} else {
						    Iterator<Oligo> it = bunk[DATA].iterator();
					        while (it.hasNext()) {
					    		oligo[DATA]        = (Oligo) it.next();
								if (evSize > 0)
									idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] = 
										(byte) (idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1]|1);
					        }
						}
			        }
				}
				oligo[DATA] = bakO; //new Oligo(bakO);
				readBunk(DATA, sz); 
				readBunk(DATB, sz); 
			} else if (compResult < 0) {
			    if (subResult == NM) {
			    	Iterator<Oligo> it = bunk[DATA].iterator();
			    	for (; it.hasNext(); ) {
			    		oligo[DATA] = new Oligo((Oligo) it.next());
				    	ofNMatch.writeOligo(oligo[DATA]);
				    }
		        }
				oligo[DATA] = bakO; //new Oligo(bakO);
				readBunk(DATA, sz); 
			} else {
				readBunk(DATB, sz); 
			}

			idxDbDat++;
			if (idxDbDat % 1000000 == 0) {// 1M 
				System.err.print(oligo[DATA].getBPat(0) + (oligo[DATA].getBPat(1) + " "));
				System.gc();
			}
		}
		System.err.println();
    	bunk[DATA].clear();
    	bunk[DATA].trimToSize();
    	bunk[DATB].clear();
    	bunk[DATB].trimToSize();
		ofOne.close();
		ofTwo.close();
		if (subResult == UM) {
			// ofMatch.close();
			if (ofMatch.bunk.size() > 0)
				ofMatch.saveSBunk(Oligo.ID);
			ofMatch.dfn = ofMatch.dfns;
			ofMatch.close();
			ofMatch = new OligoFile(mMatch + ".s.id", "", OligoFile.R);
			Oligo.sortIndex = Oligo.ID;
			ofMatch.joinSort();
			ofMatch.close();
		}
	    if (subResult == NM) {
	    	ofNMatch.close();
	    }
		ofNMatch = null;
		ofMatch  = null;
		ofOne    = null;
		ofTwo    = null;
    	System.gc();
    	int rdCount = 0;
    	int rdVal   = 0;
		if (evSize > 0) {
			IdCount = 0;
			for(int i = 0; i < evSize; i++) {
				if (idArray[2][i] > 1){
					rdCount++;
					rdVal += idArray[2][i] - 1;
				} 
				if (idArray[3][i] > 1){
					rdCount++;
					rdVal += idArray[3][i] - 1;
				} 
				if ((idArray[0][i]%2) == 1 || (idArray[1][i]%2) == 1){
					IdCount++;
				} 
				if ((idArray[0][i]%2) == 1){
					readCount++;
					S0Count++;
					if ((idArray[1][i]%2) == 1) {
						petCount++;
					}
				}
				if ((idArray[1][i]%2) == 1) {
					readCount++;
					S1Count++;
				}
			}
			if (u0bool) {
				multiMatchHist.write("i\tHist\n");
				for(int i = 0; i < mmHist.length; i++) {
					multiMatchHist.write(i + "\t" + mmHist[i] + "\n");
				}
			}
			if (subResult == UM) {
				System.err.println();
				System.err.println("S0 Count ..... :\t" + S0Count);
				System.err.println("S1 Count ..... :\t" + S1Count);
				System.err.println("S0 +  S1 Count :\t" + readCount);
				System.err.println("S0 || S1 Count :\t" + IdCount);
				System.err.println("S0 && S1 Count :\t" + petCount);
				System.err.println("redundent Count:\t" + rdCount);
				System.err.println("redundent Val  :\t" + rdVal);
			}
		}
		if (u0bool) {
			multiMatchHist.close();
		}
    }
    
    public  void subtractID(String inf, String outf) throws Exception {
    	Oligo ol;
    	
		Oligo.sortIndex = Oligo.SEQ;
    	ofOne    = new OligoFile(inf, "", OligoFile.R);
    	ofile    = new OligoFile(outf , "", OligoFile.W);
		ofOne.readOligo(oligo[DATB]);
		while(ofOne.available() > 0) {
			if ((oligo[DATB].weight > Parameters.cutWeight && !oligo[DATB].isLQ()) || (oligo[DATB].weight > 2 * Parameters.cutWeight)) {
				if (idArray[oligo[DATB].getSide()][Parameters.maxId - oligo[DATB].id1] == 0) {
					deCodeIt(DATB);
					for (byte i = 0; i < oligo[DATB].size; i++) {
						for (int j = 0; j < 4; j++) {
							if(nucl[j] != c_oligo.charAt(i)) {
								ol = new Oligo(c_oligo.substring(0, i) + nucl[j] + c_oligo.substring(i+1, oligo[DATB].size));
								ol.id1    = oligo[DATB].id1;
								ol.flg    = oligo[DATB].flg;
								ol.weight = oligo[DATB].weight;
								ol.u0     = i;
								ofile.sort_WriteOligo(ol, Oligo.SEQ); 
								idx++;
							}
						}
					}
				}
			} 
			ofOne.readOligo(oligo[DATB]);
		}
		idArray = null;
		ofOne.close();
		ofOne = null;
		ofile.saveSBunk(Oligo.SEQ);
		ofile.dfn = ofile.dfns;
		ofile.close();
		ofile = null;
		System.gc();
		ofile = new OligoFile(outf, "", OligoFile.R);
		ofile.joinSort(); 
		ofile.close();
		ofile = null;
		System.gc();
    }
    
    private void readBunk(int fd, int sz) throws Exception {
    	byte two = 2;
    	OligoFile of;
    	if (fd == DATA)
    		of = ofOne;
    	else 
    		of = ofTwo;
    	
    	Oligo bak = new Oligo(oligo[fd]);
    	bunk[fd].clear();
    	if (fd == DATA && oligo[fd].isLQ()) {
    		idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] = (byte) (idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] | two);
    	}
    	addOligo(fd);
		while(of.readOligo(oligo[fd])) {
			if (bak.equal(oligo[fd], sz) != 0) {
				break;
			}
	    	if (fd == DATA && oligo[fd].isLQ()) {
	    		idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] = (byte) (idArray[oligo[DATA].getSide()][Parameters.maxId - oligo[DATA].id1] | two);
	    	}
	    	addOligo(fd);
		}   
    }
   
    private void readIdBunk(int bk) throws Exception {
    	Oligo bak = new Oligo(oligo[bk]);
    	bunk[bk].clear();

    	addOligo(bk);
		while(ofOne.readOligo(oligo[bk])) {
			if (bak.id1 == oligo[bk].id1 && bak.getSide() == oligo[bk].getSide()) {
		    	addOligo(bk);
			} else {
				break;
			}
		}   
    }

    public void extend(int u) throws Exception {
    	int i;
    	String ef = Parameters.pathOut + "s.m"   + u + ".s.id";
    	System.err.println("Extend:\t" + ef);
    	int umCount = 0;
    	int mmCount = 0;
    	int mmumCount = 0;
    	int mmmmCount = 0;
    	float sumWeights;
    	int oldSortIdx = Oligo.sortIndex;
    	Oligo.sortIndex = Oligo.WEIGHT;
		ofOne  = new OligoFile(ef, "", OligoFile.R);
		ofTwo  = new OligoFile(Parameters.pathOut + "s.e.m" + u + ".s.id", "", OligoFile.W);
		openPrbIn(Parameters.pathData + Parameters.outFile + ".prb");
		bunk[DATB].clear();
		ofOne.readOligo(oligo[DATB]);
		// addOligo(DATA);
		int ums;
		int countAll = 0;
		seq.id = Parameters.maxId + 1;
    	for (int ii = 0; ii < 93; ii++) {
    		for (int q = -40; q <= 40; q++) {
    			q2prb[ii][q+40] = (float) (1/(1+Math.pow(10,(((float)-q)/10))));
    		}
    	}
		while(ofOne.available() > 0) {
			countAll++;
			readIdBunk(DATB);
		    Oligo r = new Oligo(bunk[DATB].get(0));
		    if (seq.id > r.id1 || (seq.id == r.id1 && seq.side <= r.getSide())) {
			    while(!read_PrbExtend(r.id1, r.getSide()));
			    if (countAll%1000000 == 0) {
			    	System.err.println(countAll/1000000 + "M:\tSeq_id: " + seq.id + ", " + seq.side + ",\tR_id: " + r.id1 + ", " + r.getSide());
			    }
			    i = 0;
			    ums = 0;
		        while (i < bunk[DATB].size()) {
		    		oligo[DATA] = bunk[DATB].get(i);
		    		if (oligo[DATA].extend(seq) > 1) { // <<<------------------------------------------- extend oligo
		    			ums++;
		    			oligo[DATA].setUM(true);
		    		} else {
		    			oligo[DATA].setUM(false);
		    		}
		    		// Adjust weight
		    		if(oligo[DATA].isRC()) {
		    			oligo[DATA].start += Parameters.dbOligoSize - oligo[DATA].size;
		    			oligo[DATA].revers();
			    		updateOligoWeightEx();
		    			oligo[DATA].comp();
			    		i++;
		    		} else {
		    			i++;
		    			updateOligoWeightEx();
		    		}
		        }
		        if (bunk[DATB].size() == 1) {
	        		(bunk[DATB].get(0)).setUM(true);
	        		ofTwo.writeOligo(bunk[DATB].get(0));
	        		umCount++;
		        } else if (bunk[DATB].size() > 1 ) { // ===============================================
		        	if (ums == 1) {
		        		i = 0;
		        		while (i < bunk[DATB].size()) {
		        			if (bunk[DATB].get(i).isUM()) {
		        				ofTwo.writeOligo(bunk[DATB].get(i));
		        				break;
		        			}
		        			i++;
		        		}
		        	} else if (ums > 1) {
		        		sumWeights = 0;
		        		i = 0;
		        		while (i < bunk[DATB].size()) {
		        			if (!bunk[DATB].get(i).isUM()) {
		        				bunk[DATB].remove(i);
		        			} else {
		        				sumWeights += bunk[DATB].get(i).weight;
		        				i++;
		        			}
		        		}
		        		i = 0;
		        		if (sumWeights > 1) {
			        		while (i < bunk[DATB].size()) {
		        				bunk[DATB].get(i).weight = bunk[DATB].get(i).weight / sumWeights;
			        			i++;
			        		}
		        		}
		        		Collections.sort(bunk[DATB]);
		        		if ((bunk[DATB].get(0).weight / bunk[DATB].get(1).weight) >= 2 && u == 0) {
		        			ofTwo.writeOligo(bunk[DATB].get(0));
		        			mmumCount++;
		        		} else {
			        		i = 0;
			        		mmmmCount++;
			        		while (i < bunk[DATB].size()) {
			        			bunk[DATB].get(i).setUM(false);
			        			ofTwo.writeOligo(bunk[DATB].get(i));
			        			i++;
			        		}
		        		}
		        	} else { // ums == 0
		        		mmCount++;
		        		sumWeights = 0;
		        		i = 0;
		        		while (i < bunk[DATB].size()) {
	        				sumWeights += bunk[DATB].get(i).weight;
	        				i++;
		        		}
		        		if (sumWeights > 1) {
			        		i = 0;
			        		while (i < bunk[DATB].size()) {
		        				bunk[DATB].get(i).weight = bunk[DATB].get(i).weight / sumWeights;
			        			i++;
			        		}
		        		}
		        		i = 0;
		        		while (i < bunk[DATB].size()) {
		        			ofTwo.writeOligo(bunk[DATB].get(i));
		        			i++;
		        		}
			        } 
		        }
			} else {
	        	System.err.println(" Pass <Seq: " + seq.id + ", " + seq.side + ">,\t<R: " + r.id1 + ", " + r.getSide() + ">");
	        	Thread.sleep(500);
	        }
		}
		System.err.println("umCount:\t" + umCount + "\tmmCount:\t" + mmCount + "\tmmumCount:\t" + mmumCount + "\tmmmmCount:\t" + mmmmCount);
		ofTwo.close();
		ofOne.close();
		Oligo.sortIndex = oldSortIdx;
    }
    
    public  void seperate(String[] files) throws Exception { // , int[][] resultCount
    	if (Parameters.ETag == Parameters.SET) {
       		seperate_SET(); // , resultCount
    	} else {
    		System.err.println("Start seperatePET1");
    		seperatePET1(files); // to generate petHist from s.e.m0.id
    		System.err.println("Start seperatePET2");
    		SeparatePET2();
    		System.err.println("Start seperatePET3");
    		SeparatePET3(files);
    	}
    	// processMM()
    }
    
    int[][] sdist = new int[2][10000];
    private void Aligne2(String line, int count, OligoFile ofNx, int chr) throws Exception {
		int  startRefSeg;
		int  endRefSeg;
		int  start2; // start other ol
		int  size2;  // size  other ol
		int  id2;    // Id    other ol
		int  start1; // start this ol
		byte side1;  // Side  this ol
		int  size1;  // size  this ol <end - start>
		byte[] ref  = new byte[100]; // to pass to saveOligoP as parameter
		byte[] mp   = new byte[100]; // to pass to saveOligoP as parameter
		byte[] prbs = new byte[400]; // to pass to saveOligoP as parameter
		int[] maxScore = {0, 0};
		int[] score = new int[11 * Parameters.petSpanMax];
		StringTokenizer st = new StringTokenizer(line);
		size1  = st.countTokens() / 4 - 1;
		start2 = new Integer(st.nextToken()).intValue();
		size2  = new Integer(st.nextToken()).intValue();
		id2    = new Integer(st.nextToken()).intValue();
		side1  = (byte) new Integer(st.nextToken()).intValue();
		for(int i = 0; i < size1; i++) {
			for(int b = 0; b < 4; b++) {
				bfr[0][b][i] = (byte) new Integer(st.nextToken()).intValue();
			}
		}
		endRefSeg = start2 + 2 * Parameters.petSpanMax;
		if (endRefSeg >= (aligne2RefSize - size1))
			endRefSeg = aligne2RefSize - size1;
		startRefSeg = start2 - 2 * Parameters.petSpanMax;
		if (startRefSeg < 0)
			startRefSeg = 0;
		int i0;
		int jr;
		int x;
		for (int i2 = startRefSeg; i2 < endRefSeg; i2++) {
			i0 = i2 - startRefSeg;
			score[2 * i0] = score[2 * i0 + 1] = 0;
			// ------------------------------------------------------------------------------------
			for (int jf = 0; jf < size1; jf++) {
				x = i2+jf;
				if (aligne2Ref[x] < 4 && aligne2Ref[x] >= 0) {
					score[2 * i0] += bfr[0][aligne2Ref[x]][jf];
				} else
					score[2 * i0] += -5;
			}
			score[2 * i0] = score[2 * i0] / size1;
			if (score[2 * i0] > score[maxScore[1]]) {
				if (score[2 * i0] > score[maxScore[0]]) {
					maxScore[1] = maxScore[0];
					maxScore[0] = 2 * i0;
				} else {
					maxScore[1] = 2 * i0;
				}
			}
			// -------------------------------- (byte) Consensus.getQCal(				
			for (int jf = 0; jf < size1; jf++) {
				x = i2+jf;
				jr = size1 - jf - 1;
				if (aligne2Ref[x] < 4 && aligne2Ref[x] >= 0) {
					score[2 * i0 + 1] += bfr[0][iComp[aligne2Ref[x]]][jr];
				} else
					score[2 * i0 + 1] += -5;
			}
			score[2 * i0 + 1] = score[2 * i0 + 1] / size1;
			if (score[2 * i0 + 1] > score[maxScore[1]]) {
				if (score[2 * i0 + 1] > score[maxScore[0]]) {
					maxScore[1] = maxScore[0];
					maxScore[0] = 2 * i0 + 1;
				} else {
					maxScore[1] = 2 * i0 + 1;
				}
			}
		}

		if (score[maxScore[0]] > 2 * score[maxScore[1]]) {
			start1 = startRefSeg + maxScore[0]/2;
			m_atches[score[maxScore[0]] + 40]++;
			if (score[maxScore[0]] > 15) {
				sdist[maxScore[0]%2][maxScore[0]/2]++;
				for (int j = 0; j < size1; j++) {
					ref[j] = aligne2Ref[start1 + j];
					mp[j]  = aligne2Map[start1 + j];
				}
				if (maxScore[0]%2 == 0) {
					for (int j = 0; j < size1; j++) {
						for (int b = 0; b < 4; b++) {
							prbs[j*4 + b] = bfr[0][b][j];
						}
					}
				} else {
					for (int j = 0; j < size1; j++) {
						for (int b = 0; b < 4; b++) {
							prbs[j*4 + comp[b]] = bfr[0][b][size1 - j - 1];
						}
					}					
				}
				ofNx.saveOligoP(id2, chr, start1, maxScore[0]%2, size1, side1, start2, size2, mp, ref, prbs, Oligo.ID);
			}
		}
    }

    int[] m_atches = new int[81];
    private void SeparatePET3(String[] files) throws Exception {
    	//*
    
    	System.err.println("PET3 Part 1, create s.nx.s.id from chr.prbNexts");
    	loadPETSpan();
    	File next;
    	for (int i = 0; i < 1000; i++) {
    		sdist[0][i] = sdist[1][i] = 0;
    	}
    	Consensus.loadQ2P();
    	String line;
    	int count = 0;
    	for (int i = 0; i < 81; i++)
    		m_atches[i] = 0;
		OligoFile ofNx = new OligoFile(Parameters.pathOut + "s.nx.s.id", Parameters.pathRefSeq + "", OligoFile.W);
		Oligo.sortIndex = Oligo.ID;
    	for (int f = 0; f < files.length; f++) {
    		System.err.println(files[f] + "\t" + Parameters.refSeqSize[f]);
    		loadRef(files[f], f);
    		next = new File(Parameters.pathOut + files[f] + ".prbNext");
    		BufferedReader readerNext = new BufferedReader(new FileReader(next));
    		while(readerNext.ready()) {
    			count++;
    			line = readerNext.readLine();
    			Aligne2(line, count, ofNx, f);
    		}
    		readerNext.close();
    		// next.delete();
    	}
		ofNx.saveSBunkP(Oligo.ID);
		ofNx.dfn = ofNx.dfns;
		ofNx.closeP();
		Oligo.sortIndex = Oligo.ID;
		ofNx = new OligoFile(Parameters.pathOut + "s.nx.s.id", Parameters.pathRefSeq + "", OligoFile.R);
		ofNx.joinSortP();
		ofNx.closeP();
		BufferedWriter bw = new BufferedWriter(new FileWriter(Parameters.pathOut + "sdist.txt"));
    	for (int i = 0; i < 1000; i++) {
    		bw.write(sdist[0][i] + "\t" + sdist[1][i] + "\n");
    	}
    	bw.close();
    	//*/
    	// --------------------------------------------------------------------------
    	System.err.println("PET3 Part 2, separate s.um.s.id & s.nx.s.id into chr.um.s.l");
    	// --------------------------------------------------------------------------
    	int nxt;
		OligoFile[] of = new OligoFile[2];
		of[0] = new OligoFile(Parameters.pathOut + "s.um.s.id", "",	OligoFile.R);
		of[1] = new OligoFile(Parameters.pathOut + "s.nx.s.id", "",	OligoFile.R);
		OligoFile[] ofOut = new OligoFile[Parameters.SEQ_VAR.length];
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			String file = Parameters.pathOut + Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2 + ".um.s.l";
			ofOut[chromosome] = new OligoFile(file, Parameters.pathRefSeq + "",	OligoFile.W);
		}

		PriorityQueue<OligoFile.OligoFileTop> pQueue = new PriorityQueue<OligoFile.OligoFileTop>();
		OligoFile.OligoFileTop oft;
		int idLast = -1;
		boolean loaded = false;
		Oligo[] ols = new Oligo[2];
		
		Oligo.sortIndex = Oligo.ID;
		for (int j = 0; j < of.length; j++) {
			oft = of[j].new OligoFileTop();
			Oligo ol = new Oligo();
			of[j].readOligo(ol);
			oft.ol = ol;
			oft.id = j;
			pQueue.add(oft);
		}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			if (oft.id != idLast) {
				if (loaded) {
					ols[oft.id] = new Oligo(oft.ol);				
					ols[0].setPET(true);
					ofOut[ols[1].ch].sort_WriteOligoP(ols[1], Oligo.LOCATION);
					ofOut[ols[0].ch].sort_WriteOligoP(ols[0], Oligo.LOCATION);
					ols[0] = null;
					ols[1] = null;
					idLast = -1;
					loaded = false;
				} else {
					ols[oft.id] = new Oligo(oft.ol);		
					idLast = oft.id;
					loaded = true;
				}
			} else {
				if (loaded) {
					ofOut[ols[idLast].ch].sort_WriteOligoP(ols[idLast], Oligo.LOCATION);
					ols[idLast] = null;
					ols[oft.id] = new Oligo(oft.ol);		
				}
				ols[oft.id] = new Oligo(oft.ol);	
				loaded = true;
			}
			nxt = oft.id;
			oft.ol = new Oligo();
			if (of[nxt].readOligo(oft.ol)) {
				oft.id = nxt;
				pQueue.add(oft);
			}
		}
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			ofOut[chromosome].saveSBunkP(Oligo.LOCATION);
			ofOut[chromosome].dfn = ofOut[chromosome].dfns;
			ofOut[chromosome].closeP();
			Oligo.sortIndex = Oligo.LOCATION;
			ofOut[chromosome] = new OligoFile(Parameters.pathOut + Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2 + ".um.s.l", "", OligoFile.R);
			ofOut[chromosome].joinSortP();
			ofOut[chromosome].closeP();
		}
		of[0].close();
		of[1].close();

    }
    
    private void loadRef(String fl, int f) throws Exception {
    	String line;
    	aligne2Ref = new byte[Parameters.refSeqSize[f]];
    	aligne2Map = new byte[Parameters.refSeqSize[f]];
		for(int i = 0; i < Parameters.refSeqSize[f]; i++) {
			aligne2Ref[i] = -1;
			aligne2Map[i] = -1;
		}
		aligne2RefSize = Parameters.refSeqSize[f];
    	String sapFl = Parameters.pathDB + Parameters.DATABASE + fl + ".map0";
    	DataInputStream		strIns = new DataInputStream(new BufferedInputStream(new FileInputStream(sapFl), 127000));
		for (int i = 0; i < Parameters.refSeqSize[f] - Parameters.seedSize; i++) {
			aligne2Map[i] = strIns.readByte();
		}
		strIns.close();
		int ofset = 0;
		sapFl = Parameters.pathRefSeq + fl;
		BufferedReader reader = new BufferedReader(new FileReader(sapFl));
		while(reader.ready()) {
			line = reader.readLine();
			if (line.length() > 0 && line.charAt(0) != '>') {
				for (int i = 0; i < line.length(); i++) {
					aligne2Ref[ofset] = (byte) bp2Int(line.charAt(i));
					ofset++;
				}
			}
		}
		System.err.println(ofset + "\t" + Parameters.refSeqSize[f]);
		reader.close();
    }
    // Parameters.pathOut, Parameters.outFile,
    public  void seperate_SET() throws Exception { // , int[][] resultCount
    	int next;
    	int biggest_num_index = 0;
    	um0 = 0;
    	mm0 = 0;
    	mmum0 = 0;
    	um1 = 0;
    	mm1 = 0;
    	mmum1 = 0;
		float[] petHist = new float[Parameters.petHistSize];
		
		// -------------------------------------- 
		OligoFile[] ofUm = new OligoFile[Parameters.SEQ_VAR.length];
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			String file = Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2;
			ofUm[chromosome] = new OligoFile(Parameters.pathOut + file + ".um.s.l", Parameters.pathRefSeq + file, OligoFile.W);
		}
		OligoFile[] of = new OligoFile[2];
		of[0] = new OligoFile(Parameters.pathOut + "s.e.m0.s.id", "",	OligoFile.R);
		of[1] = new OligoFile(Parameters.pathOut + "s.e.m1.s.id", "",	OligoFile.R);
		strInPrb = new DataInputStream(new BufferedInputStream(new FileInputStream(Parameters.pathData + Parameters.outFile + ".prb"), 256000));
    	int oldSt = Oligo.sortIndex;
    	Oligo.sortIndex = Oligo.ID;
		PriorityQueue<OligoFile.OligoFileTop> pQueue = new PriorityQueue<OligoFile.OligoFileTop>();
		OligoFile.OligoFileTop oft;

		for (int j = 0; j < of.length; j++) {
			oft = of[j].new OligoFileTop();
			Oligo ol = new Oligo();
			of[j].readOligo(ol);
			oft.ol = ol;
			oft.id = j;
			pQueue.add(oft);
		}

		ArrayList<Oligo> temp = new ArrayList<Oligo>(); //((int) (of[0].numberOfOligos + of[1].numberOfOligos));

		int ID;
		thisSize = 0;
		thisId = 0;
		thisSide = 3;

		oft = pQueue.poll();
		next = oft.id;
		ID = oft.ol.id1;
		loadPrb(oft.ol);
		// ------------------------------------------------------ loadPrb
		temp.add(oft.ol);
		oft.ol = new Oligo();
		if (of[next].readOligo(oft.ol)) {
			oft.id = next;
			pQueue.add(oft);
		}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			loadPrb(oft.ol);
			// ------------------------------------------------------ loadPrb
			if (oft.ol.id1 == ID) {
				temp.add(oft.ol);
			} else {
				saveMSET(temp, ofUm); // this will, amung other, fill mm.mm and um.nm
				temp.clear();
				ID = oft.ol.id1;
				temp.add(oft.ol);
			}
			oft.ol = new Oligo();
			if (of[next].readOligo(oft.ol)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		saveMSET(temp, ofUm); // this will, amung other, fill mm.mm and um.nm
		of[0].close();
		of[1].close();
		strInPrb.close();
    	oldSt = Oligo.sortIndex;
		Oligo.sortIndex = Oligo.LOCATION;
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			ofUm[chromosome].saveSBunkP(Oligo.LOCATION);
			ofUm[chromosome].dfn = ofUm[chromosome].dfns;
			ofUm[chromosome].closeP();
			
			System.err.println(Parameters.pathOut + Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2 + ".um.s.l");
			
			ofUm[chromosome] = new OligoFile(Parameters.pathOut + Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2 + ".um.s.l", "", OligoFile.R);
			ofUm[chromosome].joinSortP();
			ofUm[chromosome].closeP();
		}
		Oligo.sortIndex = oldSt;
		temp.clear();
		Oligo.sortIndex = oldSt;
		System.err.println("\tum\tmm\tmmum");
		System.err.println("\t" + um0 + "\t" + mm0 + "\t" + mmum0);
		System.err.println("\t" + um1 + "\t" + mm1 + "\t" + mmum1);
   }   

    int totalPETs;
    
    public  void seperatePET1(String[] files) throws Exception { // to generate petHist from s.e.m0.id
    	int totalVPETs = 0;
    	int xx = 0;
    	totalPETs  = 0;
    	int[] hist = new int[Parameters.petHistSize];
    	int dist = 0;
    	for (int i = 0; i < Parameters.petHistSize; i++) {
    		hist[i] = 0;
    	}
		Oligo lastO;
		lastO    = new Oligo();
		lastO.ch = -1;

		Oligo[] ol = new Oligo[2];

    	ofMatch    = new OligoFile(Parameters.pathOut + Parameters.outFile + ".e.m0.s.id", "", OligoFile.R);
    	
    	//*
		while(ofMatch.available() > 0) { 
			ofMatch.readOligo(oligo[DATA]);
			if (oligo[DATA].isUM()) {
				ol[oligo[DATA].getSide()] = new Oligo(oligo[DATA]);
				if (ol[0] != null && ol[1] != null) {
					if (ol[0].id1 == ol[1].id1 && ol[0].ch == ol[1].ch) {
						dist = getDist(ol[0], ol[1]);
						if (petIsP(ol[0], ol[1])) {
			    			if (dist > 0 && dist < (Parameters.petHistSize - 1)) {
			    				hist[dist]++;
			    				totalVPETs++;
			    			} else if (dist >= (Parameters.petHistSize - 1)) {
			    				hist[Parameters.petHistSize - 1]++;
			    			} 
						} else {
							xx++;
						}
	    				totalPETs++;
					} 
					ol[0] = ol[1] = null;
				}
			}
		}
		ofMatch.close();
    	PrintWriter out;
		out = new PrintWriter(new BufferedWriter(new FileWriter(new File(Parameters.pathOut + "/petHist"))));
		out.println(totalPETs); //  + "\t" + totalVPETs
		for (int i = 0; i < Parameters.petHistSize; i++) {
			out.println( i + "\t" + hist[i]);
		}
		out.close();
    }   
    //*
    private boolean petIsP(Oligo o0, Oligo o1) {
    	if (o0.isRC()) {
    		if(o0.start < o1.start) {
    			return false;
    		} else {
    			return true;
    		}
    	} else {
    		if(o0.start < o1.start) {
    			return true;
    		} else {
    			return false;
    		}
    	}
    }
    //*/
    public int getDist(Oligo o1, Oligo o2) {
    	if(o1.start > (o2.start + o2.size)) {
    		return Math.abs(o1.start + o1.size - o2.start);
    	} else if (o2.start > (o1.start + o1.size)) {
    		return Math.abs(o2.start + o2.size - o1.start);
    	} else {
    		return -1;
    	}
    }
    
    private void loadPETSpan () {
    	String line;
		// write(Parameters.petSpanMin + "\t" + Parameters.petSpanMax + "\t" + biggest_num_index + "\t" + petHist[biggest_num_index] + "\n");
    	try {
    		BufferedReader mmbr = new BufferedReader(new FileReader(Parameters.pathOut + "mm.mm"));
    		line = mmbr.readLine();
    		StringTokenizer st = new StringTokenizer(line);
    		Parameters.petSpanMin = new Integer(st.nextToken()).intValue();
    		Parameters.petSpanMax = new Integer(st.nextToken()).intValue();
    		// Parameters.petSpanMin = new Integer(st.nextToken()).intValue();
    		mmbr.close();
    	} catch (Exception e) {
    		
    	}
    }
    
    long um0;
    long mm0;
    long mmum0;
    long um1;
    long mm1;
    long mmum1;
    // int  gap;
    public void SeparatePET2() throws Exception { // generate mm.mm, um.nm
    	int next;
    	int biggest_num_index = 0;
    	um0 = 0;
    	mm0 = 0;
    	mmum0 = 0;
    	um1 = 0;
    	mm1 = 0;
    	mmum1 = 0;
		float[] petHist = new float[Parameters.petHistSize];
		PrintWriter[] prbNext = new PrintWriter[Parameters.SEQ_VAR.length]; 
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			String file = Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2;
			System.err.println(Parameters.pathOut + file + ".prbNext");
			prbNext[chromosome] = new  PrintWriter(new FileWriter(new File(Parameters.pathOut + file + ".prbNext")));
		}
		PrintWriter   outUmNm = new PrintWriter(new FileWriter(new File(Parameters.pathOut + "um.nm")));
		PrintWriter   outMmMm = new PrintWriter(new FileWriter(new File(Parameters.pathOut + "mm.mm")));
		// -------------------------------------- loading prtHist info
		BufferedReader br = new BufferedReader(new FileReader(new File(Parameters.pathOut + "petHist")));
		float total_reads = (float) Integer.parseInt(br.readLine());
		total_reads = total_reads - (int) petHist[Parameters.petHistSize -1];
		for (int j = 0; j < Parameters.petHistSize; j++) {
			String st = br.readLine();
			if (j > Parameters.seedSize) {
				String[] sub_string = st.split("\t");
				petHist[j] = ((float) Integer.parseInt(sub_string[1])) / total_reads;
				if (j > 50 && j < 500 && petHist[j] > petHist[biggest_num_index])
					biggest_num_index = j;
			}
		}
		// biggest_num_index -= Parameters.seedSize; totalPETs
		br.close();
		Parameters.petSpanMax = 0;
		float sumPets = 0;
		for (int j = 0; j < Parameters.seedSize; j++) {
			sumPets += petHist[j];
		}
		for (int j = Parameters.seedSize; j < Parameters.petHistSize; j++) {
			sumPets += petHist[j];
			if (Parameters.petSpanMin == 0 && (petHist[j] > (petHist[biggest_num_index]/100) || sumPets >= 0.1)) {
				Parameters.petSpanMin = j;
			}
			if (Parameters.petSpanMax == 0 && j > biggest_num_index && (petHist[j] < (petHist[biggest_num_index]/100) || sumPets >= 0.9)) {
				Parameters.petSpanMax = j;
			}
		}
		Parameters.petSpanVar = Parameters.petSpanMax - Parameters.petSpanMin;
		outMmMm.write(Parameters.petSpanMin + "\t" + Parameters.petSpanMax + "\t" + biggest_num_index + "\t" + petHist[biggest_num_index] + "\n");
		System.err.println("petSpanMost:\t" + biggest_num_index + ",\tpetSpanMost%:\t" + petHist[biggest_num_index] + 
				"\npetSpanMin:\t" + Parameters.petSpanMin + "\tpetSpanMax:\t" + Parameters.petSpanMax + 
				",\tpetSpanVar:\t" + Parameters.petSpanVar + "\t<" + petHist[Parameters.petHistSize -1]);
		
		// -------------------------------------- 
		// DataOutputStream[] dosPrb = new DataOutputStream[Parameters.SEQ_VAR.length];
		/*
		OligoFile[] ofUm = new OligoFile[Parameters.SEQ_VAR.length];
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			String file = Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2;
			ofUm[chromosome] = new OligoFile(Parameters.pathOut + file + ".um.s.l", Parameters.pathRefSeq + file, OligoFile.W);
			// dosPrb[chromosome] = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(Parameters.pathOut + file + ".prb"), 256000));
		}
		//*/
		// String file = Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2;
		OligoFile ofUm = new OligoFile(Parameters.pathOut + "s.um.s.id", Parameters.pathRefSeq + "", OligoFile.W);
		OligoFile[] of = new OligoFile[2];
		of[0] = new OligoFile(Parameters.pathOut + "s.e.m0.s.id", "",	OligoFile.R);
		of[1] = new OligoFile(Parameters.pathOut + "s.e.m1.s.id", "",	OligoFile.R);

		int oldSt = Oligo.sortIndex;
    	Oligo.sortIndex = Oligo.ID;
		PriorityQueue<OligoFile.OligoFileTop> pQueue = new PriorityQueue<OligoFile.OligoFileTop>();
		OligoFile.OligoFileTop oft;

		for (int j = 0; j < of.length; j++) {
			oft = of[j].new OligoFileTop();
			Oligo ol = new Oligo();
			of[j].readOligo(ol);
			oft.ol = ol;
			oft.id = j;
			pQueue.add(oft);
		}

		ArrayList<Oligo> temp = new ArrayList<Oligo>(); //((int) (of[0].numberOfOligos + of[1].numberOfOligos));

		strInPrb = new DataInputStream(new BufferedInputStream(new FileInputStream(Parameters.pathData + Parameters.outFile + ".prb"), 256000));
		nextId = strInPrb.readInt();
		int olIDlast;
		thisId = 0;
    	Oligo.sortIndex = Oligo.ID;

		oft = pQueue.poll();
		next = oft.id;
		olIDlast = oft.ol.id1;
		loadPrbBfr(oft.ol);
		// ------------------------------------------------------ loadPrb
		temp.add(oft.ol);
		oft.ol = new Oligo();
		if (of[next].readOligo(oft.ol)) {
			oft.id = next;
			pQueue.add(oft);
		}
    	Oligo.sortIndex = Oligo.ID;
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			// System.err.println(oft.ol);
			// ------------------------------------------------------ loadPrb
			if (oft.ol.id1 == olIDlast) {
				loadPrbBfr(oft.ol);
				temp.add(oft.ol);
				// System.err.println("oft.ol.id1\t" + oft.ol.id1 + "\tside " + oft.ol.getSide());
			} else {
				saveMPET(temp, ofUm, prbNext, outUmNm, outMmMm, petHist); // this will, amung other, fill mm.mm and um.nm
				temp.clear();
				olIDlast = oft.ol.id1;
				loadPrbBfr(oft.ol);
				temp.add(oft.ol);
			}
			//*
			oft.ol = new Oligo();
			if (of[next].readOligo(oft.ol)) {
				oft.id = next;
				pQueue.add(oft);
			}
			//*/
		}
		saveMPET(temp, ofUm, prbNext, outUmNm, outMmMm, petHist); // this will, amung other, fill mm.mm and um.nm
		of[0].close();
		of[1].close();
		strInPrb.close();
    	// oldSt = Oligo.sortIndex;
		// Oligo.sortIndex = Oligo.LOCATION;
		/*
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			ofUm[chromosome].saveSBunkP(Oligo.LOCATION);
			// ofUm[chromosome].minID = id - 1;
			ofUm[chromosome].dfn = ofUm[chromosome].dfns;
			ofUm[chromosome].closeP();
			// dosPrb[chromosome].close();
			
			System.err.println(Parameters.pathOut + Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2 + ".um.s.l");
			
			ofUm[chromosome] = new OligoFile(Parameters.pathOut + Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2 + ".um.s.l", "", OligoFile.R);
			ofUm[chromosome].joinSortP();
			ofUm[chromosome].closeP();
			
		}
		//*/
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			prbNext[chromosome].close();
		}
		// String line;
		ofUm.closeP();
		Oligo.sortIndex = oldSt;
		temp.clear();
		/*
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			String file = Parameters.pathOut + Parameters.ch_name1 + Parameters.SEQ_VAR[chromosome] + Parameters.ch_name2 + ".prbNext";
			String files = file + "s";
			DataOutputStream sOut = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(files), Parameters.buffSize));
			String cmd = "sort -n " + file; // + " 2> " + files;
			System.err.println(cmd);
			Process ls_proc = Runtime.getRuntime().exec(cmd);
		    DataInputStream ls_in = new DataInputStream(ls_proc.getInputStream());
	    	while ((line = ls_in.readLine()) != null) { //ls_in.available() > 0) {
	    		sOut.writeBytes(line + "\n");
	    		System.err.println(line);
	    	}
	    	sOut.close();
	    	ls_in.close();
		}
		//*/
		/*
		for (int chromosome = 0; chromosome < Parameters.SEQ_VAR.length; chromosome++) {
			ofUm[chromosome].close();
			ofUm[chromosome].writeMeta();
		} //*/
		outUmNm.close();
		outMmMm.close();
		Oligo.sortIndex = oldSt;
		System.err.println("\tum\tmm\tmmum");
		System.err.println("\t" + um0 + "\t" + mm0 + "\t" + mmum0);
		System.err.println("\t" + um1 + "\t" + mm1 + "\t" + mmum1);
    }
    
    public void savePrb() throws Exception {
    	filePrb.writeInt(seq.id);
    	filePrb.writeByte(0);
    	filePrb.writeShort((short) seq.prbSize);
    	for (int ii = 0; ii < seq.prbSize; ii++) {
    		for (int jj = 0; jj < 4; jj++) {
    			filePrb.writeByte(seq._q[0][ii][jj]);
    		}
    	}
    }
    
    public boolean loadPrb(Oligo ol) throws Exception {
		byte pSide;
		int pId;
		short pSize;
		// System.err.println(ol);
		if (ol.id1 == thisId && ol.getSide() == thisSide) {
			ol.prb = new byte[4][thisSize];
			ol.prbSize = (short) thisSize;
    		if (ol.isRC()) {
		    	for (int ii = 0; ii < thisSize; ii++) {
		    		for (int b = 0; b < 4; b++) {
		    			ol.prb[comp[b]][ii] = (byte) Consensus.getQCal(lastQs[ii*4+b], ii);
		    		}
		    	}
    		} else {
		    	for (int ii = 0; ii < thisSize; ii++) {
		    		for (int b = 0; b < 4; b++) {
		    			ol.prb[b][ii] = (byte) Consensus.getQCal(lastQs[ii*4+b], ii);
		    		}
		    	}    			
    		}
		} else {
			while (true) {
				pId  = strInPrb.readInt();
				pSide = strInPrb.readByte();
				pSize = strInPrb.readShort();
				// System.err.println("pId:\t" + pId + "\tpSide:\t" + pSide + "\tpSize:\t" + pSize);
    			if (ol.id1 == pId && ol.getSide() == pSide) { 
    				thisSize = pSize;
    				thisId = pId;
    				thisSide = pSide;
    				ol.prb = new byte[4][pSize];
    				ol.prbSize = (short) pSize;
    				if (ol.isRC()) {
	    		    	for (int ii = 0; ii < pSize; ii++) {
	    		    		for (int b = 0; b < 4; b++) {
	    		    			lastQs[ii*4+b] = strInPrb.readByte();
	    		    			ol.prb[comp[b]][ii] = (byte) Consensus.getQCal(lastQs[ii*4+b], ii);
	    		    		}
	    		    	}
    				} else {
	    		    	for (int ii = 0; ii < pSize; ii++) {
	    		    		for (int b = 0; b < 4; b++) {
	    		    			lastQs[ii*4+b] = strInPrb.readByte();
	    		    			ol.prb[b][ii] = (byte) Consensus.getQCal(lastQs[ii*4+b], ii);
	    		    		}
	    		    	}	    					
    				}
    		    	break; //return; // true
    			} else {
    		    	for (int ii = 0; ii < pSize; ii++) {
    	    			strInPrb.readInt();
    		    	}	
    			}
    			if (ol.id1 > pId) {
    				System.err.println("\t<prbFileXXRC:\t" + pId + ", " + pSide + ">\t,Oligo:\t" + ol.id1 + "\t" + ol.getSide());
    				Thread.sleep(2000);
    				// return false;
    			}
	   		}
		}
		return true;
    }
    
    public boolean loadPrbBfr(Oligo ol) throws Exception {
		// System.err.println(" >> ------------------------ loadPrbBfr oligo: " + ol);
		// System.err.println(" >> ------------------------ loadPrbBfr oligo: " + nextId);
		if (ol.id1 < nextId) { // -------------------------------- move
			bfrLoadedSides = 0;
			while (ol.id1 < nextId) { 
				strInPrb.skipBytes(1);
				thisSize = strInPrb.readShort();
				strInPrb.skipBytes(thisSize * 4);
				try {
					nextId = strInPrb.readInt();
				} catch (Exception e) {
					nextId = -1;
				}
			}
		}
		if (ol.id1 == nextId) { // --------------------------- load bfr
			bfrLoadedSides = 0;
			while (ol.id1 == nextId ) { 
				thisId   = nextId;
				thisSide = strInPrb.readByte();
				thisSize = strInPrb.readShort();
				bfrLoadedSides++;
		    	for (int ii = 0; ii < thisSize; ii++) {
		    		for (int b = 0; b < 4; b++) {
		    			bfr[thisSide][b][ii] = strInPrb.readByte();
		    		}
		    	}
				try {
					nextId = strInPrb.readInt();
				} catch (Exception e) {
					nextId = -1;
				}
				// System.err.println("Loding:\t" + thisId + "\t" + thisSide + "\tnextId " + nextId);
			}
		}
		if (ol.id1 == thisId) {
			ol.prb = new byte[4][thisSize];
			ol.prbSize = (short) thisSize;
    		if (ol.isRC()) {
		    	for (int ii = 0; ii < thisSize; ii++) {
		    		for (int b = 0; b < 4; b++) {
		    			ol.prb[comp[b]][ii] = (byte) Consensus.getQCal(bfr[ol.getSide()][b][ii], ii);
		    		}
		    	}
    		} else {
		    	for (int ii = 0; ii < thisSize; ii++) {
		    		for (int b = 0; b < 4; b++) {
		    			ol.prb[b][ii] = (byte) Consensus.getQCal(bfr[ol.getSide()][b][ii], ii);
		    		}
		    	}    			
    		}
		}
		// Thread.sleep(100);
		return true;
    }
    
	private static void saveUmSET(ArrayList<Oligo> temp, OligoFile[] ofUm, int numberOfRedundent) throws Exception {
		int k = temp.size();
		if (k == 1 && !temp.get(0).isUM())
			temp.get(0).setUM(true);
			// System.err.println("saveUmSET\t" + k);
		if (k > 1) {
			if (!temp.get(0).isUM()) {
				temp.clear();
				return;
			}
	    	int oldSt       = Oligo.sortIndex;
	    	Oligo.sortIndex = Oligo.WEIGHT;
			Collections.sort(temp);
			Oligo.sortIndex = oldSt;
			numberOfRedundent++;
			if (!temp.get(0).isUM()) {
				temp.clear();
				return;
			}
		}
		ofUm[temp.get(0).ch].writeOligo(temp.get(0));
		temp.clear();
	}

	private void saveMSET(ArrayList<Oligo> temp, OligoFile[] of) throws Exception { // this will, amung other, fill mm.mm and um.nm
		// int oid;
		// int gap;
		// float weight;
		int oSortIndex;
		int noo = temp.size();
		// int distance;
		String RC0 = "";
		String RC1 = "";
		
		// oid = temp.get(0).id1;
		if (noo < 1) {
			return;
		} else if (noo == 1) {
			if (temp.get(0).isU0())
				um0++;
			else
				um1++;
			temp.get(0).setPET(false);
			temp.get(0).setUM(true);
			of[temp.get(0).ch].sort_WriteOligoP(new Oligo(temp.get(0)), Oligo.LOCATION);
			return;
		} else if ( noo > 1) {
			oSortIndex = Oligo.sortIndex;
			Oligo.sortIndex = Oligo.WEIGHT;
			Collections.sort(temp);
			Oligo.sortIndex = oSortIndex;
			if(temp.get(0).weight > 2 * temp.get(1).weight) {
				if (temp.get(0).isU0())
					um0++;
				else
					um1++;
				temp.get(0).setPET(false);
				temp.get(0).setUM(true);
				of[temp.get(0).ch].sort_WriteOligoP(new Oligo(temp.get(0)), Oligo.LOCATION);
			}
			return;
		} 
	}

	private void saveMPET(ArrayList<Oligo> temp, OligoFile of, PrintWriter[] prbNext, PrintWriter outUmNm, 
			PrintWriter outMmMm, float[] petHist) throws Exception { // this will, amung other, fill mm.mm and um.nm
		int oid;
		int gap;
		// int sGap;
		// float weight;
		int oSortIndex;
		int noo = temp.size();
		String RC0 = "";
		String RC1 = "";
		
		oid = temp.get(0).id1;
		// System.err.println("saveMPET\t" + oid + "\tsides\t" + bfrLoadedSides);

		if (noo < 1) {
			return;
		} else if (noo == 1) {
			if (!temp.get(0).isUM())
				return;
			if (temp.get(0).isU0())
				um0++;
			else
				um1++;
			temp.get(0).setPET(false);
			of.writeOligoP(new Oligo(temp.get(0)));
			if (temp.get(0).isRC())
				RC0 = "R";
			else
				RC0 = "F";
			// System.err.println("One read SET\t" + oid + "\tsides\t" + bfrLoadedSides);
			// Thread.sleep(100);
			outUmNm.write(oid + "\t" + temp.get(0).getSide() + "\t" + temp.get(0).ch + "\t" + temp.get(0).start + "\t" + temp.get(0).size + "\t" + RC0 + "\n");
			if (bfrLoadedSides == 2) {
				prbNext[temp.get(0).ch].write(temp.get(0).start + "\t" + temp.get(0).size + "\t" + temp.get(0).id1 + "\t" + temp.get(0).getCompSide() + "\t");
				if (temp.get(0).id1 != thisId)
					System.err.println(" A Error:\tOligo: " + temp.get(0).id1 + "\tthisId: " + thisId);
				for (int p = 0; p < thisSize; p++) {
					for (int b = 0; b < 4; b++) {
						prbNext[temp.get(0).ch].write(bfr[temp.get(0).getCompSide()][b][p] + "\t");
					}
				}
				prbNext[temp.get(0).ch].write("\n");
			}
			return;
		} else if ( noo == 2) {
			if(temp.get(0).getSide() != temp.get(1).getSide()) {
				for (int uu = 0; uu < 2; uu++) {
					if (temp.get(uu).isU0())
						um0++;
					else
						um1++;
				}
				temp.get(0).setUM(true);
				temp.get(1).setUM(true);
				oSortIndex = Oligo.sortIndex;
				Oligo.sortIndex = Oligo.ID;
				Collections.sort(temp);
				Oligo.sortIndex = oSortIndex;
				gap = getDist(temp.get(0), temp.get(1));
				if (Parameters.petSpanMin <= gap && gap <= Parameters.petSpanMax && temp.get(0).ch == temp.get(1).ch
						&& temp.get(0).isRC() != temp.get(1).isRC() && petIsP(temp.get(0), temp.get(1))) { 
					temp.get(0).setPET(true);
					temp.get(1).setPET(true);
					temp.get(0).petSize = (short) gap;
					temp.get(1).petSize = (short) gap;
				} else {
					temp.get(0).setPET(false);
					temp.get(1).setPET(false);
					if (temp.get(0).isRC()) {
						RC0 = "R";
					} else {
						RC0 = "F";
					}
					if (temp.get(1).isRC()) {
						RC1 = "R";
					} else {
						RC1 = "F";
					}
					outMmMm.write(oid + "\t1\t1\t" + temp.get(0).ch + "\t" + temp.get(0).start + "\t" + temp.get(0).size + "\t" + RC0 + "\t" + 
							temp.get(1).ch + "\t" + temp.get(1).start + "\t" + temp.get(1).size + "\t" + RC1 + "\n");
				}
				of.writeOligoP(new Oligo(temp.get(0)));
				of.writeOligoP(new Oligo(temp.get(1)));
			}  else { 
				if (temp.get(0).isU0())
					mm0++;
				else
					mm1++;
			} 
			return;
		} else { // ------------------------------------------------------------ noo >= 3
			int[] countSid = new int[2];
			countSid[0] = 0;
			countSid[1] = 0;
			for (int q = 0; q < noo; q++) {
				countSid[temp.get(q).getSide()]++;
			}
			if (countSid[0] == 0 || countSid[1] == 0) {                                             // this for later
				if (temp.get(0).isU0())
					mm0++;
				else
					mm1++;
				return;					
			} else if (countSid[0] == 1 || countSid[1] == 1) { // one to many
				oSortIndex = Oligo.sortIndex;
				Oligo.sortIndex = Oligo.ID;
				Collections.sort(temp);
				Oligo.sortIndex = oSortIndex;
				ArrayList<Oligo> possibleOligos = new ArrayList<Oligo>(noo - 1);
				int uqSide;
				int uqSideIdx = 0;
				int uqChr = -1;
				int distance = 0;

				if (countSid[0] == 1) {
					uqSide    = 0;
					uqSideIdx = 0;
				} else {
					uqSide    = 1;
					uqSideIdx = noo - 1;
				}
				if (temp.get(uqSideIdx).isU0())
					um0++;
				else
					um1++;
				uqChr = temp.get(uqSideIdx).ch;

				//  ----------------------------- save all to mm file
				outMmMm.write(oid + "\t" + countSid[0] + "\t" + countSid[1]);
				for (int j = 0; j < noo; j++) {
					if (temp.get(j).isRC())
						RC0 = "R";
					else
						RC0 = "F";
					outMmMm.write("\t" + temp.get(j).ch + "\t" + temp.get(j).start + "\t" + temp.get(j).size + "\t" + RC0); 
				}
				outMmMm.write("\n");

				for (int j = 0; j < noo; j++) {
					if ( j != uqSideIdx) {
						if (temp.get(j).ch == uqChr) {
							distance = getDist(temp.get(j), temp.get(uqSideIdx));
							if (distance > 0 && distance <= Parameters.petSpanMax &&
									petIsP(temp.get(j), temp.get(uqSideIdx)))
								possibleOligos.add(temp.get(j));
						}
					}
				}
				
				if (possibleOligos.size() == 0) { 
					of.writeOligoP(new Oligo(temp.get(uqSideIdx)));
					if (temp.get(uqSideIdx).isRC())
						RC0 = "R";
					else
						RC0 = "F";
					outUmNm.write(oid + "\t" + temp.get(uqSideIdx).getSide() + "\t" + temp.get(uqSideIdx).ch + "\t" + 
							temp.get(uqSideIdx).start + "\t" + temp.get(uqSideIdx).size + "\t" + RC0 + "\n");
					if (bfrLoadedSides == 2) {
						if (temp.get(uqSideIdx).id1 != thisId)
							System.err.println(" B Error:\t" + temp.get(uqSideIdx).id1 + "\t" + thisId);
						prbNext[temp.get(uqSideIdx).ch].write(temp.get(uqSideIdx).start + "\t" + temp.get(uqSideIdx).size + "\t" + temp.get(uqSideIdx).id1 + 
								"\t" + temp.get(uqSideIdx).getCompSide() + "\t");
						for (int p = 0; p < thisSize; p++) {
							for (int b = 0; b < 4; b++) {
								prbNext[temp.get(uqSideIdx).ch].write(bfr[temp.get(uqSideIdx).getCompSide()][b][p] + "\t");
							}
						}
						prbNext[temp.get(uqSideIdx).ch].write("\n");
					}
					return;
				} else if (possibleOligos.size() == 1) { // (1/1)
					if (possibleOligos.get(0).isU0())
						mmum0++;
					else
						mmum1++;
					possibleOligos.get(0).setPET(true);
					temp.get(uqSideIdx).setPET(true);
					gap = getDist(possibleOligos.get(0), temp.get(uqSideIdx));
					possibleOligos.get(0).petSize = (short) gap;
					temp.get(uqSideIdx).petSize = (short) gap;
					if(uqSide == 0) {
						of.writeOligoP(new Oligo(temp.get(uqSideIdx)));
					}
					of.writeOligoP(new Oligo(possibleOligos.get(0)));
					if(uqSide == 1) {
						of.writeOligoP(new Oligo(temp.get(uqSideIdx)));
					}
					return;
				} else { // procces (1/n)
					float[] probabilities = new float[possibleOligos.size()];
					int highest_index = -1;
					for ( int j = 0; j < possibleOligos.size(); j++) {
						distance = getDist(temp.get(j), temp.get(uqSideIdx));
						if(distance > 0 && distance < probabilities.length) {
							probabilities[j] = petHist[distance] * possibleOligos.get(j).weight;
							if (highest_index < 0 || probabilities[j] > probabilities[highest_index]) {
								highest_index = j;
							}
						}
					}
					if (highest_index >= 0) {
						possibleOligos.get(highest_index).setPET(true);
						possibleOligos.get(highest_index).setUM(false);
						temp.get(uqSideIdx).setPET(true);
						temp.get(uqSideIdx).setUM(true);
						gap = getDist(possibleOligos.get(highest_index), temp.get(uqSideIdx));
						possibleOligos.get(highest_index).petSize = (short) gap;
						temp.get(uqSideIdx).petSize = (short) gap;
						if(uqSide == 0) {
							of.writeOligoP(new Oligo(temp.get(uqSideIdx)));
						}
						of.writeOligoP(new Oligo(possibleOligos.get(highest_index)));
						if(uqSide == 1) {
							of.writeOligoP(new Oligo(temp.get(uqSideIdx)));
						}
						if (possibleOligos.get(highest_index).isU0())
							mmum0++;
						else
							mmum1++;
					} else {
						temp.get(uqSideIdx).setPET(false);
						temp.get(uqSideIdx).setUM(true);
						of.writeOligoP(new Oligo(temp.get(uqSideIdx)));
					}
					return;
				} 
			} // else many to many
		} 
	}
    
	private void update_OligoWeightSNPs(int d) { // uses oligo[DATA]
		if (!(oligo[d].isPET() && oligo[d].isUM())) {
			oligo[d].weight = oligo[d].weight * Parameters.setMaxWeight;
		}
		return;
	}
	
	private void updateOligoWeightEx() { // uses oligo[DATA]
		float w;
		float sum = 0;
		int max;
		int seqPoint;
		if (oligo[DATA].u0 < oligo[DATA].size)
			max = oligo[DATA].u0;
		else
			max = oligo[DATA].size;
		w = 1;
		for (int i = 0; i < max; i++) {
			if(oligo[DATA].isRC()) {
				seqPoint = oligo[DATA].size - 1 - i;
			} else {
				seqPoint = i;
			}
			w = w * seq.p[i][Oligo.getBpInt(oligo[DATA].getBPat(seqPoint), 1)];
		}
		for (int i = max; i < oligo[DATA].size; i++) {
			if(oligo[DATA].isRC()) {
				seqPoint = oligo[DATA].size - 1 - i;
			} else {
				seqPoint = i;
			}
			sum += seq.p[i][Oligo.getBpInt(oligo[DATA].getBPat(seqPoint), 1)];
		}
		sum = sum * 0.999f / oligo[DATA].u0;
		oligo[DATA].weight = w + sum ; //((float) 0.5 + w / 2) * sum / seq.prbSize;
		if (oligo[DATA].weight > 0.999)
			oligo[DATA].weight = (float) 0.999;
		return;
	}
	
	
	private int seedCut;
	
	public void readPrbCal() {
		try {
		BufferedReader readerQCAL2 = new BufferedReader(new FileReader(Parameters.pathOut + "prbCal2d"));
		// BufferedReader readerQCAL1 = new BufferedReader(new FileReader(Parameters.pathOut + "prbCal1d"));
		String line = readerQCAL2.readLine();
		StringTokenizer st = new StringTokenizer(line);
		int prbSizeMax = st.countTokens() - 2;
		while (readerQCAL2.ready()) {
			line = readerQCAL2.readLine();
		}
		for (int i3 = 0; i3 < 81; i3++) {
			for (int i2 = 0; i2 < prbSizeMax; i2++) { // Parameters.seedSize
				if (prbCal[0][i2][i3] + prbCal[1][i2][i3] > 0)
					q2prb[i2][i3] = prbCal[0][i2][i3] / (prbCal[0][i2][i3] + prbCal[1][i2][i3]);		
				else
					q2prb[i2][i3] = (float) 0.0001;
				if (q2prb[i2][i3] < 0.0001)
					q2prb[i2][i3] = Parameters.Q2P((int) ((i3-40) * 0.93));
				if (q2prb[i2][i3] > 0.9999)
					q2prb[i2][i3] = Parameters.Q2P((int) ((i3-40) * 0.93));
			}
		}/*
		for (int i3 = 0; i3 < 81; i3++) {
			prbCal1[0][i3] = prbCal1[1][i3] = 0;
			writerQCAL2.write((i3 - 40) + "\t" + Parameters.trim(Parameters.Q2P(i3-40) * 100) + "\t");
			for (int i2 = 0; i2 < prbSizeMax; i2++) { // Parameters.seedSize
				writerQCAL2.write(Parameters.P2Q(q2prb[i2][i3]) + "\t");
				prbCal1[0][i3] += prbCal[0][i2][i3];
				prbCal1[1][i3] += prbCal[1][i2][i3];
			}
			if (prbCal1[0][i3] + prbCal1[1][i3] > 0)
				tmp = prbCal1[0][i3] / (prbCal1[0][i3] + prbCal1[1][i3]);
			else
				tmp = 0;
			writerQCAL1.write((i3 - 40) + "\t" + Parameters.trim(Parameters.Q2P(i3-40) * 100) + 
					"\t" + Parameters.P2Q(tmp) + "\t" + Parameters.trim(tmp * 100) + "\n");
			writerQCAL2.write("\n");
		}//*/
		readerQCAL2.close();
		// writerQCAL1.close();
		} catch (Exception e) {
			
		}
	}
	/*
	public  void prbCall() throws Exception {
		boolean drc = false;
		long countReads = 0;
		long countExtra = 0;
		int prbPoint;
		int bs;
		int readsUsed = Parameters.prbCallUse / Parameters.SEQ_VAR.length;
		int readCall;
		for (int i1 = 0; i1 < 2; i1++)
			for (int i2 = 0; i2 < 93; i2++)
				for (int i3 = 0; i3 < 81; i3++)
					prbCal[i1][i2][i3] = 0;
		// int sidPoint;
		// int seqPoint;
		// int direction;
		int prbSizeMax = Parameters.seedSize;
		for(int i = 0; i < Parameters.SEQ_VAR.length; i++) {
			readCall = 0;
			openPrbIn(Parameters.pathData + Parameters.outFile + ".prb");
			String file = Parameters.ch_name1 + Parameters.SEQ_VAR[i] + Parameters.ch_name2;
			ofOne    = new OligoFile(Parameters.pathOut + file + ".um.s.id", "", OligoFile.R); // read to DATA
			while(ofOne.available() > 0) {
				if (readCall >= readsUsed) {
					break;
				} else {
					readCall++;
				}
				countReads++;
				if (countReads % 5000000 == 0) {
					System.err.println(file + "\tRead:\t" + countReads/1000000 + "M");
				}
				ofOne.readOligo(oligo[DATA]);
				if (oligo[DATA].isU0() && oligo[DATA].isPET() && !oligo[DATA].isLQ() ) { // && !oligo[DATA].isRC()
					
					if (drc) 
						System.err.println(oligo[DATA] + "\t" + seq.size);
					
					while(!read_Q_Comp(oligo[DATA].id1, oligo[DATA].getSide(), oligo[DATA].isRC()));
					if (seq.prbSize > prbSizeMax)
						prbSizeMax = seq.prbSize;
					
					for (int i2 = 0; i2 < oligo[DATA].size; i2++) { // seq.size
						if (oligo[DATA].isRC()) { 
							prbPoint = oligo[DATA].size - 1 - i2; 
						} else {
							prbPoint = i2;
						}
						bs = Oligo.getBpInt(oligo[DATA].getBPat(i2), 1); // i2
						if (drc) 
							System.err.print(i2 + "\t" + bs);
						for (int b = 0; b < 4; b++) {
							if (drc)
								System.err.print("\t" + seq._q[0][prbPoint][b]);
							if (b == bs)
								prbCal[0][prbPoint][seq._q[0][prbPoint][b]+40]++;
							else
								prbCal[1][prbPoint][seq._q[0][prbPoint][b]+40]++;
						}
						if (drc)
							System.err.println();
					} 
					for (int i2 = oligo[DATA].size; i2 < seq.prbSize; i2++) {
						bs = Oligo.getBpInt(oligo[DATA].getBPat(i2), 1); // i2
						countExtra++;
						for (int b = 0; b < 4; b++) {
							if (b == bs)
								prbCal[0][i2][seq._q[0][i2][b]+40]++;
							else
								prbCal[1][i2][seq._q[0][i2][b]+40]++;
						}
					}
					if (drc) 
						Thread.sleep(1000);
				}
			} // end reading from oligoFile
			closePrbIn();
		}
		System.err.println("countExtra:\t" + countExtra);
		// write prb quality to files
		float tmp;
		float [][] prbCal1 = new float[2][101];
		BufferedWriter writerQCAL2 = new BufferedWriter(new FileWriter(Parameters.pathOut + "prbCal2d"));
		BufferedWriter writerQCAL1 = new BufferedWriter(new FileWriter(Parameters.pathOut + "prbCal1d"));
		writerQCAL2.write("Q\tP");
		for (int i2 = 0; i2 < prbSizeMax; i2++) 
			writerQCAL2.write("\t" + i2);
		writerQCAL2.write("\n");
		//*
		for (int i3 = 0; i3 < 81; i3++) {
			for (int i2 = 0; i2 < prbSizeMax; i2++) { // Parameters.seedSize
				if (prbCal[0][i2][i3] + prbCal[1][i2][i3] > 0)
					Consensus.iq2prb[i2][i3] = q2prb[i2][i3] = prbCal[0][i2][i3] / (prbCal[0][i2][i3] + prbCal[1][i2][i3]);		
				else
					Consensus.iq2prb[i2][i3] = q2prb[i2][i3] = (float) 0.0001;
			}
		}
		///
		for (int i3 = 0; i3 < 81; i3++) {
			prbCal1[0][i3] = prbCal1[1][i3] = 0;
			writerQCAL2.write((i3 - 40) + "\t" + Parameters.trim(Parameters.Q2P(i3-40) * 100) + "\t");
			for (int i2 = 0; i2 < prbSizeMax; i2++) { // Parameters.seedSize
				writerQCAL2.write(Parameters.P2Q(q2prb[i2][i3]) + "\t");
				prbCal1[0][i3] += prbCal[0][i2][i3];
				prbCal1[1][i3] += prbCal[1][i2][i3];
			}
			if (prbCal1[0][i3] + prbCal1[1][i3] > 0)
				tmp = prbCal1[0][i3] / (prbCal1[0][i3] + prbCal1[1][i3]);
			else
				tmp = 0;
			writerQCAL1.write((i3 - 40) + "\t" + Parameters.trim(Parameters.Q2P(i3-40) * 100) + 
					"\t" + Parameters.P2Q(tmp) + "\t" + Parameters.trim(tmp * 100) + "\n");
			writerQCAL2.write("\n");
		}
		writerQCAL2.close();
		writerQCAL1.close();
		// End write prb quality to files
		int tmpQ;
		for (int i3 = 0; i3 < 81; i3++) {
			for (int i2 = 0; i2 < Parameters.seedSize; i2++) { // Parameters.seedSize
				tmpQ = i3-40;
				if (tmpQ > Parameters.q_Cal[i2]) {
					tmpQ = Parameters.q_Cal[i2];
				} else if (tmpQ < -Parameters.q_Cal[i2]) {
					tmpQ = -Parameters.q_Cal[i2];
				} 
				Consensus.iq2prb[i2][i3] = q2prb[i2][i3] = Parameters.Q2P(tmpQ);  // q2prb[idx][q+40]
			}
		}
	}
	//*/
	/*
    public  void generate_Map(String refName) throws Exception {
		int ID0    = -1;
		int start0 = 0;
		boolean isRC0 = true;
		int[] p = {0, 0};
    	// long  idxCv = 0;
    	int start1 = 0;
		int span;
		int notPets = 0;
		int countPets = 0;
		double sAvrg = 0;
		int refPoint;
    	int prbPoint;
		int direction;
		long countReads = 0;
		long prbSize  = 0;
		long readSize = 0;

		spanAvr   = 0;
		petGapMin = Parameters.petSpanMin - 2 * Parameters.seedSize;
		petGapMax = Parameters.petSpanMax - 2 * Parameters.seedSize;
		if (petGapMin < 0)
			petGapMin = 0;
		System.err.println("Parameters.petSpanMin:\t" + Parameters.petSpanMin 
				+ ",\tParameters.petMost:\t" + Parameters.petSpanMost 
				+ ",\tParameters.petSpanMax:\t" + Parameters.petSpanMax
				+ ",\tParameters.petSpanVar:\t" + Parameters.petSpanVar);
		System.err.println(Parameters.pathOut + refName + ".um.s.id");
		ofOne    = new OligoFile(Parameters.pathOut + refName + ".um.s.id", "", OligoFile.R); // read to DATA
    	getMeta(ofOne);
		openPrbIn(Parameters.pathData + Parameters.outFile + ".prb");
		while(ofOne.available() > 0) { 
			countReads++;
			if (countReads % 5000000 == 0) {
				System.err.println(refName + "\tRead:\t" + countReads/1000000 + "M");
			}
			ofOne.readOligo(oligo[DATA]);
			if(oligo[DATA].isRC()) {
				while(!readPrb2dCallComp(oligo[DATA].id1, oligo[DATA].getSide()));
				start1 = oligo[DATA].start + oligo[DATA].size - Parameters.seedSize;
			} else { 
				while(!readPrb2dCall(oligo[DATA].id1, oligo[DATA].getSide()));
				start1 = oligo[DATA].start;
			}
			updateOligoWeightSNPs(DATA);
			prbSize += seq.prbSize;
			readSize += oligo[DATA].size;
			//*    ---------------------------------- this section is to fill the gshift array
			if(Parameters.ETag == Parameters.PET) {
				if (oligo[DATA].id1 == ID0) {
					if (start0 > start1) {
						p[1] = start0;
						p[0] = start1 + Parameters.seedSize;
					} else {
						p[1] = start1;
						p[0] = start0 + Parameters.seedSize;
					}
					if (p[0] > p[1])
						p[0] = p[1];
					span = p[1] - p[0];

					if (!oligo[DATA].isPET()) { // for gShift
						if (span > 0 && span <= (petGapMax + Parameters.petSpanVar) && isRC0 != oligo[DATA].isRC()) {
							if (span <= petGapMin)
								for (int i = p[0]; i < p[1]; i++) {
									map.gshift[i]--;
								}
							else if (span >= petGapMax)
								for (int i = p[0]; i < p[1]; i++) {
									map.gshift[i]++;
								}
							notPets++;
						}
					} else {
						sAvrg += span;
						countPets++;
					}
					ID0   = -1;
				} else {
					start0 = start1;
					ID0    = oligo[DATA].id1;
					isRC0  = oligo[DATA].isRC();
				} 
			}
			// --------------------------------------------------------------------------------------------------
			for (int j = 0; j < oligo[DATA].size; j++) { 
				refPoint = oligo[DATA].start + j;
				if (oligo[DATA].isRC()) { 
					prbPoint = oligo[DATA].size - 1 - j; 
					direction = 1;
				} else {
					prbPoint = j;
					direction = 0;
				}

				for (int i = 0; i < 4; i++) {
					readP[i] = seq.p[prbPoint][i];
				}
				map.bidx[refPoint] += prbPoint;
				map.cnt[2][refPoint]++;
				map.info[refPoint] += seq.baseInfo(prbPoint);
				map.updatePb(refPoint, prbPoint, 2);
			}
			for (int j = oligo[DATA].size; j < seq.prbSize; j++) { 
				if (oligo[DATA].isRC()) { 
					direction = 1;
					refPoint = oligo[DATA].start + oligo[DATA].size - 1 - j;
				} else {
					direction = 0;
					refPoint = oligo[DATA].start + j;
				}
				for (int i = 0; i < 4; i++) {
					readP[i] = seq.p[j][i];
				}
				map.updatePb(refPoint, j, direction);
				map.cnt[direction][refPoint]++;
			}
		} // end reading from oligoFile
		ofOne.close();
		closePrbIn();

		map.avrCnt[0] = map.avrCnt[1] = map.avrCnt[2] = 0;
		for (int i = 0; i < map.size; i++) {
			for (int j = 0; j < 3; j++) {
				map.avrCnt[j] += map.cnt[j][i];				
			}
			for (int cv = 0; cv < map.avrCrv.length; cv++) {
				if (map.cnt[2][i] > cv) {
					map.avrCrv[cv]++;
				}
			}
			if(map.cnt[2][i] > 0) {
				map.bidx[i] = map.bidx[i] / map.cnt[2][i];
				// clean bidx after shifting cnt
			} else {
				map.bidx[i] = Parameters.seedSize / 2;
			}
		}
		for (int j = 2; j >= 0; j--) {
			map.avrCnt[j] = map.avrCnt[j] / map.size;				
		}
		for (int cv = 0; cv < map.avrCrv.length; cv++) {
			map.avrCrv[cv] = map.avrCrv[cv] / map.size;
		}
		prbAvrSize   = (int) Math.round(((double) prbSize ) / countReads);
		readAvrSize  = (int) Math.round(((double) readSize) / countReads);
		prbAvrSizeF  = (float) ((double) prbSize ) / countReads;
		readAvrSizeF = (float) ((double) readSize) / countReads;
    }
    //*/
    //* locateSNPs // getMPBPInt(int dr, int l)
    
    public void markSNPs(String file) throws Exception {
    	int bidxLow  = Math.round((Parameters.seedSize / 2) * (float) 0.7) + 1;
    	int bidxHigh = Math.round((Parameters.seedSize / 2) * (float) 1.3) + 1;
    	System.err.println("seedSize:\t" + Parameters.seedSize + "\tbidxLow:\t" + bidxLow + "\tbidxHigh:\t" + bidxHigh);
    	for (int i = 0; i < map.size; i++) {
    		if(map.info[i] > Parameters.minInfo) {
           		if (map.detectSnp(i, 2, Parameters.cutSNP0)) { 
           			map.setSnp(2, i, true);
        		}
			}
    	} // end for loop
    	int pL = 0;
    	int pH = 0;
    	int countSNPs   = 0;
    	int toBeDeleted = 0;
    	for (pH = Parameters.exSize; pH < map.size; pH++) {
    		if (map.isSnp(2, pH)) {
    			countSNPs++;
        		if (countSNPs > 1) {
        			if (toBeDeleted == 0)
        				toBeDeleted++;
        			toBeDeleted++;
        		}
    		}
    		if (map.isSnp(2, pL)) {
    			countSNPs--;
        		if (toBeDeleted > 0) {
        			map.setSnp(2, pL, false);
        			toBeDeleted--;
        		}
    		}    		
    		pL++;
    	}
    	int countCrv;
    	for (int direction = 0; direction < 2; direction++) {
        	countSNPs = 0;
        	countCrv  = 0;
        	pL        = -1;
        	for (pH = Parameters.exSize; pH < map.size; pH++) {
        		if (map.cnt[direction][pH] > 0) {
        			if (pL == -1)
        				pL = pH;
        			countSNPs++;
        			countCrv += map.cnt[direction][pH];
            		if (countCrv > countSNPs + 1 && pH > pL + 1) {
            			while (pL <= pH) {
            				map.setSnp(direction, pL, true);
            				pL++;
            			}
            		} 
        		} else {
        			pL        = -1;
        			countCrv  = 0;
        			countSNPs = 0;
        		}
        	}
        	for (int i = Parameters.exSize; i < map.size; i++) {
        		if (map.isSnp(direction, i)) {
               		if (!map.detectSnp(i, direction, Parameters.cutSNP0)) { 
               			map.setSnp(direction, i, false);
            		}
        		} else {
        			map.cnt[direction][i] = 0;
        		}
        	}
    	}
    }
    
    // int gapHeteroCrv;
	public void divideSeq_1(String file) throws Exception {
		int     smothing = 3;
		boolean gap = true;
		long    sumSeedCnt = 0;
		int     countSub   = 0;
		int     lastSub    = 0;
		// int coveredCut = Math.round(UX * 2 / 3);
		// gapHeteroCrv = Parameters.snpMinCoverHetero / 2;
		// if (gapHeteroCrv < 1)
		// 	gapHeteroCrv = 1;
		float  bidxCut = Parameters.seedSize / 4;

		System.err.println("UX .......:\t" + Parameters.trim(UXf));
		System.err.println("seedCut ..:\t" + seedCut);
		System.err.println("bidxCut ..:\t< " + (bidxCut-1) + "\t>" + (Parameters.seedSize - bidxCut));
		System.err.println("--------------------------------------------");

		for (int i = 0; i < smothing; i++) {
			sumSeedCnt += map.cnt[2][i] - map.cnt[1][i] - map.cnt[0][i];
		}
		for (int i = smothing; i < map.size; i++) {
			sumSeedCnt += (map.cnt[2][i]   - map.cnt[2][i - smothing]) - (map.cnt[1][i]   - map.cnt[1][i - smothing]) - (map.cnt[0][i]   - map.cnt[0][i - smothing]);
			if (!gap && sumSeedCnt / smothing < seedCut &&
					(map.bidx[i - 1] > ((float) Parameters.seedSize - bidxCut) || 
					sumSeedCnt == 0 ||
					i == map.size -1)) { 
				gap = true;
			} else if (gap && 
					(map.bidx[i - 1] < (bidxCut - 1) || 
					sumSeedCnt / smothing >= seedCut)) { // Parameters.snpMinCoverHomo
				gap = false;
				countSub++;
			}
			map.setGap(i - 1, gap);
			if(gap) {
				map.subSeq[i - 1] = 0;
			} else {
				map.subSeq[i - 1] = countSub;
			}
		}
    	//*
    	// gap = false;
		// --------------------------------- GAPs pass 2
    	lastSub = 1;
    	for (int i = 3; i < map.size - 1; i++) {
    		if (map.isGap(i)) {
    			if (!map.isGap(i - 1) && map.isGap(i + 1)) {
        			if (  (map.bidx[i] > ((float) Parameters.seedSize / 2 + 1)) &&
        				  (  ((map.bidx[i-4] + map.bidx[i-3] + map.bidx[i-2] + map.bidx[i-1])/4 < map.bidx[i]) || 
        					 (map.bidx[i] <= map.bidx[i + 1])
        				  )
        			   ) {
        				map.subSeq[i] = lastSub;
        				map.setGap(i, false);
        			}
        		}
    			if (!map.isGap(i - 1) && map.isGap(i)) {
    				lastSub++;
    			}
    		} else {
    			map.subSeq[i] = lastSub;
    		}
    	}
		int j = 0;
		for (int i = 0; i < map.size; i++) {
			if (map.isGap(i)) {
				map.setNeighbourhood(i, true);
				j = 0;
			} else {
				if (j < (Parameters.maxPrbLine - Parameters.seedSize)) {
					map.setNeighbourhood(i, true);
					j++;
				}
			}
		}
		for (int i = map.size - 1; i >= 0; i--) {
			if (map.isGap(i)) {
				j = 0;
			} else {
				if (j < (Parameters.maxPrbLine - Parameters.seedSize)) {
					map.setNeighbourhood(i, true);
					j++;
				}
			}
		}
    	//*/
		System.err.println("countSub:\t" + countSub);
	}
//*/
	//*
	public void fillPET(String file) throws Exception {
		// -----------------------------------------------------------------------------------------------
		//*
		int countOl = 0;
		boolean isRC0 = true;
		int ID0    = -1;
		int start0 = 0;
		int start1 = 0;
		int span;
		int[] p = {0, 0};
    	for (int i = 0; i < map.size; i++) {
			detectSNP(i);
    	}

    	countOl = 0;
		OligoFile of    = new OligoFile(Parameters.pathOut + file + ".um.s.id", "", OligoFile.R);
		Oligo ol = new Oligo();
		
		System.err.println(" ------------------------------------");
		System.err.println("petEnds + Parameters.petSpan:\t" + (petGapMax + Parameters.petSpanVar));
		long countWhat  = 0;
		long countWhatP = 0;
		long countWhatN = 0;
		while(of.available() > 0) { 
			countOl++;
			if (countOl % 5000000 == 0) {
				System.err.println("2nd Pass: " + file + "\tRead:\t" + countOl/1000000 + "M");
			}
			of.readOligo(ol);
			if (ol.id1 == ID0) {
				if (ol.isRC()) {
					start1 = ol.start + ol.size - Parameters.seedSize;
				} else {
					start1 = ol.start;
				}
				if (start0 > start1) {
					p[1] = start0;
					p[0] = start1;
				} else {
					p[1] = start1;
					p[0] = start0;
				}
				p[0] += Parameters.seedSize;
				if (p[0] > p[1])
					p[0] = p[1];
				if (isRC0 != ol.isRC()) {
					span = p[1] - p[0];
					if (span > 0 && span <= (petGapMax + Parameters.petSpanVar)) {
						for (int i = p[0]; i < p[1]; i++) {
							if (map.gshift[i] < 0 && span <= petGapMax) {
								countWhatN++;
								map.gval[i] = map.gval[i] + span;
								map.gcnt[i]++;
							} else if (map.gshift[i] > 0 && span >= petGapMin) {
								countWhatP++;
								map.gval[i] = map.gval[i] + span;
								map.gcnt[i]++;
							} else if (map.gshift[i] == 0 && span >= petGapMin && span <= petGapMax) {
								countWhat++;
								map.gval[i] = map.gval[i] + span;
								map.gcnt[i]++;
							} 
						}
					}
				}
				ID0   = -1;
			} else {
				if (ol.isRC()) {
					start0 = ol.start + ol.size - Parameters.seedSize;
				} else {
					start0 = ol.start;
				}
				ID0    = ol.id1;
				isRC0  = ol.isRC();
			}
		}
		of.close();
	}

    float wsnpp;
	float nzCvr;
    
    public  void detectSNP(int idx) throws Exception {
    	/*
    	map.setSnpType(idx, 'N');
    	if ((map.snp_Sc[2][idx] + map.snp_Sc[1][idx] < Parameters.snpMinCoverHomo  * 0.99 && !map.isNeighbourhood(idx))) {
    		return;
    	}

    	int heteroCvr;
    	int homoCvr;
    	float heteroPer;
    	float homoPer;

    	if (map.isNeighbourhood(idx)) { 
    		heteroCvr = gapHeteroCrv;
    		heteroPer = Parameters.snpPerWHetero / 2;
        	if (map.isGap(idx)) { 
        		homoCvr = Parameters.snpMinCoverHomo;
        		homoPer = Parameters.snpPerWHomo;
        	} else {
        		homoCvr = Parameters.snpMinCoverHomo;
        		homoPer = Parameters.snpPerWHomo;
        	}
    	} else {
    		heteroCvr = Parameters.snpMinCoverHetero;
    		heteroPer = Parameters.snpPerWHetero;
    		homoCvr = Parameters.snpMinCoverHomo;
    		homoPer = Parameters.snpPerWHomo;
    	}

    	scoreSNP(idx); // map.wsnpp[l]; map.setDirection(l, 'R'); map.setMPBP(l, mpbp); map.setSnp(l, false);
   		if (map.snp_Sc[2][idx] + map.snp_Sc[1][idx] > homoCvr && map.wsnpp[idx] > homoPer) {
      		map.setSnpType(idx, 'M');
      		map.setSnp(idx, true);
   		} else if ( map.snp_Sc[2][idx] + map.snp_Sc[1][idx] > heteroCvr && map.wsnpp[idx] > heteroPer) {
       	    map.setSnpType(idx, 'T');
       	    map.setSnp(idx, true);
   		}
    	if (map.isNeighbourhood(idx) || (map.isSnp(idx) && map.getSnpType(idx).charAt(1) == 'e')) { 
    		scoreSNP_N(idx);
    	}
    	//*/
       	return;
    }
    
    public void scoreSNP_N(int l) {
    	/*
		for (int dr = 1; dr < 3; dr++) {
			if (map.snp_Sc[dr][l] > 0.5) {
				getMPB(dr-1, l);
				map.set_MPBP(dr-1, l, mpbp);
				if (map.snp_Sc[dr][l] >= gapHeteroCrv * 0.9) {
					map.setSnp(dr-1, l, true);
				} else {
					map.setSnp(dr-1, l, false);
				}
			} else {
				map.type[1][l] = (byte) (map.type[1][l] & (byte) (0 - (1+dr)));
			}
		} //*/
    	return;
    }
    
    public void scoreSNP(int l) {
    	/*
    	map.setSnp(l, false);
    	nzCvr = map.snp_Sc[1][l] + map.snp_Sc[2][l];
		if (nzCvr <= 0) {
			map.wsnpp[l] = 0;
			return;
		} 
		getMPB(l);
		float ratDirSNP = map.snp_Sc[1][l] / nzCvr; 
		if (ratDirSNP < Parameters.snpDirected) {
			map.setDirection(l, 'F');
		} else if (ratDirSNP > (1-Parameters.snpDirected)) {
			map.setDirection(l, 'R');
		} else { 
			map.setDirection(l, 'B');
		}
		map.setMPBP(l, mpbp);
   		map.wsnpp[l] =  (map.Pb[0][mpbp][l] * map.snp_Sc[1][l] + map.Pb[1][mpbp][l] * map.snp_Sc[2][l]) / (map.snp_Sc[0][l] + map.snp_Sc[1][l] + map.snp_Sc[2][l]);
    	return; //*/
    }
    
    public  void snpFileTop(BufferedWriter writer) throws Exception {
		writer.write("Dataset: " + Parameters.DATASET + "\tDatabase: " + Parameters.DATABASE + "\n");
		writer.write("coverage of Ref %:");
		for(int cv = 0; cv < map.avrCrv.length; cv++) {
			writer.write("\t" + Parameters.trim(map.avrCrv[cv]));
		}
		writer.write("\n");
		writer.write("coverage X:");
		for(int cnt = map.avrCnt.length - 1; cnt >= 0; cnt--) {
			writer.write("\t" + Parameters.trim(map.avrCnt[cnt]));
		}
		writer.write("\n");
    	writer.write("SNPid\tLocat\tRef\tMPB\tCount\tInfo\tP(A)\tP(C)\tP(G)\tP(T)\tbidx\tmap\n");
    	writer.write("-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\n");
    }
    
    public  void edgesFileTop(BufferedWriter writer) throws Exception {
		writer.write("Dataset: " + Parameters.DATASET + "\tDatabase: " + Parameters.DATABASE + "\n");
		writer.write("coverage of Ref %:");
		for(int cv = 0; cv < map.avrCrv.length; cv++) {
			writer.write("\t" + Parameters.trim(map.avrCrv[cv]));
		}
		writer.write("\n");
		writer.write("coverage X:");
		for(int cnt = map.avrCnt.length - 1; cnt >= 0; cnt--) {
			writer.write("\t" + Parameters.trim(map.avrCnt[cnt]));
		}
		writer.write("\n");
    	writer.write("Direct\tEID\tStarts\tSize\tCover\t#SNPs\tChange\tgCnt\tSequence\n");
    	// writer.write("\n");
    }

    public  void displaySNPs(String path, String file, int chr, BufferedWriter writerSNPs) throws Exception {
    	//*
		System.gc();
    	int sid;
    	String fl = path + "ScoreResults/" + file;
    	int eId       = 0;
    	int[] countSNPs = {0, 0};
    	int[] countBases= {0, 0};
    	int[] countCrv  = {0, 0};
    	int[] countCrvRf= {0, 0};
    	int[] pL        = {-1, -1};
    	int pH          = 0;
    	int[] rL        = {-1, -1};
    	int[] oRL       = {-1, -1};
		int[] weight    = {0, 0};
		int[] weightRf  = {0, 0};
		char oBase = ' ';
		int   countSNPs2= 0;
		int   countBoth = 0;
    	int bpi;
		BufferedWriter writerEdges = new BufferedWriter(new FileWriter(fl + ".edges"));
		System.err.println(fl + ".edges");
		edgesFileTop(writerEdges);
    	for (pH = 0; pH < map.size; pH++) {
    		for (int direction = 0; direction < 2; direction++) {
        		if (map.cnt[direction][pH] > 0) {
        			countCrv[direction]   += map.cnt[direction][pH];
        			countCrvRf[direction] += map.cnt[2][pH];
        			countBases[direction]++;
        			if (map.isSnp(direction, pH))
        				countSNPs[direction]++;
           			if (pL[direction] == -1)
        				pL[direction] = pH;
        		} else { 
        			weight[direction]   = Math.round(((float) countCrv[direction]) / countBases[direction]);
        			weightRf[direction] = Math.round(((float) countCrvRf[direction]) / countBases[direction]);
            		if (countSNPs[direction] > 0 && pH > pL[direction] + 1) {
            			countSNPs2 = 0;
            			countBoth  = 0;
            			for (int s = pL[direction]; s < pH; s++) {
            				if (map.isSnp(2,s)) {
            					countSNPs2++;
                				if (map.isSnp(direction,s)) {
                					countBoth++;
                				}
            				}
            			}
            			if (!(countSNPs2 == 1 && countSNPs[direction] == 1 && countBoth == 1)) {
            				if (countSNPs2 > 0) { // ------------------------------ mark SNPs
                    			for (int s = pL[direction]; s < pH; s++) {
                    				if (map.isSnp(2,s)) {
                    					map.setSnpEdge(s, true);
                    				}
                    			}
            				}
                			eId++;
                			if (direction == 0) {
                				writerEdges.write("->>>\t");
                			} else {
                				writerEdges.write("<<<-\t");	            				
                			}
                			writerEdges.write(eId + "\t" + (pL[direction]+1) + "\t" + countBases[direction] + "\t" + weight[direction] + "\t" + 
                					countSNPs[direction] + "\t");
                			writerEdges.write(Math.round(((float) map.gval[pL[direction]+1]) / map.gcnt[pL[direction]+1] - Parameters.petSpanMost + (2 * Parameters.seedSize)) + "\t" + map.gcnt[pL[direction]+1] + "\t");
                			pL[direction]--;
                			rL[direction] = pL[direction];
                			while (pL[direction] <= pH) { 
                				bpi = map.getMPBPInt(direction, pL[direction]);
                				if (bpi < 0) {
                					if (direction == 0 && pL[direction] < pH)
                						bpi = 5;
                					else if (direction == 1 && pL[direction] == pH)
                						bpi = 6;
                					else
                						bpi = 7;
                				}
                				writerEdges.write(nucl[bpi]);
                				pL[direction]++;
                			}
                			writerEdges.write("\n");
                			// ------------------ end Edge,   start normal
                			oRL[direction] = rL[direction];
                			if (direction == 0) {
                				writerEdges.write("->>\t");
                			} else {
                				writerEdges.write("<<-\t");	            				
                			}
                			writerEdges.write("Obs\t" + (rL[direction]+1) + "\t" + (countBases[direction] + 2) + "\t" + weightRf[direction] + "\t" + countSNPs2 + "\t-\t-\t");
                			while (rL[direction] <= pH) { 
                				if (map.cnt[2][rL[direction]] > 1) {
    	            				bpi = map.getMPBPInt(2, rL[direction]);
    	            				if (bpi < 0) {
    	            					bpi = 7;
    	            				}
    	            				oBase = nucl[bpi];
    	            				if (!map.isSnp(2, rL[direction])) 
    	            					oBase = Character.toLowerCase(oBase);
                				} else {
                					oBase = '-';
                				}
                				writerEdges.write(oBase);
                				rL[direction]++;
                			}	            			
                			rL[direction] = oRL[direction];
                			writerEdges.write("\n");
                			// ------------------ end normal, start Reference
                			if (direction == 0) {
                				writerEdges.write("->\t");
                			} else {
                				writerEdges.write("<-\t");	            				
                			}
                			writerEdges.write("Ref\t" + (rL[direction]+1) + "\t" + (countBases[direction] + 2) + "\t-\t-\t-\t-\t");
                			while (rL[direction] <= pH) { 
                				writerEdges.write(map.charSeq[rL[direction]]);
                				rL[direction]++;
                			}
                			writerEdges.write("\n");
                		}
            		}
            		countBases[direction]= 0;
        			countSNPs[direction] =  0;
        			countCrv[direction]  =  0;
        			countCrvRf[direction]  =  0;
        			pL[direction]        = -1;
        		}
	    	}
		}
    	writerEdges.close();
    	BufferedWriter writerCover = null;
    	if (Parameters.displayCover) {
    		writerCover = new BufferedWriter(new FileWriter(fl + ".cover"));
    		System.err.println(fl + ".cover");
    	}
		BufferedWriter writer     = new BufferedWriter(new FileWriter(fl + ".snp"));
		System.err.println(fl + ".snp");
		snpFileTop(writer);
    	sid = 0;
    	for (int i = 0; i < map.size; i++) {
			if (map.isSnp(2, i)) { 
				sid++;
				displaySNP(writer, i, sid + "");
        	}			
			if (Parameters.displayCover) {
				displayCoverPoint(writerCover, i);
			}
    	}
    	writer.close();
    	writerCover.close();
    }
    
    private void displayCoverPoint(BufferedWriter writer0, int i) throws Exception {
		writer0.write((i + 1) + "\t" + map.charSeq[i] + "\t");
		for (int dr = 2; dr >= 0; dr--) {
			writer0.write(map.cnt[dr][i] + "");
			for (int b = 0; b < 4; b++) {
				writer0.write("\t" + Parameters.trim(map.Pb[dr][b][i]));
			}
			if (dr > 0)
				writer0.write("\t|\t");
		}
		writer0.write("\n");
    }
    private void displaySNP(BufferedWriter writer, int i, String sid) throws Exception {
    	writer.write(sid);
		writer.write("\t" + (i + 1) + "\t");
		if (i>1)
			writer.write(Character.toLowerCase(map.charSeq[i-2]));
		else
			writer.write(" ");
		if (i>0)
			writer.write(Character.toLowerCase(map.charSeq[i-1]));
		else
			writer.write(" ");
		writer.write(Character.toUpperCase(map.charSeq[i]));
		if (i< map.size-1)
			writer.write(Character.toLowerCase(map.charSeq[i+1]));
		else
			writer.write(" ");
		if (i< map.size-2)
			writer.write(Character.toLowerCase(map.charSeq[i+2]) + "\t");
		else
			writer.write(" \t");
		writer.write(map.getMPBP(2, i) + "\t");
		writer.write(map.cnt[2][i] + "\t");
		writer.write(Parameters.trim(map.info[i]/map.cnt[2][i]) + "\t");
		for(int b = 0; b < 4; b++) {
			writer.write("" + Parameters.trim(map.Pb[2][b][i]));
			if (b < 3)
				writer.write("\t");
		}
		writer.write("\t" + Math.round(map.bidx[i]) + "\t" + map.refMap[i]);
		if (map.isSnpEdge(i)) {
			writer.write("\tEdge\n");
		} else {
			writer.write("\n");
		}
		if (map.bidx[i] >=0 && map.bidx[i] < bidxHist.length) {
			bidxHist[Math.round(map.bidx[i])]++;
		}
    }
    
    private  void compareSNPs1(String path, String file, String multiPath) throws Exception {
    	/*
		System.gc();
    	int sid;
    	long[] xx = new long[4];
    	long[] ax = new long[4];
    	int x;
    	int z;
    	String fl = multiPath + "/" + file; // <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    	ax[0] = ax[1] = xx[0] = xx[1] = xx[2] = 0;
    	for (int i = 0; i < map.size - Parameters.seedSize; i++) {
    		if ((map.snp_Sc[0][i] + map.snp_Sc[1][i] + map.snp_Sc[2][i]) > 0) {
    			xx[3]++;
    		}
    		for (int l = 0; l < 3; l++) {
    			if (map.snp_Sc[l][i] > 0) {
    				xx[l]++;
    				ax[l] += map.snp_Sc[l][i];
    			}
    		}
    	}
    	String sapFl = Parameters.pathDB + Parameters.DATABASE + file + ".map." + (Parameters.seedSize + Parameters.mapAdjust); 
    	FileInputStream 	fis    = new FileInputStream(sapFl);
    	BufferedInputStream bis    = new BufferedInputStream(fis, 127000);
    	DataInputStream		strIns = new DataInputStream(bis);
		for (int i = 0; i < map.size - Parameters.seedSize; i++) {
			map.refMap[i] = strIns.readByte();
		}
		strIns.close();
		bis.close();
		fis.close();

		System.err.println(fl + ".snp");
		BufferedWriter writer = new BufferedWriter(new FileWriter(fl + ".snp"));
    	writer.write("total  coverage:\t(\t" + ax[0] + "\t+\t" + ax[1] + "\t+\t" + ax[2] + "\t) =\t" + (ax[0] + ax[1] + ax[2]) + "\tX =\t" + Parameters.trim((float) (ax[0] + ax[1] + ax[2]) / (float) (map.size - Parameters.seedSize)) + "\n");
    	writer.write("actual coverage:\t(\t" + xx[0] + "\t+\t" + xx[1] + "\t+\t" + xx[2] + "\t) =\t" + (xx[0] + xx[1] + xx[2]) + "\tX =\t" + Parameters.trim((float) (xx[3]) / (float) (map.size - Parameters.seedSize)) + "\n");
    	writer.write("Location\tmap\tcmCount\tmCount\tnmCount\twSNP%\tcSNP%\tncSNP%\tRef\tMPB\n");
    	writer.write("-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\t-----\n");
    	sid = 0;
    	float wsnpp;
    	for (int i = 0; i < map.size; i++) {
    		getMPB(i);
    		wsnpp = ((( (float) map.snp_Sc[1][i] + map.snp_Sc[2][i]) * map.Pb[0][mpbp][i] / 100) / 
    				((float) map.snp_Sc[0][i] + map.snp_Sc[1][i] + map.snp_Sc[2][i]));
			z = getBpInt(map.charSeq[i]); 
			if(z < 0) {
				System.err.println("z:\t" + z + "\tchar\t" + map.charSeq[i]);
			}
			getMPB(i);

			if (mpbp == z)
				mpbp = npbp;
			writer.write((i+1) + "\t" + map.refMap[i] + "\t" + map.snp_Sc[0][i] + "\t" + map.snp_Sc[2][i] + "\t" + map.snp_Sc[1][i] + "\t");
			writer.write(Parameters.trim(100 * wsnpp) + "\t");
			writer.write(Parameters.trim(100 * (map.snp_Sc[1][i] + map.snp_Sc[2][i]) / (map.snp_Sc[0][i] + map.snp_Sc[1][i] + map.snp_Sc[2][i])) + "\t");
			writer.write(Parameters.trim(100 * ((float) map.Pb[1][mpbp][i]) / ((float) map.Pb[1][getBpInt(map.charSeq[i])][i] + map.Pb[1][mpbp][i])) + "\t");
			writer.write(map.charSeq[i] + "\t" + nucl[mpbp]);
			sid++;
			writer.write("\n");   		
    	}
    	writer.close();
    	System.runFinalization();
		System.gc();
		//*/
    } // --------------------------------------------------------------------------------------- compareSNPs1
//*
    private int  getAddress() throws Exception {
    	int x = 10000000;
    	String line = reader.readLine(); 
		if (line != null) {
			StringTokenizer tokens = new StringTokenizer(line);
			// tokens.nextToken();
			x = Integer.parseInt(tokens.nextToken(), 10);
		} else {
			x = -1;
		}
		x--;
    	return x;
    }
//*/
	private class Sequencability {
		byte[] C_map;
		public Sequencability(int size) {
			C_map = new byte[size];
    		for (int j = 0; j < size; j++) {
    			C_map[j] = 0;
    		}
 		}
	}
//*
    public  void dbChart(String fdb) throws Exception {
    	long count = 0;
    	float[] fHist = new float[Parameters.dbHistSize];
    	long[][] hist = new long[2][Parameters.dbHistSize];
    	for (int i = 0; i < Parameters.dbHistSize; i++) {
    		hist[0][i] = hist[1][i] = 0;
    		fHist[i] = 0;
    	}
    	OligoFile ofile = new OligoFile(fdb, "", OligoFile.R);
  		while(ofile.readOligo(oligo[DATA])) {
  			count++;
  			if (count % 10000000 == 0)
  				System.err.println(count/1000000 + "M");
  			if (oligo[DATA].u0 < Parameters.dbHistSize) {
  				hist[0][oligo[DATA].u0]++;
  			} else if (oligo[DATA].u0 >= Parameters.dbHistSize) {
  				hist[0][Parameters.dbHistSize-1]++;
  			} else {
  				System.err.println("Error in dbChart:\t" + oligo[DATA]);
  			}
  		}
		ofile.close();
		System.err.println(fdb + "/dbHist.txt");
    	BufferedWriter writer = new BufferedWriter(new FileWriter(fdb + "/dbHist.txt"));
    	for (int i = 1; i < Parameters.dbHistSize; i++) {
    		hist[1][i] = hist[0][i] + hist[1][i-1];
    		fHist[i] = (float) ((double) hist[1][i] / count);
    	}
    	for (int i = 0; i < Parameters.dbHistSize; i++) {
    		writer.write((i + 1) + "\t" + hist[0][i] + "\t" + hist[1][i] + "\t" + fHist[i] + "\n");
    	}
    	writer.close();
    }
    //*/
    public  void sequencability(String fdb, String[] seqNames, int[] seqSize, int sz) throws Exception {
    	long ofidx = 0;
    	Sequencability[] sq = new Sequencability[seqNames.length];
    	for (int i = 0; i < seqNames.length; i++) {
    		sq[i] = new Sequencability(seqSize[i]);
    	}
    	if (display == true)
    		System.err.println("sequencability:\t" + fdb + "\tSize:\t" + sz);
    	// int        x = 2;
    	long[] histo   = new long[2 * sz + 1];
		for (int j = 0; j < histo.length; j++)
			histo[j] = 0;
		// System.err.println("Will be Writing map to:\t" + fdb + seqNames[0] + ".map." + sz);
		ofidx = 0;
    	OligoFile ofile = new OligoFile(fdb, "", OligoFile.R);
  		while(ofile.readOligo(oligo[DATA])) { // ------------------------------------------------------
   			ofidx++;
   			if (ofidx % 100000000 == 0)
   				System.err.println("\t" + (ofidx/100000000) + "00,000,000");
   			//*
			if (oligo[DATA].isRC()) {
				if(oligo[DATA].u0 < sz) {
					for (int i = 0; i < sz; i++) {
						sq[oligo[DATA].ch].C_map[oligo[DATA].start + Parameters.dbOligoSize - i - 1] += (byte) (oligo[DATA].weight + 0.2);
					}
				}								
			} else {
				if(oligo[DATA].u0 < sz) {
					for (int i = 0; i < sz; i++) {
						sq[oligo[DATA].ch].C_map[oligo[DATA].start + i] += (byte) (oligo[DATA].weight + 0.2);
					}
				}				
			}
			//*/
		}		
		for (int ifile = 0; ifile < seqNames.length; ifile++) {
			for (int j = 0; j < seqSize[ifile]; j++) {
				if (sq[ifile].C_map[j] > ( 2 * sz)) {
					System.err.println("i0\t" + sq[ifile].C_map[j] + ",\tchr\t" + ifile + "\tidx\t" + j);
				} else {
					histo[sq[ifile].C_map[j]]++;
				}
			}
		}
		
		ofile.close();	
		// ----------------------------------------------------------------------- Writing Results
		for (int ifile = 0; ifile < seqNames.length; ifile++) {
			System.err.println("\nWriting map to:\t" + fdb + seqNames[ifile] + ".map." + sz);
			FileOutputStream     fwOne = new FileOutputStream(fdb + seqNames[ifile] + ".map." + sz);
			BufferedOutputStream bwOne = new BufferedOutputStream(fwOne);
			DataOutputStream     file  = new DataOutputStream(bwOne);			
			for (int i = 0; i < seqSize[ifile]; i++) {
				file.writeByte(sq[ifile].C_map[i]);
			}
		    file.close();
		    bwOne.close();
		    fwOne.close();
	 	}
		sq = null;
		System.gc();
		/*
		System.out.println("sequencability\t# of Noclutides");
		for (int i = 0; i < histo.length; i++) {
			System.out.println("\t" + i + "\t" + histo[i]);
		}
		//*/
    }

    public  void sequencability0(String fdb, String[] seqNames, int[] seqSize) throws Exception {
    	long ofidx = 0;
    	Sequencability[][] sq = new Sequencability[seqNames.length][2];
    	for (int i = 0; i < seqNames.length; i++) {
			// System.err.println("Chr:\t" + (i+1));
    		sq[i][Parameters.F] = new Sequencability(seqSize[i]);
    		sq[i][Parameters.R] = new Sequencability(seqSize[i]);
    	}
		ofidx = 0;
    	OligoFile ofile = new OligoFile(fdb, "", OligoFile.R);
  		while(ofile.readOligo(oligo[DATA])) { // ------------------------------------------------------
   			ofidx++;
   			if (ofidx % 100000000 == 0)
   				System.err.println("\t" + (ofidx/100000000) + "00,000,000");
   			//*
			if (oligo[DATA].isRC()) {
				// System.err.println(oligo[DATA].start + Parameters.dbOligoSize + "\t" + sq[oligo[DATA].ch][Parameters.R].C_map.length);
				if(sq[oligo[DATA].ch][Parameters.R].C_map[oligo[DATA].start + Parameters.dbOligoSize - 1] < oligo[DATA].u0) {
					sq[oligo[DATA].ch][Parameters.R].C_map[oligo[DATA].start + Parameters.dbOligoSize - 1] = oligo[DATA].u0;
				}								
			} else {
				if(sq[oligo[DATA].ch][Parameters.F].C_map[oligo[DATA].start] < oligo[DATA].u0) {
					sq[oligo[DATA].ch][Parameters.F].C_map[oligo[DATA].start] = oligo[DATA].u0;
				}				
			}
			//*/
		}		
		ofile.close();	
		// ----------------------------------------------------------------------- Writing Results
		for (int ifile = 0; ifile < seqNames.length; ifile++) {
			FileOutputStream     fwOne = new FileOutputStream(fdb + seqNames[ifile] + ".u0");
			BufferedOutputStream bwOne = new BufferedOutputStream(fwOne);
			DataOutputStream     file  = new DataOutputStream(bwOne);			
			for (int i = 0; i < seqSize[ifile]; i++) {
				file.writeByte(sq[ifile][Parameters.F].C_map[i]);
				file.writeByte(sq[ifile][Parameters.R].C_map[i]);
			}
		    file.close();
		    bwOne.close();
		    fwOne.close();
	 	}
		sq = null;
		System.gc();
    }

    /*
    private  void filterRedundant(String fin, String fout) throws Exception {
    	if (display == true)
    		System.err.println("\n -------- sortFilterRedundant " + fin);
    	Oligo oligo2 = new Oligo();
    	ofOne    = new OligoFile(fin, "", OligoFile.R);
    	ofTwo    = new OligoFile(fout, "", OligoFile.W);
		readMeta(ofOne);     // ----------------------------------------------------------
		clearMeta();
		idx = 0;
		bunk[0].clear();
		oligo2.id1 = -100;
		while(ofOne.available() > 0) { 
			ofOne.readOligo(oligo[DATA]); 
			if (oligo[DATA].id1 != oligo2.id1) {
				if (oligo[DATA].isUQ())
					numberOfUQ++;
				else 
					numberOfUD++;
				if (oligo[DATA].isUM()) {
					weightOfUM += oligo[DATA].weight;
				} else {
					weightOfMM += oligo[DATA].weight;
				}
				ofTwo.writeOligo(oligo[DATA]);
				idx++;
				oligo2 = new Oligo(oligo[DATA]);
				numberOfOligos++;
			}
		}
		ofOne.close(); 
		ofTwo.close();
	    // System.out.println("From file " + fin + " removed oligos \t" + (oldNumberOfOligos - numberOfOligos));
	    if (display == true)
	    	System.err.println(" -------- sortFilterRedundant " + fin + " End\n");
	
    }

    //*/

    public int split(String line) {
		int rdMxSz;
    	StringTokenizer tokens = new StringTokenizer(line);
    	int lineSize = tokens.countTokens();
		if (Parameters.ETag == Parameters.PET) {
			if (Parameters.skipCycles > 0) {
				rdMxSz =  (lineSize/4)/2 - Parameters.skipCycles;
			} else {
				rdMxSz =  (lineSize/4)/2;
			}
	    	seq.prbSize = rdMxSz;
			for (int sk = 0; sk < Parameters.skipCycles; sk++) { 			// ------------------------------------------------ skip
		    	for (int j = 0; j < 4; j++) {
		    		tokens.nextToken();
		    	}
			}
	    	for (int i = 0; i < rdMxSz; i++) {
	    		for (int j = 0; j < 4; j++) {
	       			seq._q[0][i][j] = (byte) Integer.parseInt(tokens.nextToken(), 10);
	    		}
	    	}    	   	
			for (int sk = 0; sk < Parameters.skipCycles; sk++) { 			// ------------------------------------------------ skip
		    	for (int j = 0; j < 4; j++) {
		    		tokens.nextToken();
		    	}
			}
	    	for (int i = 0; i < rdMxSz; i++) {
	    		for (int j = 0; j < 4; j++) {
	    			seq._q[1][i][j] = (byte) Integer.parseInt(tokens.nextToken(), 10);
	    		}
	    	}    	  
		} else {
	    	rdMxSz =  (lineSize/4) - Parameters.skipCycles;
	    	seq.prbSize = (short) rdMxSz;
			for (int sk = 0; sk < Parameters.skipCycles; sk++) { 			// ------------------------------------------------ skip
		    	for (int j = 0; j < 4; j++) {
		    		tokens.nextToken();
		    	}
			}
	    	for (int i = 0; i < rdMxSz; i++) {
	    		for (int j = 0; j < 4; j++) {
	       			seq._q[0][i][j] = (byte) Integer.parseInt(tokens.nextToken(), 10);
	    		}
	    	}    	   				
		}
		// seq.size = Parameters.seedSize;
    	return rdMxSz;
    }
 
    private void getMeta(OligoFile of) {
    	numberOfOligos = of.numberOfOligos;
    	weightOfUM     = of.weightOfUM;
    	weightOfMM     = of.weightOfMM;
    	numberOfUQ     = of.numberOfUQ;
    	numberOfUD     = of.numberOfUD;
    }
    private void readMeta(OligoFile of) throws Exception {
		numberOfOligos = of.numberOfOligos;
		numberOfUQ     = of.numberOfUQ; 
		numberOfUD     = of.numberOfUD; 
		weightOfUM     = of.weightOfUM;
		weightOfMM     = of.weightOfMM;
		if (display == true) {
			System.out.println(
					"Number of Us ...:\t" + "\n" +
					"Number of Oligos:\t" + numberOfOligos + "\n" + 
					"Number of UQ ...:\t" + numberOfUQ + "\n" + 
					"Number of UD ...:\t" + numberOfUD + "\n" + 
					"Weight of Oligos:\t" + Parameters.trim(weightOfUM  + weightOfMM) + "\n" + 
					"Weight of UM ...:\t" + Parameters.trim(weightOfUM) + "\n" + 
					"Weight of MM ...:\t" + Parameters.trim(weightOfMM) + "\n" +
					"Ratio UM/MM ....:\t" + ((float) weightOfUM / (float) weightOfMM));
		}
    }  
    private void clearMeta() {
    	numberOfOligos = 0;
    	weightOfUM     = 0;
    	weightOfMM     = 0;
    	numberOfUQ     = 0;
    	numberOfUD     = 0;
    }
	// rs.createMap(<<Parameters.pathRefSeq + files[i]>>, <<files[i]>>, i);
	public  void createMap(String fl, String outf, int fInt) throws Exception {
		map = new Map(refSeqSize[fInt], fl + outf, outf, fInt);
	}

	/*
    public void fillU(String fl) throws Exception {
		System.err.println(" fillU " + fl);
    	Oligo oligo0;
    	// int u0 = 0;
    	byte u1;
    	byte[] l = new byte[2];
    	openStreamIn(fl + ".db.s.sq", DATA);
		readMeta(DATA, fl + ".db.s.sq");
		idx = 0;
		bunk[0].clear();
		while(strIn[DATA].available() > 0) {
			readOligoEx(DATA);
			addOligo(0);
			idx++;
		}
		
		closeStreamIn(DATA);
	    // int k = 1;
		if (Oligo.nou > 1)
	    for ( int u = 0; u < oligoSize; u++) {
	    	System.err.println("itation:\t" + u + "/" + (oligoSize-1));
	    	Oligo.clearMask();
	    	Oligo.setMask(u);
	    	Collections.sort(bunk[0]); 
	    	Iterator it = bunk[0].iterator();
	    	l[1] = 0;
	    	oligo0 = (Oligo) it.next ();
	    	for (; it.hasNext(); ) {
	    		oligo[DATA] = (Oligo) it.next ();
	    		l[0] = l[1];
	    		l[1] = oligo[DATA].match(oligo0);
	    		if (l[0] > l[1]) {
	    			u1 = l[0];
	    		} else {
	    			u1 = l[1];
	    		}
	    		if (u1 > oligo0.u[1])
	    			oligo0.u[1] = u1;
	    		oligo0 = oligo[DATA];
	    	}
	    	if (Oligo.nou > 2) {
	    	    for ( int u2 = u+1; u2 < oligoSize; u2++) {
	    	    	Oligo.clearMask();
	    	    	Oligo.setMask(u);
	    	    	Oligo.setMask(u2);
	    	    	Collections.sort(bunk[0]); 
	    	    	Iterator it2 = bunk[0].iterator();
	    	    	l[1] = 0;
	    	    	oligo0 = (Oligo) it2.next ();
	    	    	for (; it2.hasNext(); ) {
	    	    		oligo[DATA] = (Oligo) it2.next ();
	    	    		l[0] = l[1];
	    	    		l[1] = oligo[DATA].match(oligo0);
	    	    		if (l[0] > l[1]) {
	    	    			u1 = l[0];
	    	    		} else {
	    	    			u1 = l[1];
	    	    		}
	    	    		if (u1 > oligo0.u[2])
	    	    			oligo0.u[2] = u1;
	    	    		oligo0 = oligo[DATA];
	    	    	}
	    	    	if (Oligo.nou > 3) {
	    	    	    for ( int u3 = u+1; u3 < oligoSize; u3++) {
	    	    	    	Oligo.clearMask();
	    	    	    	Oligo.setMask(u);
	    	    	    	Oligo.setMask(u2);
	    	    	    	Oligo.setMask(u3);
	    	    	    	Collections.sort(bunk[0]); 
	    	    	    	Iterator it3 = bunk[0].iterator();
	    	    	    	l[1] = 0;
	    	    	    	oligo0 = (Oligo) it3.next ();
	    	    	    	for (; it3.hasNext(); ) {
	    	    	    		oligo[DATA] = (Oligo) it3.next ();
	    	    	    		l[0] = l[1];
	    	    	    		l[1] = oligo[DATA].match(oligo0);
	    	    	    		if (l[0] > l[1]) {
	    	    	    			u1 = l[0];
	    	    	    		} else {
	    	    	    			u1 = l[1];
	    	    	    		}
	    	    	    		if (u1 > oligo0.u[3])
	    	    	    			oligo0.u[3] = u1;
	    	    	    		oligo0 = oligo[DATA];
	    	    	    	}
	    	    	    }
	    	    	}

	    	    }
	    	}
	    }
	    
	    openStreamOut(fl + ".db.s.sq", DATA);
	    saveBunk(DATA);
	    closeStreamOut(DATA);
	    
		System.err.println(" fillU End\n");
    } 

//*/	
	public  void test(String umf) throws Exception {
		Sequencability[] tt;
		tt = new Sequencability[100];
		for (int i = 0; i < 100; i++) {
			tt[i] = new Sequencability(i);
		}
	}
}

